/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

/*

This experiment demonstrates building a live model with QtSparql and QtSparqlTracker.
The model is highly specialized to a specific query and cannot be used as a
general-purpose live model.

Other assumptions made for this experiment:
- It's ok to do one synchronous query on startup to get ontology data
  (tracker:id's of classes and properties)
- When something related to a contact changes, it's ok to query all the data
  related to that contact again. (E.g., if the name changes, it's ok to also
  query the phone number again.)
- If the phone number of a contact changes, a new nco:PhoneNumber object is
  created and the contact's hasPhoneNumber property is changed to point into
  that object. (So, the nco:phoneNumber property of nco:PhoneNumber never
  changes.
- We query only one phone number per contact.  In real life there can be
  multiple phone numbers.  We would need a subquery to get only one phone number
  per contact.
- We ignore the collation.

*/

#include <TrackerChangeNotifier>
#include <QtSparql>

#include <QObject>
#include <QDebug>
#include <QStandardItemModel>

// How many contacts are queried in the same query if there are lots of changes
// at the same time. A small batch size is only for demonstration purposes (we
// don't need to add so many contacts to get multiple batches). In real life,
// the batch size needs to be small enough so that the limit of variables in a
// query is not exceeded.
#define BATCH_SIZE 10

class LiveContacts : public QStandardItemModel
{
    Q_OBJECT
public:
    LiveContacts()
        : conn("QTRACKER_DIRECT"),
          initialResult(0), initialQueryDone(false),
          changeNotifier("http://www.semanticdesktop.org/ontologies/"
                         "2007/03/22/nco#PersonContact"),
          basicQuery("SELECT tracker:id(?u) "
                     "fn:string-join((?ng, ?nf), \" \") "
                     "?n "
                     "{ ?u a nco:PersonContact . "
                     "OPTIONAL { ?u nco:nameGiven ?ng . } "
                     "OPTIONAL { ?u nco:nameFamily ?nf . } "
                     "OPTIONAL { ?u nco:hasPhoneNumber ?pn . "
                     "?pn nco:phoneNumber ?n } "
                     "FILTER(bound(?ng) || bound(?nf)) "
                     "%1 } ORDER BY ?ng ?nf") // %1 is for additional FILTER

    {
        // Initial query: what are the current contacts?
        if (!conn.isValid()) {
            qWarning() << "Driver not found, available drivers:";
            qWarning() << QSparqlConnection::drivers();
            return ;
        }

        // Start listening to the changes related to the PersonContact class
        qDebug() << "Connecting tracker signal: " << connect(&changeNotifier,
                SIGNAL(changed(QList<TrackerChangeNotifier::Quad>,
                               QList<TrackerChangeNotifier::Quad>)),
                SLOT(changed(QList<TrackerChangeNotifier::Quad>,
                             QList<TrackerChangeNotifier::Quad>)));

        // Do a synchronoues query to get ontology info
        QSparqlQuery ontoQuery("select tracker:id(nco:PersonContact) "
                               "tracker:id(rdf:type) "
                               "tracker:id(nco:nameGiven) "
                               "tracker:id(nco:nameFamily) "
                               "tracker:id(nco:hasPhoneNumber) { }");
        QSparqlResult* ontoResult = conn.exec(ontoQuery);
        ontoResult->waitForFinished();
        if (ontoResult->hasError())
            qWarning() << "ontology query failed:" <<
                ontoResult->lastError().message();
        else if (!ontoResult->next())
            qWarning() << "no ontology data";
        else {
            personContactId = ontoResult->value(0).toInt();
            typeId = ontoResult->value(1).toInt();
            nameGivenId = ontoResult->value(2).toInt();
            nameFamilyId = ontoResult->value(3).toInt();
            hasPhoneNumberId = ontoResult->value(4).toInt();
        }
        delete ontoResult;

        QSparqlQuery query(basicQuery.arg(""));
        qDebug() << "Running initial query:\n" << query.preparedQueryText();
        initialResult = conn.exec(query);
        connect(initialResult, SIGNAL(finished()),
                SLOT(initialQueryFinished()));

        connect(this, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
                SLOT(printModel()));
        connect(this, SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                SLOT(printModel()));
        connect(this, SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
                SLOT(printModel()));
    }
private:
    void deleteContact(int id)
    {
        if (idToRow.contains(id)) {
            if (idToRow[id].isValid())
                removeRow(idToRow[id].row());
            idToRow.remove(id);
        }
        //qDebug() << idToRow;
    }
    void insertContact(int id, const QString& name, const QString& phone)
    {
        // If the contact already exists, just update it.  But if the name of
        // the contact has changed, then its position in the model must change,
        // too.
        if (idToRow.contains(id) && idToRow[id].isValid()) {
            int row = idToRow[id].row();
            if (item(row, 1)->text() == name) {
                setItem(row, 2, new QStandardItem(phone));
                return;
            }
            else {
                removeRow(row);
                idToRow.remove(id);
            }
        }

        // Look for a row *before which* the new item will be inserted. Here we
        // use the knowledge that the model is ordered by name (secondary
        // ordering: tracker id), and search the correct place for the insertion
        // with binary search.

        // When searching the correct place, the ordering criteria need to be
        // exactly the same as the ordering criteria for the initial data.
        // (Note: collation.)

        int start = 0, end = rowCount();
        int mid;
        while (start < end) {
            mid = (start + end) / 2;
            int res = QString::compare(item(mid, 1)->text(), name);
            if (res == 0) {
                int oldId = item(mid, 0)->text().toInt();
                if (oldId == id) {
                    // This shouldn't happen; idToRow should contain id in this
                    // case.
                    qWarning() << "Unexpected existence of a contact to add";
                    return;
                }
                if (oldId < id)
                    start = mid;
                else
                    end = mid;
            }
            else if (res < 0) {
                if (mid == start) // avoid endless loop here
                    start = end;
                else
                    start = mid;
            }
            else
                end = mid;
        }

        insertRow(start);
        setItem(start, 0, new QStandardItem(QString::number(id)));
        setItem(start, 1, new QStandardItem(name));
        setItem(start, 2, new QStandardItem(phone));
        idToRow.insert(id, QPersistentModelIndex(index(start, 0)));
        //qDebug() << idToRow;
    }

private Q_SLOTS:
    void initialQueryFinished()
    {
        initialQueryDone = true;
        if (initialResult->hasError())
            qWarning() << initialResult->lastError().message();
        else {
            int row = 0;
            while (initialResult->next()) {
                setItem(row, 0,
                        new QStandardItem(
                            initialResult->value(0).toString()));
                setItem(row, 1,
                        new QStandardItem(
                            initialResult->value(1).toString()));
                setItem(row, 2,
                        new QStandardItem(
                            initialResult->value(2).toString()));
                idToRow[initialResult->value(0).toInt()] =
                    QPersistentModelIndex(index(row, 0));
                ++row;
            }
        }
        initialResult->deleteLater();
        initialResult = 0;

        // React if some change signals have arrived before the initial query
        processChanges(pendingDeletes, pendingInserts);
        pendingDeletes.clear();
        pendingInserts.clear();

        //qDebug() << idToRow;
    }

    void addedQueryFinished()
    {
        QSparqlResult* result = qobject_cast<QSparqlResult*>(QObject::sender());
        if (result->hasError())
            qWarning() << result->lastError().message();
        else {
            while (result->next()) {
                insertContact(result->value(0).toInt(),
                              result->value(1).toString(),
                              result->value(2).toString());
            }
        }
        result->deleteLater();
    }

    /// Goes through the list of deleted and inserted triples.  Does the
    /// corresponding changes into our model.  If more data is needed (e.g.,
    /// names are literals and not included in the signal), constructs
    /// additional queries and executes them.
    void processChanges(QList<TrackerChangeNotifier::Quad> deletes,
                        QList<TrackerChangeNotifier::Quad> inserts)
    {
        qDebug() << "GraphUpdated:" << deletes << inserts;
        Q_FOREACH(const TrackerChangeNotifier::Quad& q, deletes) {
            if (q.predicate == typeId && q.object == personContactId) {
                // The contact has been deleted altogeter; remove from the model
                deleteContact(q.subject);
            }
        }
        QSet<int> changedContactsSet;
        Q_FOREACH(const TrackerChangeNotifier::Quad& q, inserts) {
            if (q.predicate == typeId &&
                q.object == personContactId) {
                // New contact was added
                changedContactsSet.insert(q.subject);
            }
            else if (q.predicate == nameGivenId ||
                     q.predicate == nameFamilyId ||
                     q.predicate == hasPhoneNumberId) {
                // Name was added
                changedContactsSet.insert(q.subject);
            }
        }
        // Query the added contacts...
        QList<int> changedContacts = changedContactsSet.toList();
        for (int j = 0; j < changedContacts.size(); j+= BATCH_SIZE) {
            QString filter = "FILTER (tracker:id(?u) in (";
            bool first = true;

            for (int i = j; i < j + BATCH_SIZE && i < changedContacts.size(); ++i) {
                if (!first)
                    filter.append(", ");
                first = false;
                filter.append(QString::number(changedContacts[i]));
            }
            filter.append("))");

            QSparqlQuery addedQuery(basicQuery.arg(filter));
            addedQuery.bindValue("FILTER", filter);
            qDebug() << "Running query:\n" << addedQuery.preparedQueryText();
            QSparqlResult* addedResult = conn.exec(addedQuery);
            connect(addedResult, SIGNAL(finished()),
                    SLOT(addedQueryFinished()));
        }
    }

    void changed(QList<TrackerChangeNotifier::Quad> deletes,
                 QList<TrackerChangeNotifier::Quad> inserts)
    {
        if (initialQueryDone) {
            processChanges(deletes, inserts);
        }
        else {
            pendingDeletes.append(deletes);
            pendingInserts.append(inserts);
        }
    }

    void printModel() const
    {
        qDebug() << "-----------";
        //qDebug() << rowCount() << columnCount();
        for (int row = 0; row < rowCount(); ++row) {
            for (int col = 0; col < columnCount(); ++col) {
                qDebug() << row << col << data(index(row, col)).toString();
            }
        }
    }

private:
    QSparqlConnection conn;
    QSparqlResult* initialResult;
    bool initialQueryDone;
    TrackerChangeNotifier changeNotifier;
    int personContactId, typeId, nameGivenId, nameFamilyId, hasPhoneNumberId;
    QList<TrackerChangeNotifier::Quad> pendingDeletes, pendingInserts;
    QString basicQuery;
    QHash<int, QPersistentModelIndex> idToRow; // tracker:id -> row in the model
};
