/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "trackerchangenotifier.h"
#include "trackerchangenotifier_p.h"

#include "trackerdbus_p.h"
#include "debug.h"

#include <QCoreApplication>
#include <QThreadStorage>
#include <QDebug>
#include <QDBusMetaType>

// D-Bus marshalling
QDBusArgument &operator<<(QDBusArgument &argument,
                          const TrackerChangeNotifier::Quad &t)
{
    argument.beginStructure();
    argument << t.graph << t.subject << t.predicate << t.object;
    argument.endStructure();
    return argument;
}

// D-Bus demarshalling
const QDBusArgument &operator>>(const QDBusArgument &argument,
                                TrackerChangeNotifier::Quad &t)
{
    argument.beginStructure();
    argument >> t.graph >> t.subject >> t.predicate >> t.object;
    argument.endStructure();
    return argument;
}

#ifndef QT_NO_DEBUG_STREAM
QDebug operator<<(QDebug dbg, const TrackerChangeNotifier::Quad &q)
{
    dbg.nospace() << "(" << q.graph << ", " << q.subject
                  << ", " << q.predicate << ", " << q.object << ")";
    return dbg.space();
}
#endif

/// Starts listening to GraphUpdated signals related to the ontolgy class \a
/// className. \a className needs to be the long name of the class, e.g.,
/// http://www.semanticdesktop.org/ontologies/2007/03/22/nco#PersonContact.
/// TrackerChangeNotifier will use the default session bus connection of QtDBus.
TrackerChangeNotifier::TrackerChangeNotifier(const QString& className,
                                             QObject* parent)
    : QObject(parent)
{
    qDBusRegisterMetaType<TrackerChangeNotifier::Quad>();
    qDBusRegisterMetaType<QList<TrackerChangeNotifier::Quad> >();

    d = new TrackerChangeNotifierPrivate(className, this);
}

TrackerChangeNotifier::~TrackerChangeNotifier()
{
}

TrackerChangeNotifierPrivate::TrackerChangeNotifierPrivate(
    const QString& className,
    TrackerChangeNotifier* q)
    : QObject(q), q(q), className(className),
      bufferActive(false), bufferSize(0)
{
    // Start listening to the actual signal
    using namespace TrackerDBus;
    bool ok =
        QDBusConnection::sessionBus().connect(service, resourcesPath,
                                              resourcesInterface, changedSignal,
                                              QStringList() << className,
                                              changedSignature,
                                              this, SLOT(changed(QString,
                                                                 QList<TrackerChangeNotifier::Quad>,
                                                                 QList<TrackerChangeNotifier::Quad>)));

    if (!ok) {
        TE_WARNING << "TrackerChangeNotifer: Cannot connect to the GraphUpdated signal";
    }
}

void TrackerChangeNotifierPrivate::emitPendingSignals()
{
    TE_DEBUG << "TrackerChangeNotifer: emitPendingSignals() - deletions" << deletedBuffer.size()
                << "insertions" << insertedBuffer.size();
    if(deletedBuffer.size() || insertedBuffer.size())
    {
        emit q->changed(deletedBuffer, insertedBuffer);
        deletedBuffer.clear();
        insertedBuffer.clear();
    }
}

/// Returns the name of the watched ontology class.
QString TrackerChangeNotifier::watchedClass() const
{
    return d->className;
}

/// Enables or disables the internal signal buffer
void TrackerChangeNotifier::setSignalsEnabled(bool enable)
{
    d->bufferActive=!enable;
    if(enable)
    {
        d->emitPendingSignals();
    }
}

///Sets the internal signal buffer size
void TrackerChangeNotifier::setBufferSize(int newSize)
{
    d->bufferSize=newSize;
}

void TrackerChangeNotifierPrivate::changed(QString,
                                    QList<TrackerChangeNotifier::Quad> deleted,
                                    QList<TrackerChangeNotifier::Quad> inserted)
{
    // D-Bus will filter based on the class name, so we only get relevant
    // signals here.
    if(!bufferActive)
    {
        TE_DEBUG << "TrackerChangeNotifer: emit changed - deletions" << deleted.size()
                    << "insertions" << inserted.size();
        emit q->changed(deleted, inserted);
    }
    else
    {
        deletedBuffer << deleted;
        insertedBuffer << inserted;
        TE_DEBUG << "TrackerChangeNotifer: buffered - deletions" << deleted.size()
                    << "insertions" << inserted.size();
        if(deletedBuffer.size() > bufferSize || insertedBuffer.size() > bufferSize)
        {
            TE_DEBUG << "TrackerChangeNotifer: emit changed (buffered) - deletions" << deletedBuffer.size()
                        << "insertions" << insertedBuffer.size();
            emit q->changed(deletedBuffer, insertedBuffer);
            deletedBuffer.clear();
            insertedBuffer.clear();
        }
    }
}

Q_DECLARE_METATYPE(TrackerChangeNotifier::Quad)
Q_DECLARE_METATYPE(QList<TrackerChangeNotifier::Quad>)
