/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

/*!

\mainpage QtSparqlTrackerLive

\brief A library building live models on top of QtSparql.

\warning Due to the short development time of TrackerLiveQuery, we don't promise
API / ABI stability! If we need API changes for supporting more use cases or
fixing bugs, we will implement the changes.

\section concepts Live concepts

To be able to use QtSparqlTrackerLive successfully, the user must know about:
- <a href="http://www.w3.org/TR/rdf-sparql-query/">SPARQL</a>, the query
  language. With TrackerLiveQuery, the user controls which SPARQL queries will
  be executed.
- <a href="http://library.gnome.org/devel/ontology/unstable/"> Ontologies</a>
  used by Tracker

QtSparqlTrackerLive works as follows:

- First, the \ref concepts_originalquery is executed and the data from it is
  inserted into the model.

- When data changes in Tracker, it sends the <a
  href="http://live.gnome.org/Tracker/Documentation/SignalsOnChanges">
  GraphUpdate signal</a> over D-Bus.

- The data in the signal is used for determining whether we need to run an \ref
  concepts_updatequery and what is the correct query to run.  The update query
  is constructed based on
   - the update query template given by the user
   - the data in the GraphChanged signal

- The update query is executed.

- Data from the update query is merged into the original model.

\subsection concepts_originalquery Original query

The original query determines the initial data that is inserted to the model.
It must also meet certain requirements that enable TrackerLiveQuery to merge
data from the update queries into the model.

In SPARQL, each resource is identified by it's URI.  Internally in the Tracker
data base, each resource has an integer id called tracker:id.  The GraphUpdated
signal contains these tracker:ids as data, and that's why the operation of the
live queries is based on the tracker:ids rather than URI's. URI's can also be
present in the original query, but they are not used by TrackerLiveQuery.

Example 1 (e-mails and titles): We have a model where column 0 contains
tracker:ids of e-mails and column 1 contains the message subject of the
corresponding e-mail.  Each e-mail can have only one subject.  E-mails without a
subject are not included in the model.  The model is ordered by the subjects of
the e-mails.

The original query is:
\code
SELECT tracker:id(?e) ?s
{
  ?e a nmo:Email ;
    nmo:messageSubject ?s .
}
ORDER BY ?s
\endcode

\subsection concepts_identitycolumns Identity columns

Identity columns define the identity for a row in the model.

Example 1 (e-mails and subjects): the column 0 containing the tracker:ids of
the e-mails is the only identity column in the model.  If the subject of an
e-mail changes, the e-mail is still the same (its identity has not changed).
But two e-mails with different tracker:ids are different, even if they have the
same title.

The model will never contain several rows with the same identity.

The set of identity columns is used to determine whether a row in the update
query result corresponds to an existing row (in which case the existing row
should be updated) or not (in which case the new row should be added).

\subsection concepts_collationcolumns Collation columns

The set of collation columns determine the ordering between the rows in the
model. The columns are used:
- when data in the model changes and a row needs to be relocated
- when a new row is added.

The user can define the list of collation columns.  For each column, the user
can define the type of the sorting (e.g., integer sorting, string sorting) and
whether the sorting is ascending or descending.

Example 1 (e-mails and subjects): The collation column is column 1 which
contains the subjects of the e-mails. The type of the sorting is ascending
string sorting.

\subsection concepts_whentoreact When to run updates

The user can specify that the TrackerLiveQuery should react to the changes in a
specified ontology class and a subset of its properties (or all properties).

Note: To listen to the changes in a specific ontology class, the class needs to
have the property \c tracker:notify set to \c true.

Example 1 (e-mails and subjects): The TrackerLiveQuery needs to monitor the changes in
the \c nmo:Email class.  The interesting properties are \c nmo:messageSubject
and \c rdf:type.  \c rdf:type is the property used for specifying that a
resource belongs to an ontology class.  To identify added and removed e-mails,
we need to monitor this property.  The changes in other properties can be
ignored, since they don't affect the content of the model.

\subsection concepts_updatequery Update query

When the data in Tracker changes, TrackerLiveQuery executes an update query.  Normally
the update query is the original query with additional filters.  The filters
limit the query to resources which have actually changed.

The update query needs to contain the same columns as the original query.  The
update query must follow the same sorting as the original query.

Example 1 (e-mails and subjects): Let's assume we have the following results for
the original query:

\verbatim
tracker:id(?e)  |  messageSubject
------------------------------------------
103             |  "hello"
100             |  "remember me?"
101             |  "special offer for you"
102             |  "welcome to my party"
\endverbatim

The GraphUpdated signal contains information in the form of triples (subject,
predicate, object). Let's assume the data is as follows:
- deleted triple: (100, nmo:messageSubject, ?)
- insterted triple: (100, nmo:messagSubject, ?) (these two changes occur
  together when the subject of the e-mail 100 has changed)
- deleted triple: (101, rdf:type, nmo:Email) (this means that the e-mail has
  been deleted)
- inserted triple: (104, rdf:type, nmo:Email)
- inserted triple: (104, rdf:messageSubject, ?) (these two changes occur when a
  new e-mail with a message subject has been inserted)

The GraphUpdated signal doesn't contain the literals (e.g., the messageSubject
of an e-mail). It only contains information that the messageSubject has been
deleted / inserted.

The update query to run is:
\code
SELECT tracker:id(?e) ?s
{
  ?e a nmo:Email ;
    nmo:messageSubject ?s .
  FILTER (tracker:id(?e) in (100, 101, 104))
}
order by ?s
\endcode
where \c (100, 101, 104) is the list of tracker:ids in the GraphUpdated signal
(in the "subject" part of the triples).

The update queries are not limited to be this simple.  With complex original
queries, more complex update queries are needed.

The update query template (given by the user) is an arbitrary string which
contains a placeholder for the "FILTER".  The placeholder is replaced with a
FILTER statement which is an "OR" of one or more filter snippets.  Each filter
snippet is an arbitrary string which contains a placeholder for the list of
tracker:ids.

The general format of the update query is "... %FILTER ..." (meaning: free text
with exactly one "%FILTER" placeholder inside it). %FILTER is expanded to
"FILTER ( snippet1 || snippet2 || ...)". Each snippet is of the form "... %LIST
... " (meaning: free text with exactly one "%LIST" placeholder inside it). %LIST
is replaced with a list of tracker:ids.

The final FILTER statement will be of the form "FILTER (tracker:id(?a) in
(1, 2, 3) || tracker:id(?b) in (4, 5, 6))".

See \ref updatequeries for examples of more complicated update queries.

To merge the update query results correctly into the model, the TrackerLiveQuery needs
information about \ref concepts_identitycolumns, \ref concepts_collationcolumns
and \ref concepts_affectedrows.

\subsection concepts_affectedrows Affected rows

To be able to merge the results back to the model (especially, to remove rows
which are no longer supposed to be included in the model), the TrackerLiveQuery needs
information about which existing rows in the model are affected by the update
query results.

The affected rows are expressed as rules of the following form: "column <number>
contains one of the id's in this list", where the list of id's is constructed
from the GraphChanged signal.  The list can contain data either from the
subjects of the triples (usual) or the objects of the triples (rare).

For specifying the affected rows, the user must give the:
- the column number
- whether the list of id's is constructed from the "subjects" or the "objects"
  of the GraphChanged signal.

Example 1 (e-mails and subjects): The affected rows are those which have one of
the changed subjects on the column 0. The signal contained the following
subjects: (100, 101, 104). The affected rows are marked in the model below:

\verbatim
tracker:id(?e)  |  messageSubject
------------------------------------------
103             |  "hello"
100             |  "remember me?"          << this row is affected
101             |  "special offer for you" << this row is affected
102             |  "welcome to my party"
\endverbatim

The removal of a row from the model happens usually when a resource is removed
from Tracker.  But it can also happen because a resource no longer meets the
criteria for the resources in the model.  (E.g,. "images with the favourite
tags", "e-mails whose subject begins with A".)

\subsection concepts_merging Merging back the results

When the update query returns, the results of it are merged into the original
model.  Merging means three different things:
- if an affected row is not present in the update query results, it is removed
  from the original model.
- other affected rows are updated with the data in the update query results.
  They are also relocated to the correct place, if the data affecting the
  sorting of the rows has changed.
- if a row is present in the update query results but is not present in the
  original model, it is added to the original model.

Example 1 (e-mails and subjects): The update query results are:

\verbatim
tracker:id(?e)  |  messageSubject
------------------------------------------
100             |  "changed subject"
104             |  "new e-mail"
\endverbatim

The data after merging these results is:

\verbatim
tracker:id(?e)  |  messageSubject
------------------------------------------
100             |  "changed subject"       << this row also got relocated
103             |  "hello"
104             |  "new e-mail"            << this new row in inserted in the correct place
102             |  "welcome to my party"
\endverbatim

\section updatequeries Correct update queries for different types of models

Goals:
- to keep the data of the model correct no matter how the data in Tracker
  changes

- to minimize the number of update queries needed

The previous section demonstrated the query type "one resource and its
properties".  The following sections explain the usage of TrackerLiveQuery for
more complicated models and its limitations.

\subsection one_resource_and_properties One resource and its properties

See the Example 1 of \ref concepts.

\subsection two_watched_classes Two watched classes

The following example requires listening to changes in 2 ontology classes, but
the update query for reacting to the changes is the same.

The original query retrives the tracker:id and nie:url of each photo and video.
Ordering is not specified, so the resulting model will be in arbitrary order,
and new rows will be added to arbitrary places when new data is added to the
model.

Original query:

\code
SELECT tracker:id(?u) nie:url(?u)
{
    {
        ?u a nmm:Photo .
    }
    UNION
    {
        ?u a nmm:Video .
    }
}
\endcode

Update query:
\code
SELECT tracker:id(?u) nie:url(?u)
{
    {
        ?u a nmm:Photo .
    }
    UNION
    {
        ?u a nmm:Video .
    }
    FILTER(tracker:id(?u) in (...))
}
\endcode

Identity columns: 0
Collation columns: none (sorting of the resources is arbitrary)

React to the following changes:
- class \c nmm:Photo, properties: \c rdf:type, \c nie:url
- class \c nmm:Video, properties: \c rdf:type, \c nie:url

List of id's is gathered from subjects in the GraphUpdated signal.

Affected rows are those whose value on column 0 is one of the id's.

\subsection two_resources Two related resources and their properties

Original query:

\code
SELECT tracker:id(?c) ?ct tracker:id(?m) ?s
{
    ?c a mfo:FeedChannel ;
        nie:title ?ct .
    ?m a mfo:FeedMessage ;
        nmo:messageSubject ?s ;
        nie:isLogicalPartOf ?c .
}
\endcode

Identity columns: 0, 2
Collation columns: none, sorting is arbitrary

React to the following changes:
- class \c mfo:FeedChannel, properties: \c rdf:type, \c nie:title
- class \c mfo:FeedMessage, properties: \c rdf:type, nmo:messageSubject

Update query:
\code
SELECT tracker:id(?c) ?ct tracker:id(?m) ?s
{
    ?c a mfo:FeedChannel ;
        nie:title ?ct .
    ?m a mfo:FeedMessage ;
        nmo:messageSubject ?s ;
        nie:isLogicalPartOf ?c .
    FILTER (tracker:id(?c) in list1 || tracker:id(?m) in list2)
}
\endcode

list1 is gathered from the subjects of the GraphUpdated signal related to \c
mfo:FeedChannel, and list2 is gathered from the subjects of the GraphUpdated
signal related to \c mfo:FeedMessage.

Affected rows:
Rows that have a value in list1 on column 0 \b or a value in list2 on column 2.

The trick is to minimize the number of queries if both resources "change at the
same time" (e.g., both a primary resource and a related secondary resource are
added at the same time).  That's why we have only one udpate query where we
filter both channels and messages.

When a new \c mfo:FeedMessage and a new \c mfo:FeedChannel are added at the same
time, the following inserted triples are received:
- (200, rdf:type, mfo:FeedChannel)
- (200, nie:title, ?)
- (201, rdf:type, mfo:FeedMessage)
- (201, messageSubject, ?)

The following update query is ran:
\code
SELECT tracker:id(?c) ?ct tracker:id(?m) ?s
{
    ?c a mfo:FeedChannel ;
        nie:title ?ct .
    ?m a mfo:FeedMessage ;
        nmo:messageSubject ?s ;
        nie:isLogicalPartOf ?c .
    FILTER (tracker:id(?c) in (200) || tracker:id(?m) in (201))
}
\endcode

Affected rows is an empty set, since the resources have just been added and are
not part of the previous model.

The update query returns only one row:
\verbatim
tracker:id(?c)  |   ?ct          |   tracker:id(?m)  |  ?s
--------------------------------------------------------------------
200             | "new channel"  |   201             | "new message"
\endverbatim

The row is then correctly added to the existing model.

\subsection aggregates Aggregates

The original query retrieves the tracker:ids of artists, and for each artist,
the number of songs, and the length of the shortest song of that artist.

Original query:

\code
SELECT tracker:id(?a) count(?mp) min(?l)
{
  ?a a nmm:Artist ;
     nmm:artistName ?an .
  OPTIONAL {
    ?mp a nmm:MusicPiece ;
        nmm:performer ?a ;
        nmm:length ?l .
  }
} GROUP BY ?a ORDER BY ?an
\endcode

The correct update query for this case is delightfully non-trivial:
\code
SELECT tracker:id(?a) ?an count(?mp) min(?l)
{
  ?a a nmm:Artist ;
     nmm:artistName ?an.
  OPTIONAL {
    ?mp a nmm:MusicPiece ;
        nmm:performer ?a ;
        nmm:length ?l .
  }
  FILTER (EXISTS { ?mp_filter a nmm:MusicPiece ;
                       nmm:performer ?a .
                   FILTER(?mp_filter in list1)
                 } << limits the query to the artists who have a specific song
          || ?a in list2) << limits the query to specific artists
} GROUP BY ?a ORDER BY ?an
\endcode

To keep the aggregate computation correct, the update query must not change the
solution set for the part that is shared with the original query.  This update
query fulfills the requirement: it doesn't modify the main pattern of the
original query:

\code
  ?a a nmm:Artist ;
     nmm:artistName ?an.
  OPTIONAL {
    ?mp a nmm:MusicPiece ;
        nmm:performer ?a ;
        nmm:length ?l .
  }
\endcode
but only adds a FILTER to it.  This means that the set of solutions (all the
possible combinations of resources to which ?a, ?an, ?mp and ?l can be bound)
remains the same.

\subsection aggregates_limitation1 Aggregates: limitation 1

\warning A limitation: removing resources from the model cannot be guaranteed to
succeed if the criteria for being included in the model is complex.

If we remove the OPTIONAL block from the previous example, we require that all
artists included in the model must have at least one song.

When a song is deleted, the following deleted triples are received:
- (300, rdf:type, nmm:MusicPiece)

Specifically, the triple
- (300, nmm:performer, 400)
is not included in the set of deleted triples.

When a song is deleted, we cannot any more determine whose song it was.
TrackerLiveQuery does not store data about the artist-song-relationship, and
since the song has been deleted from Tracker, we cannot query more information
about it.

As a result, we cannot determine the affected rows for the update we will run.
The update result will be empty, since the deleted song no longer exists.  If
some artist should be deleted from the model, the TrackerLiveQuery cannot know
that, and fails to delete the artist.

This problem will disappear if Tracker is changed to include all the properties
of the removed resources into the GraphChanged signal.

The problem can also be solved by adding a TrackerFullUpdater which re-runs the
original query when this change happens.

\subsection aggregates_limitation2 Aggregates: limitation 2

\warning A limitation: removing resources from the model cannot be guaranteed to
succeed if the criteria for being included in the model is complex.

The following example demonstrates this limitation.

Example: The model contains artists and their number of songs, but only artists
whose minimum length song is longer than 1 minute are included in the model.

The change: the length of a song changes.  Some artist might need to be deleted
from the model since he no longer has a song longer than 1 minute.

We can re-run the original query limited to the artists who have the changed
song. But we cannot determine the \ref concepts_affectedrows : The model stores
no data about the relationship between the artists and the songs.  Thus, we
don't know which artists to delete from the model, if they don't appear in the
update query results.

Unlike the problem described in \ref aggregates_limitation1, this problem will
not disappear even if Tracker includes more data in the signal.

The problem can also be solved by adding a TrackerFullUpdater which re-runs the
original query when this change happens.

\section api The API

The following classes are needed:
- TrackerLiveQuery
- TrackerPartialUpdater
- TrackerFullUpdater

\section example Example

The following example demonstrates the use of TrackerLiveQuery and
TrackerPartialUpdater.

\dontinclude genericlive/main.cpp
\skip TrackerLiveQuery
\until return
\line }

The following commands can be used for modifying data in Tracker while the
example is running.

\skip tracker-sparql
\until Resource

In the example, TrackerFullUpdater is not needed, since all changes can be
handled without re-querying the full data.

*/
