/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "merger.h"
#include "usermodifiablemodel.h"
#include "identity.h"
#include "debug.h"

#include <QAbstractItemModel>
#include <QList>
#include <QVariant>
#include <QSet>
#include <QMap>
#include <QDateTime>

/*
TODO:
- form a proper index of the new data (come up with a hash for QList<int>)
 */

namespace { // unnamed namespace for helper functions

int compareVariants(const QVariant& v1, const QVariant&v2, QVariant::Type type)
{
    // FIXME: this is just a draft!
    // TODO: support all data types
    if (v1 == v2)
        return 0;

    // Types from the tracker direct driver:
    // QUrl, QString, qlonglong, double, QDateTime, bool
    switch (type) {
    case QVariant::Url: // fall-through
    case QVariant::String:
        return QString::localeAwareCompare(v1.toString(), v2.toString());
        break;
    case QVariant::LongLong: // TODO: when Int, etc. use this larger data type
        return (v1.toLongLong() < v2.toLongLong() ? -1 : 1 );
        break;
    case QVariant::Double:
        return (v1.toDouble() < v2.toDouble() ? -1 : 1 );
        break;
    case QVariant::DateTime:
        return (v1.toDateTime() < v2.toDateTime() ? -1 : 1 );
        break;
    case QVariant::Bool:
        return (v1.toBool() < v2.toBool() ? -1 : 1 );
        break;
    case QVariant::Int:
        return (v1.toInt() < v2.toInt() ? -1 : 1 );
        break;
    default:
        break;
    }
    // Oops, type not recognized.
    LQ_WARNING << "LiveQuery: unrecognized data type" << type;
    return QString::localeAwareCompare(v1.toString(), v2.toString());
}

int compareRows(const QAbstractItemModel& m1, int row1,
                const QAbstractItemModel& m2, int row2,
                const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    Q_FOREACH(const TrackerLiveQuery::CollationColumn& col, collationColumns) {
        int abs = (col.order == Qt::AscendingOrder ? 1 : -1);
        QVariant v1 = m1.data(m1.index(row1, col.column));
        QVariant v2 = m2.data(m2.index(row2, col.column));
        int res = compareVariants(v1, v2, col.type);
        if (res != 0)
            return abs*res;
    }
    return 0;
}

/// Finds the correct place for row \a row from model \a source in the model \a
/// target, row range [start, end[.
int findLocationInRange(const QAbstractItemModel& source, int row,
                 const QAbstractItemModel& target, int start, int end,
                 const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    // TODO: replace this ugly and error-prone binary search with something
    // standard.  (Wrapping collationColumns somehow).
    int mid;
    while (start < end) {
        mid = (start + end) / 2; // integer division; "rounded" down
        int res = compareRows(source, row, target, mid, collationColumns);
        if (res == 0) {
            // Comparison-wise, the rows are the same. (They might still not
            // have the same identity.)
            return mid;
        }
        else if (res > 0)
            start = mid + 1; // skip mid (start is inclusive)
        else
            end = mid; // skip mid (end is not inclusive)
    }
    return start;
}

/// Finds the correct place for row \a row from model \a source in the model \a
/// target.
int findLocation(const QAbstractItemModel& source, int row,
                 const QAbstractItemModel& target,
                 const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    return findLocationInRange(source, row,
                               target, 0, target.rowCount(),
                               collationColumns);
}

void removeRows(UserModifiableModel& model,
                QLinkedList<IdentityList>& affectedRows,
                const QHash<IdentityList, int>& newDataIndex)
{
    // Go through the affected rows, remove those which are not in the new rows.
    // The rows need to be iterated in the same order they appear in the
    // affectedRows list.
    model.beginLazyRowRemovals();
    QLinkedList<IdentityList>::iterator it = affectedRows.begin();
    while (it != affectedRows.end()) {
        if (!newDataIndex.contains(*it))  {
            // New data has no row of that identity -> remove the original row
            // from the model.
            model.lazyRemoveRow(*it);
            it = affectedRows.erase(it);
        }
        else {
            ++it;
        }
    }
    model.endLazyRowRemovals();
}

void updateRows(UserModifiableModel& model,
                const QLinkedList<IdentityList>& affectedRows,
                const QAbstractItemModel& newData,
                QHash<IdentityList, int>& newDataIndex,
                const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    // Go through the affected rows (which weren't yet processed), update the
    // values with the corresponding row in newData, find the new place for the
    // row, relocate.

    QLinkedList<IdentityList>::const_iterator it;
    for (it = affectedRows.begin(); it != affectedRows.end(); ++it) {
        // *it is the identity of the row we're updating

        // The index of the new data row (in newData) containg the updated data
        const int ixInNewData = newDataIndex.take(*it);

        // Whether the data on that row has really changed.
        bool changed = false;

        // Update the values
        int ixInModel = model.identityToRow(*it).row();
        for (int c = 0; c < model.columnCount(); ++c) {
            QVariant newDataItem =
                newData.data(newData.index(ixInNewData, c));
            if (model.data(model.index(ixInModel, c)) != newDataItem) {
                model.setValue(model.index(ixInModel, c), newDataItem);
                changed = true;
            }
        }
        if (changed) {
            // The row has changed, so it needs possibly to be relocated. We
            // have already updated the data on the row, so the model is no
            // longer sorted. Binary search doesn't work if the data is not
            // sorted, so we cannot search for the new location just by doing a
            // binary search over the model.

            // It is also important that we don't relocate the row if it is not
            // strictly needed. E.g., the previous and the next row might have
            // the identical sorting data as the target row. In this case,
            // there's no guarantee that the any search for the correct location
            // would return the current location of the row (it can return any
            // location that is acceptable).

            if (ixInModel > 0 && compareRows(model, ixInModel - 1,
                                       model, ixInModel,
                                       collationColumns) > 0) {
                // The row should be moved backwards; look for a new location in
                // range [0, ixInModel[.
                int target = findLocationInRange(model, ixInModel,
                                                 model, 0, ixInModel,
                                                 collationColumns);
                model.moveRow(ixInModel, target);

            }
            else if (ixInModel < model.rowCount() - 1 &&
                     compareRows(model, ixInModel,
                                 model, ixInModel + 1,
                                 collationColumns) > 0) {
                // The row should be moved forward; look for a new location in
                // range [ixInModel + 1, rowCount()[.
                int target = findLocationInRange(model, ixInModel,
                                                 model, ixInModel + 1, model.rowCount(),
                                                 collationColumns);
                model.moveRow(ixInModel, target);
            }
            // Else: the current place for the row is still ok. E.g., the data
            // affecting the sorting didn't change, or changed so that the
            // current place is still fine.
        }
    }
}

void addRows(UserModifiableModel& model,
             const QAbstractItemModel& newData,
             QHash<IdentityList, int>& newDataIndex,
             const QList<int>& identityColumns,
             const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    // First make sure that none of the rows we're trying to insert exist in the
    // model.  If we cannot add a row since it exists yet, we gather the info
    // here, to be able to run updateRows for that row.
    QLinkedList<IdentityList> newAffectedRows;

    QMap<int, int> reallyNewRows; // we use this as a sorted set
    QHash<IdentityList, int>::const_iterator i;
    for (i = newDataIndex.constBegin(); i != newDataIndex.constEnd(); ++i) {
        // i.key() = identity, i.value() = the row index in newData
        IdentityList identity = i.key();

        // Safeguard: never add 2 rows with the same identity, no matter what.
        // (This guards against cases where we weren't able to resolve
        // affectedRows properly.)  Instead, we need to update values on that
        // row.
        if (model.hasRowWithIdentity(identity)) {
            newAffectedRows.append(identity);
        }
        else {
            // the row i.value() of newData is really a new row to be added to
            // the model later
            reallyNewRows.insert(i.value(), 1);
        }
    }
    if (newAffectedRows.isEmpty() == false) {
        updateRows(model,
                   newAffectedRows,
                   newData,
                   newDataIndex,
                   collationColumns);
    }

    // Go through the new rows (which weren't yet processed), add them to the
    // model (finding the right place).  To group the rowsInserted signals, the
    // rows must be processed in the same order they will appear in the merged
    // model.  So, we process the rows in newData, but only those which are
    // present in reallyNewRows.

    // Pre-allocate row value list to avoid re-allocating it for each result row
    QVariantList valueList;
    valueList.reserve(model.columnCount());
    for (int c = 0; c < model.columnCount(); ++c)
        valueList.push_back(QVariant());

    model.beginLazyRowInsertions();
    Q_FOREACH(int newRow, reallyNewRows.keys()) {
        int rowLocation = findLocation(newData, newRow, model, collationColumns);
        for (int c = 0; c < model.columnCount(); ++c)
            valueList[c] = newData.data(newData.index(newRow, c));
        model.lazyInsertRow(rowLocation, valueList);
    }
    model.endLazyRowInsertions();
}

} // end of unnamed namespace

void mergeResultsToModel(UserModifiableModel& model,
                         QLinkedList<IdentityList>& affectedRows, // we will modify this
                         const QAbstractItemModel& newData,
                         const QList<int>& identityColumns,
                         const QList<TrackerLiveQuery::CollationColumn>& collationColumns)
{
    // Merge the results from newData into the existing model.
    // Process the updates in the following order:

    // 1. removals: affected rows in the model which are not present in the new
    // data.
    // 2. updates, possibly involving row relocations: affected rows whose
    // values have changed in new data.
    // 3. additions: rows which are present in new data but not in the old model

    // This order minimizes the signals sent:

    // If a row has its values changed and needs to be relocated, maybe all the
    // rows between the old location and new location got removed, so no
    // relocation is needed. -> Process removals before updates.

    // If a row is added, maybe another row needs to be moved over it. If
    // updates are processed first, the new row can be added to the right place
    // right away. -> Process updates before additions.


    // --- Algorithm
    // Form an index of the new rows: id -> row number.

    QHash<IdentityList, int> newDataIndex;
    for (int i = 0; i < newData.rowCount(); ++i) {
        newDataIndex.insert(IdentityList::fromRow(newData, i, identityColumns), i);
    }
    // TODO: consider combining newDataIndex and newData into a class

    // Note: this removes rows from affectedRows, too
    removeRows(model, affectedRows, newDataIndex);

    // Note: this removes stuff from newDataIndex
    updateRows(model, affectedRows, newData, newDataIndex, collationColumns);

    addRows(model, newData, newDataIndex, identityColumns, collationColumns);
}
