/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "movablerowmodel.h"
#include "debug.h"

MovableRowModel::MovableRowModel(QObject* parent)
: QAbstractTableModel(parent), batchInProgress(false)
{
    connectDataChanged();
}

MovableRowModel::~MovableRowModel()
{
}

inline QModelIndex MovableRowModel::toModelIndex(const QModelIndex& ownIndex) const
{
    return (ownIndex.isValid() ? model.index(ownIndex.row(), ownIndex.column(), ownIndex.parent()) : QModelIndex());
}

inline QModelIndex MovableRowModel::toOwnIndex(int modelRow, int modelColumn,
                                               const QModelIndex& parent) const
{
    return toOwnIndex(model.index(modelRow, modelColumn, parent));
}

inline QModelIndex MovableRowModel::toOwnIndex(const QModelIndex& modelIndex) const
{
    return (modelIndex.isValid() ? createIndex(modelIndex.row(), modelIndex.column()) : QModelIndex());
}

void MovableRowModel::connectDataChanged()
{
    LQ_DEBUG << "MovableRowModel: connectDataChanged()";
    connect(&model, SIGNAL(dataChanged(const QModelIndex&,const QModelIndex&)),
            this, SLOT(convertDataChanged(const QModelIndex&,const QModelIndex&)));
}

void MovableRowModel::disconnectDataChanged()
{
    LQ_DEBUG << "MobableRowModel: disconnectDataChanged()";
    disconnect(&model, SIGNAL(dataChanged(const QModelIndex&,const QModelIndex&)),
               this, SLOT(convertDataChanged(const QModelIndex&,const QModelIndex&)));
}

int MovableRowModel::rowCount(const QModelIndex &parent) const
{
    return model.rowCount(parent);
}

int MovableRowModel::columnCount(const QModelIndex &parent) const
{
    return model.columnCount(parent);
}

QVariant MovableRowModel::data(const QModelIndex &index, int role) const
{
    // If we're asking for a user defined role, it's coming from
    // QML, lets deal with it and return the correct data
    if (role >= Qt::UserRole+1) {
        // We need to offset the column, first header is column 0, second 0+1 etc..
        // Since we set the roles as offsets of Qt::UserRole, we can
        // use this to get the correct column;
        int columnOffset = role - (Qt::UserRole+1);
        // create the new index
        QModelIndex correctIndex =  model.index(index.row(),
                                                index.column()+columnOffset,
                                                index.parent());
        // Ignore the role now, we don't add to the model with any role information
        // so we need to use 0
        return model.data(correctIndex, 0);
    }
    return model.data(toModelIndex(index), role);
}

QModelIndex MovableRowModel::index(int row, int column,
                                   const QModelIndex& parent) const
{
    return toOwnIndex(row, column, parent);
}

QVariant MovableRowModel::headerData(int section, Qt::Orientation orientation,
                                     int role) const
{
    return model.headerData(section, orientation, role);
}

bool MovableRowModel::setHeaderData(int section, Qt::Orientation orientation,
                                    const QVariant & value, int role)
{
    // Lets use header data as a role name for QML
    // Offset the role name by the column id 
    // Need this later to get the correct column for the role

    if(orientation == Qt::Horizontal) {
        QHash<int, QByteArray> names = roleNames();
        names[(Qt::UserRole+1)+section] = value.toString().toAscii();
        setRoleNames(names);
    }

    return model.setHeaderData(section, orientation, value, role);
}

bool MovableRowModel::addColumns(int column, int count)
{
    if ((count < 1) || (column < 0) || (column > columnCount()))
         return false;

    beginInsertColumns(QModelIndex(), column, column + count - 1);
    const bool result = model.insertColumns( column, count );
    endInsertColumns();
    return result;
}

bool MovableRowModel::addRows(int row, int count)
{
    if ((count < 1) || (row < 0) || (row > rowCount()))
         return false;

    if (!batchInProgress) {
        beginInsertRows(QModelIndex(), row, row + count - 1);
    }

    const bool result = model.insertRows( row, count );

    if (!batchInProgress) {
        endInsertRows();
    }

    return result;
}

bool MovableRowModel::addRowWithData(int row,
                                     const QAbstractItemModel& sourceModel,
                                     int sourceRow)
{
    if (row < 0 || row > rowCount())
        return false;
    if (sourceRow < 0 || sourceRow >= sourceModel.rowCount())
        return false;

    if (!batchInProgress) {
        // we only want to emit rowsInserted, no dataChanged.
        disconnectDataChanged();
        beginInsertRows(QModelIndex(), row, row);
    }

    bool result = true;
    if ((result = model.insertRows(row, 1))) {
        for (int c = 0; c < model.columnCount(); c++) {
            // We have disconnected dataChanged so this won't result in emitting
            // it
            model.setData(model.index(row, c),
                          sourceModel.data(sourceModel.index(sourceRow, c)));
        }
    }

    if (!batchInProgress) {
        endInsertRows();
        connectDataChanged();
    }

    return result;
}

bool MovableRowModel::addRowWithData(int row,
                                     const QVariantList& data)
{
    if (row < 0 || row > rowCount())
        return false;

    if (!batchInProgress) {
        // we only want to emit rowsInserted, no dataChanged.
        disconnectDataChanged();
        beginInsertRows(QModelIndex(), row, row);
    }

    bool result = true;
    if ((result = model.insertRows(row, 1))) {
        for (int c = 0; c < model.columnCount() && c < data.size(); c++) {
            // We have disconnected dataChanged so this won't result in emitting
            // it
            model.setData(model.index(row, c), data[c]);
        }
    }

    if (!batchInProgress) {
        endInsertRows();
        connectDataChanged();
    }

    return result;
}

bool MovableRowModel::deleteRows(int row, int count)
{
    if ((count < 1) || (row < 0) || ((row + count) > rowCount()))
        return false;

    beginRemoveRows(QModelIndex(), row, row + count - 1);
    const bool result = model.removeRows( row, count );
    endRemoveRows();
    return result;
}

bool MovableRowModel::moveRow(int row, int where)
{
    if (row < 0 || where < 0 || row >= model.rowCount() || where > model.rowCount())
        return false;

    if (!beginMoveRows(QModelIndex(), row, row, QModelIndex(), where))
        return false;

    model.insertRow(where);

    if (where < row)
        ++row;

    // Copy the columns
    disconnectDataChanged();
    for (int c = 0; c < columnCount(); ++c) {
        model.setData(model.index(where, c), model.data(model.index(row, c)));
    }
    connectDataChanged();

    model.removeRow(row);

    endMoveRows();
    return true;
}

bool MovableRowModel::setValue(const QModelIndex& index, const QVariant& value)
{
    return model.setData(toModelIndex(index), value);
}

void MovableRowModel::convertDataChanged(const QModelIndex& topLeft,
                                         const QModelIndex& bottomRight)
{
    LQ_DEBUG << "MovableRowModel: emitting dataChanged";
    emit dataChanged(toOwnIndex(topLeft), toOwnIndex(bottomRight));
}

void MovableRowModel::beginAddRowsBatch(int row, int count)
{
    Q_ASSERT(count > 0);
    Q_ASSERT(row >= 0);
    Q_ASSERT(row <= rowCount());

    if ((count < 1) || (row < 0) || (row > rowCount())) {
        return;
    }

    batchInProgress = true;
    disconnectDataChanged();
    beginInsertRows(QModelIndex(), row, row + count - 1);
}

void MovableRowModel::endAddRowsBatch()
{
    batchInProgress = false;
    connectDataChanged();
    endInsertRows();
}
