/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "ontology.h"
#include "debug.h"

#include <QSparqlResult>
#include <QSparqlConnection>
#include <QSparqlError>

OntologyQuerier::OntologyQuerier(QSparqlConnection& conn, QObject* parent)
    : QObject(parent), conn(conn), classesAdded(false), ontologyResult(0)
{
}

OntologyQuerier::~OntologyQuerier()
{
    // ontologyResult is a child
}

// Tell the OntologyQuerier that information about this predicate is needed.
void OntologyQuerier::addPredicate(const QString& shortPredicate)
{
    // Ontology query results contains 2 kinds of triplets:
    // 1, short name, long name of a class
    // 0, short name, tracker:id of the property

    // Limitation: we assume we don't get too many variables in the query.
    ontologySnippet.append(
        QString("1 \"%1\" tracker:id(%1) ").arg(shortPredicate));
}

// Tell the OntologyQuerier that information about this class is needed.
void OntologyQuerier::addClass(const QString& shortClass)
{
    // Ontology query results contains 2 kinds of triplets:
    // 1, short name, long name of a class
    // 0, short name, tracker:id of the property

    // Limitation: we assume we don't get too many variables in the query.
    ontologySnippet.append(QString("0 \"%1\" %1 ").arg(shortClass));

    classesAdded = true;
}

// Strts the ontology query.
void OntologyQuerier::start()
{
    QString ontologyQuery = QString("select %1 { }").arg(ontologySnippet);
    ontologySnippet.clear(); // not needed any more
    LQ_DEBUG << "Executing ontology query: " << ontologyQuery;

    ontologyResult = conn.exec(QSparqlQuery(ontologyQuery));

    ontologyResult->setParent(this);

    if (ontologyResult->hasError()) {
        LQ_WARNING << "error in ontology query"
                   << ontologyResult->lastError().message();
        delete ontologyResult;
        ontologyResult = 0;
        emit finished();
        return;
    }

    connect(ontologyResult, SIGNAL(finished()),
            SLOT(ontologyQueryFinished()));
}

// When the query has finished, this function returns the short name of a class.
QString OntologyQuerier::longClassToShort(const QString& klass) const
{
    // returns the default-constructed value if not present
    return longToShort[klass];
}

// When the query has finished, this function returns the map containing
// tracker:ids of predicates.
const QHash<QString, int>& OntologyQuerier::shortPredicateToIdMap() const
{
    return shortToId;
}

QList<QString> OntologyQuerier::longClasses() const
{
    return longToShort.keys();
}

void OntologyQuerier::ontologyQueryFinished()
{
    LQ_DEBUG << "Ontology query finished";
    if (ontologyResult->hasError()) {
        LQ_WARNING << "error in ontology query"
                   << ontologyResult->lastError().message();
    }
    else if (ontologyResult->next()) {
        // The structure of the ontology query:
        // - only one row
        // - values on that row form triplets:
        // (0, short name for a class, long name for a class) and
        // (1, short name for a predicate, tracker:id of a predicate)
        for (int i=0; i<ontologyResult->current().count()-2; i += 3) {
            switch (ontologyResult->value(i).toInt()) {
            case 0:
            {
                // this triplet describes a class: 0 shortname longname
                QString classShort = ontologyResult->value(i+1).toString();
                QString classLong = ontologyResult->value(i+2).toString();

                if (classLong.isEmpty()) {
                    LQ_WARNING << "Unrecognized class:" << classShort;
                }
                else {
                    longToShort[classLong] =
                        ontologyResult->value(i+1).toString();
                }
                break;
            }
            case 1:
            {
                // this triplet describes a property: 1 shortname tracker:id
                int id = ontologyResult->value(i+2).toInt();
                QString predicateShort = ontologyResult->value(i+1).toString();

                if (id == 0) {
                    LQ_WARNING << "Unrecognized predicate" << predicateShort;
                }
                else {
                    shortToId[predicateShort] = id;
                }
                break;
            }
            default:
                break;
            }
        }
    }

    ontologyResult->deleteLater();
    ontologyResult = 0;

    emit finished();
}
