/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "trackerlivequery.h"

#include "merger.h"
#include "updater.h"
#include "updaterule.h"
#include "usermodifiablemodel.h"
#include "ontology.h"
#include "liveperformancerecorder.h"
#include "debug.h"

#include <TrackerChangeNotifier>

#include <QSparqlResult>
#include <QSparqlQueryModel>
#include <QSparqlError>

#include <QObject>
#include <QString>
#include <QList>
#include <QSet>
#include <QStringList>
#include <QDateTime>
#include <QLinkedList>

/*
Startup of the TrackerLiveQuery:
1. perform the ontology query, async
2. ontology query returns
3. connect to the GraphUpdated signal
4. perform the initial query, async
5. initial query returns -> put the data into the model
...
6. GraphUpdated comes
7. perform the update query, async
8. update query returns
9. merge results back to the model
*/

/*!

  \class TrackerLiveQuery

  \brief Class which implements live queries with Tracker.

  TrackerLiveQuery is an interface to a QAbstractItemModel which contains the
  result of a query and which is kept up to date when the result changes.

  TrackerLiveQuery first runs the initial query, given by the user.  Then it
  connects to the GraphChanged signal which Tracker sends when data in its
  database has changed.  When the signal arrives, TrackerLiveQuery runs an
  update query to find out how the changes affect the current data, and merges
  the changes back into the data.

  The user specifies how the TrackerLiveQuery should reach to changes by
  associating updaters with it.  Each updater knows which changes it should
  react to (ontology class name and property name) and what kind of an update
  query to run.

  Two kinds of updaters can be used for keeping the TrackerLiveQuery up to date:
  - TrackerPartialUpdater runs an update query which is limited to the changed
    resources
  - TrackerFullUpdater re-runs the original query

*/

namespace
{

    template<class T>
    void resize(QList<T>& list, int size)
    {
        const int diff = size - list.size();
        list.reserve(size);
        for (int c = 0; c < diff; ++c)
            list.push_back(T());
    }

}

class TrackerLiveQueryPrivate : public QObject
{
    Q_OBJECT
public:
    TrackerLiveQueryPrivate(const QSparqlQuery& origQuery,
                            int columnCount,
                            QSparqlConnection& conn,
                            QObject* parent);
    ~TrackerLiveQueryPrivate();

    Q_INVOKABLE void runUpdates();

    void setIdentityColumns(const QList<int>& identityColumns);

public:
    QSparqlQuery origQuery;

    int columnCount;
    QList<int> identityColumns;
    QList<TrackerLiveQuery::CollationColumn> collationColumns;

    UserModifiableModel model; // data storage

    // Stores the TrackerPartialUpdaters which will update this live model.  The
    // pointers are "strong", i.e., we have increased the ref count for each
    // TrackerPartialUpdaterPrivate the pointers point to, and we must also
    // decrease it when we delete this list.
    QList<TrackerPartialUpdaterPrivate*> updaters;

    // Stores the TrackerFullUpdaters which will re-run the query to update this
    // model.
    QList<TrackerFullUpdaterPrivate*> rerunners;

    OntologyQuerier ontologyQuerier;

    // True if we currently react on the GraphChanged signal by running an
    // update query.  Updates are disabled 1) when the original query hasn't
    // finished yet 2) when the user has explicitly paused the updates.  During
    // that time, we only gather information about the update queries to be run
    // when the updates are enabled the next time.
    bool updatesEnabled;
    // This boolean stores the user's intention about whether the updates should
    // be enabled or not.  (E.g., our intention might be that updatesEnabled ==
    // false since the original query hasn't finished yet, but the user's
    // intention might change even during that time.)
    bool updatesEnabledUser;

    LivePerformanceRecorder* performanceRecorder;

private:
    QSparqlConnection& conn;
    // Original query result
    QSparqlResult *origResult;
    // Which TrackerPartialUpdaters will run their queries next time we enter
    // the main loop.  The list contains "weak pointers", i.e., we don't
    // increase the ref count of a TrackerPartialUpdaterPrivate because we put
    // it in this list.
    QSet<TrackerPartialUpdaterPrivate*> activeUpdaters;
    // If some TrackerFullUpdater wants to re-run the query now
    bool requeryNow;

    // For each running update query, how to identify "the set of "affected
    // rows": the affected rows will have in their column (rule.idColumn_) one
    // of the id's (rule.idValues_).
    QHash<QSparqlQueryModel*, QList<UpdateRule> > affectedRowRules;

Q_SIGNALS:
    void initialQueryFinished();

private Q_SLOTS:
    void ontologyQueryFinished();
    void origQueryDataReady(int count);
    void origQueryFinished();
    void updateQueryFinished();
    void requeryFinished();

    void onChanged(QList<TrackerChangeNotifier::Quad>,
                   QList<TrackerChangeNotifier::Quad>);

private:
    void connectToGraphUpdated(const QString& classLong);
    void startOriginalQuery();
};

TrackerLiveQueryPrivate::TrackerLiveQueryPrivate(const QSparqlQuery& origQuery,
                                                 int columnCount,
                                                 QSparqlConnection& conn,
                                                 QObject* parent)
    : QObject(parent), origQuery(origQuery),
      columnCount(columnCount),
      ontologyQuerier(conn),
      updatesEnabled(false), updatesEnabledUser(true),
      performanceRecorder(LivePerformanceRecorder::create(parent)),
      conn(conn),
      origResult(0),
      requeryNow(false)
{
    model.addColumns(0, columnCount);

    connect(&ontologyQuerier, SIGNAL(finished()),
            SLOT(ontologyQueryFinished()));

    if (performanceRecorder) performanceRecorder->reset();
}

TrackerLiveQueryPrivate::~TrackerLiveQueryPrivate()
{
    qDeleteAll(updaters);
    qDeleteAll(rerunners);
}

void TrackerLiveQueryPrivate::connectToGraphUpdated(const QString& classLong)
{
    TrackerChangeNotifier* n = new TrackerChangeNotifier(classLong, this);
    connect(n,
            SIGNAL(changed(QList<TrackerChangeNotifier::Quad>,
                           QList<TrackerChangeNotifier::Quad>)),
            SLOT(onChanged(QList<TrackerChangeNotifier::Quad>,
                           QList<TrackerChangeNotifier::Quad>)));
}

void TrackerLiveQueryPrivate::runUpdates()
{
    // Did the user disable the updates while runUpdates was in the queue?
    if (!updatesEnabledUser)
        return;

    // If there are some active rerunners, re-run the query and ignore
    // everything else.
    if (requeryNow) {
        LQ_DEBUG << "Rerunning the query";

        QSparqlQueryModel* newData = new QSparqlQueryModel(this);
        newData->setQuery(QSparqlQuery(origQuery), conn);

        // TODO: we cannot currently "stream in" the results of the rerun;
        // implement that
        connect(newData, SIGNAL(finished()),
                SLOT(requeryFinished()));

        activeUpdaters.clear();
        requeryNow = false;

        if (performanceRecorder) performanceRecorder->incrementReQueryCounter();
        return;
    }

    // Initiate one update query for each active TrackerPartialUpdater.  Note:
    // some updaters can have lots of id's in their lists, so this feature
    // relies on Tracker being able to handle long FILTER(?sth in (...)) lists.
    Q_FOREACH(TrackerPartialUpdaterPrivate* up, activeUpdaters) {
        QList<UpdateRule> rules;
        QString realUpdateQuery = up->buildQuery(rules);

        // This query model will contain the results of the query.  It is
        // deleted with deleteLater() when the query has finished.
        QSparqlQueryModel* newData = new QSparqlQueryModel(this);

        // Note down the rules for computing affected rows when the query
        // returns.  Here we need the newData to exist already.
        affectedRowRules.insert(newData, rules);

        LQ_DEBUG << "Running update query" << realUpdateQuery;
        if (performanceRecorder) performanceRecorder->incrementUpdateQueryCounter();

        newData->setQuery(QSparqlQuery(realUpdateQuery), conn);
        connect(newData, SIGNAL(finished()),
                SLOT(updateQueryFinished()));
    }
    activeUpdaters.clear();
}

void TrackerLiveQueryPrivate::setIdentityColumns(const QList<int>& identityColumns)
{
    Q_FOREACH(int col, identityColumns) {
        if (col < 0 || col >= columnCount) {
            LQ_WARNING
                    << "identity column index" << col
                    << "is out of range for live query:" << origQuery.query();
            return;
        }
    }

    this->identityColumns = identityColumns;
    model.setIdentityColumns(identityColumns);
}

void TrackerLiveQueryPrivate::ontologyQueryFinished()
{
    if (performanceRecorder) performanceRecorder->markOntologyQueryFinished();

    QList<QString> classes = ontologyQuerier.longClasses();

    // Connect to the GraphUpdated signal now that we know the long names of the
    // classes.
    Q_FOREACH(const QString& classLong, classes)
        connectToGraphUpdated(classLong);

    // Give the predicate information to all updaters
    Q_FOREACH(TrackerPartialUpdaterPrivate* up, updaters)
        up->convertPredicates(ontologyQuerier.shortPredicateToIdMap());

    Q_FOREACH(TrackerFullUpdaterPrivate* up, rerunners)
        up->convertPredicates(ontologyQuerier.shortPredicateToIdMap());

    startOriginalQuery();
}

void TrackerLiveQueryPrivate::startOriginalQuery()
{
    if (performanceRecorder) performanceRecorder->markOriginalQueryStarted();

    origResult = conn.exec(origQuery);
    if (origResult->hasError()) {
                LQ_WARNING << "error in query"
                   << origResult->lastError().message();
        delete origResult;
        origResult = 0;
        emit initialQueryFinished();
    }
    else {
        origResult->setParent(this);
        connect(origResult, SIGNAL(dataReady(int)),
                SLOT(origQueryDataReady(int)));
        connect(origResult, SIGNAL(finished()),
                SLOT(origQueryFinished()));
    }
}

void TrackerLiveQueryPrivate::origQueryDataReady(int count)
{
    LQ_DEBUG << "Original query data ready" << count;

    if (performanceRecorder) performanceRecorder->markOriginalQueryDataBlockReady();

    // We have previously retrieved model.rowCount, now insert more rows, if
    // available
    if (count > model.rowCount()) {
        if (performanceRecorder) performanceRecorder->markOriginalDataBlockToModelWriteStart();

        // Pre-alocate row value list to avoid re-allocating it for each result row
        QVariantList valueList;
        resize(valueList, columnCount);

        model.beginAddRowsBatch(model.rowCount(), count - model.rowCount());
        for (int row = model.rowCount(); row < count; ++row) {
            // Batch operation allows the model to emit only one rowsInserted
            if (origResult->setPos(row)) {
                for (int c = 0; c < columnCount; ++c)
                    valueList[c] = origResult->value(c);
            }
            model.addRowWithData(row, valueList);
        }
        model.endAddRowsBatch();

        if (performanceRecorder) performanceRecorder->markOriginalDataBlockToModelWriteEnd();
    }
}

void TrackerLiveQueryPrivate::origQueryFinished()
{
    LQ_DEBUG << "Original query finished";
    origResult->deleteLater();
    origResult = 0;
    emit initialQueryFinished();

    if (performanceRecorder) performanceRecorder->markInitialQueryEmitted();

    updatesEnabled = true;
    // run the pending updates, if any
    if (updatesEnabled && updatesEnabledUser)
        runUpdates();
}

void TrackerLiveQueryPrivate::updateQueryFinished()
{
    LQ_DEBUG << "Update query finished";
    QSparqlQueryModel* newData = qobject_cast<QSparqlQueryModel*>(QObject::sender());

    LQ_DEBUG << "newData.rowCount()" << newData->rowCount();
    LQ_DEBUG << "model.rowCount()" << model.rowCount();
    // FIXME: do this more efficiently. Perhaps with some kind of a index.

    // TODO: move affected rows computation to UserModifiableModel.

    // See which rows are affected.  There can be multiple rules for determining
    // the affected rows.  Rules are of the form (column, list of id's).  A row
    // is affected if it's value on the column is one of the id's in the list.

    QLinkedList<IdentityList> affectedRows =
        model.affectedRows(affectedRowRules[newData]);

    mergeResultsToModel(model, affectedRows, *newData,
                        identityColumns, collationColumns);
    affectedRowRules.remove(newData);
    LQ_DEBUG << "model.rowCount() now" << model.rowCount();
    newData->deleteLater();
}

void TrackerLiveQueryPrivate::requeryFinished()
{
    LQ_DEBUG << "Requery finished";

    // There can be multiple requeries in progress at the same time; handle them
    // gracefully.
    QSparqlQueryModel* newData = qobject_cast<QSparqlQueryModel*>(QObject::sender());

    LQ_DEBUG << "newData.rowCount()" << newData->rowCount();
    LQ_DEBUG << "model.rowCount()" << model.rowCount();

    // FIXME: have a better way to say "all rows are affected" than just putting
    // all indices to the list.

    QLinkedList<IdentityList> affectedRows;
    for (int r = 0; r < model.rowCount(); ++r) { // FIXME: slow!
        const QModelIndex index(model.index(r, 0));
        affectedRows.append(IdentityList::fromRow(model, r, identityColumns));
    }

    mergeResultsToModel(model, affectedRows, *newData,
                        identityColumns, collationColumns);
    LQ_DEBUG << "model.rowCount() now" << model.rowCount();
    newData->deleteLater();
}

void TrackerLiveQueryPrivate::onChanged(QList<TrackerChangeNotifier::Quad> deletes,
                                 QList<TrackerChangeNotifier::Quad> inserts)
{
    TrackerChangeNotifier* notifier = qobject_cast<TrackerChangeNotifier*>(
        QObject::sender());
    if (!notifier) {
        LQ_WARNING << "Unrecognized signal sender";
        return;
    }

    QString classLong = notifier->watchedClass();
    QString classShort = ontologyQuerier.longClassToShort(classLong);
    if (classShort.isEmpty()) {
        LQ_DEBUG << "unrecognized class" << classLong;
        return;
    }

    LQ_DEBUG << "changed" << classShort << deletes << inserts;

    // If activeUpdaters contains something, or requeryNow is true, the real
    // update has already been scheduled to be run when we next time enter the
    // main loop.
    bool doSchedule = (updatesEnabled && updatesEnabledUser &&
                       activeUpdaters.isEmpty() && !requeryNow);

    // Check if there are any TrackerFullUpdaters that want to re-run the whole
    // query
    Q_FOREACH(TrackerFullUpdaterPrivate* up, rerunners) {
        if(up->checkSignal(classShort, deletes, inserts)) {
            requeryNow = true;
            if (doSchedule) {
                QMetaObject::invokeMethod(this, "runUpdates", Qt::QueuedConnection);
            }
            return;
        }
    }

    // Check which TrackerPartialUpdaters target this class.  Since the number
    // of updaters is supposed to be really small (e.g., 2), we check all
    // updaters without having an index (class name -> interested updaters).
    Q_FOREACH(TrackerPartialUpdaterPrivate* up, updaters) {
        // Gather data from the signal to the Updater and schedule the update
        // query to happen when we enter the main loop the next time (and have
        // processed all GraphUpdated signals).  Note that the update query may
        // need to gather data from multiple GraphUpdated signals, to be able to
        // execute only one update query even if multiple classes have changed.
        bool active = up->storeIds(classShort, deletes, inserts);
        if (active)
            activeUpdaters.insert(up);
    }
    if (doSchedule && !activeUpdaters.isEmpty()) {
        // Schedule this function to be run when we enter the main loop.  We
        // hope that the events from all D-Bus signals (of different classes)
        // are already in the event queue.
        QMetaObject::invokeMethod(this, "runUpdates", Qt::QueuedConnection);
    }
}

/////////////////////////////////////////////////////////////////////////

TrackerLiveQuery::TrackerLiveQuery(const QString& origQuery,
                                   int columnCount,
                                   const QList<int>& identityColumns,
                                   QSparqlConnection& conn)
    : d(new TrackerLiveQueryPrivate(QSparqlQuery(origQuery), columnCount, conn, this))
{
    connect(d, SIGNAL(initialQueryFinished()), SIGNAL(initialQueryFinished()));
    d->setIdentityColumns(identityColumns);
}

TrackerLiveQuery::TrackerLiveQuery(const QSparqlQuery& origQuery,
                                   int columnCount,
                                   const QList<int>& identityColumns,
                                   QSparqlConnection& conn)
    : d(new TrackerLiveQueryPrivate(origQuery, columnCount, conn, this))
{
    connect(d, SIGNAL(initialQueryFinished()), SIGNAL(initialQueryFinished()));
    d->setIdentityColumns(identityColumns);
}

// Deprecated
TrackerLiveQuery::TrackerLiveQuery(const QString& origQuery,
                                   int columnCount,
                                   QSparqlConnection& conn)
    : d(new TrackerLiveQueryPrivate(QSparqlQuery(origQuery), columnCount, conn, this))
{
    connect(d, SIGNAL(initialQueryFinished()), SIGNAL(initialQueryFinished()));
}

// Deprecated
TrackerLiveQuery::TrackerLiveQuery(const QSparqlQuery& origQuery,
                                   int columnCount,
                                   QSparqlConnection& conn)
    : d(new TrackerLiveQueryPrivate(origQuery, columnCount, conn, this))
{
    connect(d, SIGNAL(initialQueryFinished()), SIGNAL(initialQueryFinished()));
}

TrackerLiveQuery::~TrackerLiveQuery()
{
}

// Deprecated
void TrackerLiveQuery::setIdentityColumns(const QList<int>& identityColumns)
{
    d->setIdentityColumns(identityColumns);
}

void TrackerLiveQuery::setCollationColumns(const QList<CollationColumn>& columns)
{
    Q_FOREACH(const CollationColumn& collationColumn, columns) {
        if (collationColumn.column < 0 || collationColumn.column >= d->columnCount) {
            LQ_WARNING
                    << "collation column index" << collationColumn.column
                    << "is out of range for live query:" << d->origQuery.query();
            return;
        }
    }

    d->collationColumns = columns;
}

void TrackerLiveQuery::addUpdater(const TrackerPartialUpdater& updater)
{
    if (updater.d->schemas.isEmpty()) {
        LQ_WARNING
                << "No class watchers found for update query:"
                << updater.d->updateQuery;
        return;
    }

    Q_FOREACH(const PartialUpdateSchema* us, updater.d->schemas) {
        if (us->idColumn < -1 || us->idColumn >= d->columnCount) {
            LQ_WARNING
                    << "updater id column" << us->idColumn
                    << "for class" << us->change.klass
                    << "is out of range for live query:" << d->origQuery.query();
            return;
        }
    }

    d->updaters << new TrackerPartialUpdaterPrivate(*updater.d);

    // Add the watched classes and predicates to the ontology query.

    // We need the long name of the class for connecting to the GraphUpdated
    // signal from Tracker, and we need the tracker:id's of the properties for
    // processing the signal.

    Q_FOREACH(PartialUpdateSchema* us, updater.d->schemas) {
        d->ontologyQuerier.addClass(us->change.klass);
        Q_FOREACH(const QString& p, us->change.predicates) {
            d->ontologyQuerier.addPredicate(p);
        }
    }
}

void TrackerLiveQuery::addUpdater(const TrackerFullUpdater& updater)
{
    if (updater.d->changes.isEmpty()) {
        LQ_WARNING
                << "No class watchers found for full updater of live query:"
                << d->origQuery.query();
        return;
    }

    d->rerunners << new TrackerFullUpdaterPrivate(*updater.d);

    // Add the watched classes and predicates to the ontology query.

    // We need the long name of the class for connecting to the GraphUpdated
    // signal from Tracker, and we need the tracker:id's of the properties for
    // processing the signal.

    Q_FOREACH(const Change& change, updater.d->changes) {
        d->ontologyQuerier.addClass(change.klass);
        Q_FOREACH(const QString& p, change.predicates) {
            d->ontologyQuerier.addPredicate(p);
        }
    }
}

QAbstractItemModel* TrackerLiveQuery::model() const
{
    return &d->model;
}

bool TrackerLiveQuery::start()
{
    if (d->updaters.isEmpty() && d->rerunners.isEmpty()) {
        LQ_WARNING
                << "Cannot start because updaters not set for live query:"
                << d->origQuery.query();
        return false;
    }

    if (d->identityColumns.isEmpty()) {
        LQ_WARNING
                << "Cannot start because identity columns not set for live query:"
                << d->origQuery.query();
        return false;
    }

    if (d->performanceRecorder) d->performanceRecorder->markLiveQueryStarted();
    d->ontologyQuerier.start();
    return true;
}

void TrackerLiveQuery::setUpdatesEnabled(bool enable)
{
    d->updatesEnabledUser = enable;
    // run the pending updates, if any
    if (d->updatesEnabled && d->updatesEnabledUser)
        d->runUpdates();
}

#include "trackerlivequery.moc"
