/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "updater.h"

#include <QString>
#include <QStringList>
#include "updaterule.h"
#include "debug.h"

namespace
{
    const char* const filterTag = "%FILTER";
    const char* const idListTag = "%LIST";
}

PartialUpdateSchema::PartialUpdateSchema(const Change& change,
                                         const QString& filterSnippet,
                                         TrackerPartialUpdater::PlaceInSignal idPlace,
                                         int idColumn)
    : change(change),
      filterSnippet(filterSnippet), idPlace(idPlace),
      idColumn(idColumn)
{
}

bool PartialUpdateSchema::storeIds(const QString& classShort,
        const QList<TrackerChangeNotifier::Quad>& deletes,
        const QList<TrackerChangeNotifier::Quad>& inserts)
{
    bool idsAdded = false;
    if (change.klass == classShort) {
        if (change.type & Change::Delete) {
            storeSubjectOrObjectIds(idsAdded, deletes);
        }
        if (change.type & Change::Insert) {
            storeSubjectOrObjectIds(idsAdded, inserts);
        }
    }
    return idsAdded;
}

void PartialUpdateSchema::storeSubjectOrObjectIds
        (bool& idsAdded, const QList<TrackerChangeNotifier::Quad>& quads)
{
    Q_FOREACH(const TrackerChangeNotifier::Quad& q, quads) {
        if (change.hasPredicateId(q.predicate)) {
            if (idPlace == TrackerPartialUpdater::Subject)
                ids.insert(q.subject);
            else
                ids.insert(q.object);
            idsAdded = true;
        }
    }
}

TrackerPartialUpdaterPrivate::TrackerPartialUpdaterPrivate(const QString& updateQuery)
    : updateQuery(updateQuery)
{
    const int filterTagCount = updateQuery.count(filterTag);
    if (filterTagCount <= 0) {
        LQ_WARNING << filterTag << "missing in update query:" << updateQuery;
    }
    else if (filterTagCount > 1) {
        LQ_WARNING << "more than one" << filterTag << "found in update query:" << updateQuery;
    }
}

TrackerPartialUpdaterPrivate::TrackerPartialUpdaterPrivate(const TrackerPartialUpdaterPrivate& other)
    : updateQuery(other.updateQuery)
{
    Q_FOREACH(const PartialUpdateSchema* us, other.schemas) {
        schemas << new PartialUpdateSchema(*us);
    }
}

TrackerPartialUpdaterPrivate::~TrackerPartialUpdaterPrivate()
{
    qDeleteAll(schemas);
}

void TrackerPartialUpdaterPrivate::watchClass(const QString& classShort,
                                              const QStringList& predicates,
                                              unsigned int changeType,
                                              const QString& filterSnippet,
                                              TrackerPartialUpdater::PlaceInSignal placeInSignal,
                                              int idColumn)
{
    if (classShort.isEmpty()) {
        LQ_WARNING << "class to watch is empty for update query:" << updateQuery;
        return;
    }
    const int idListTagCount = filterSnippet.count(idListTag);
    if (idListTagCount <= 0) {
        LQ_WARNING << idListTag << "missing in filter snippet:" << filterSnippet;
        return;
    }
    else if (idListTagCount > 1) {
        LQ_WARNING << "more than one" << idListTag << "found in filter snippet:" << filterSnippet;
        return;
    }
    schemas << new PartialUpdateSchema(
        Change(classShort, predicates.toSet(), changeType),
        filterSnippet, placeInSignal,
        idColumn);
}

void TrackerPartialUpdaterPrivate::convertPredicates(const QHash<QString, int>& predicateToId)
{
    Q_FOREACH(PartialUpdateSchema* us, schemas) { // modifying Q_FOREACH
        Q_FOREACH(const QString& predicate, us->change.predicates)
            us->change.predicateIds.insert(predicateToId[predicate]);
        us->change.predicates.clear();
    }
}

bool TrackerPartialUpdaterPrivate::storeIds(const QString& classShort,
                                            const QList<TrackerChangeNotifier::Quad>& deletes,
                                            const QList<TrackerChangeNotifier::Quad>& inserts)
{
    // Note: this function stores the gathered id's into the
    // PartialUpdateSchema's.
    bool becomesActive = false;
    Q_FOREACH(PartialUpdateSchema* us, schemas) { // modifying Q_FOREACH
        if (us->storeIds(classShort, deletes, inserts))
            becomesActive = true;
    }
    return becomesActive;
}

QString TrackerPartialUpdaterPrivate::buildQuery(QList<UpdateRule>& rules)
{
    // Construct the query string (utilizing the id's gathered into the
    // PartialUpdateSchemas), then clear the id's.

    // Returns the real update query, plus a set of rules for determining the
    // affected rows. (Each PartialUpdateSchema adds one rule.) Each rule is of
    // the form (column, list of id's found in that column).

    // Update queries are of the form:
    // ... FILTER ( ... (1, 2, 3) ... || ... (4, 5, 6) ...) ...

    QStringList filterList;
    Q_FOREACH (PartialUpdateSchema* us, schemas) {
        if (us->ids.isEmpty())
            continue;

        QStringList idList;
        Q_FOREACH (const int& id, us->ids) {
            idList << QString::number(id);
        }
        QString list = QString("(") + idList.join(",") + ")";
        filterList << QString(us->filterSnippet).replace(idListTag, list);

        rules.append(UpdateRule(us->idColumn, us->ids));
        us->ids.clear();
    }
    QString filter = QString("FILTER(") + filterList.join(" || ") + ")";
    QString realUpdateQuery = updateQuery; // copy(-on-write)
    realUpdateQuery.replace(filterTag, filter);
    return realUpdateQuery;
}

TrackerPartialUpdater::TrackerPartialUpdater(const QString& updateQuery)
    : d(new TrackerPartialUpdaterPrivate(updateQuery))
{
}

TrackerPartialUpdater::TrackerPartialUpdater(const TrackerPartialUpdater& other)
    : d(new TrackerPartialUpdaterPrivate(*other.d))
{
}

TrackerPartialUpdater::~TrackerPartialUpdater()
{
    delete d;
}

TrackerPartialUpdater& TrackerPartialUpdater::operator=(const TrackerPartialUpdater& other)
{
    TrackerPartialUpdaterPrivate* newData = new TrackerPartialUpdaterPrivate(*other.d);
    delete d;
    d = newData;
    return *this;
}

void TrackerPartialUpdater::watchClass(const QString& classShort,
                                       const QStringList& predicates,
                                       const QString& filterSnippet,
                                       PlaceInSignal placeInSignal,
                                       int idColumn)
{
    d->watchClass(classShort, predicates, Change::Insert | Change::Delete,
                  filterSnippet, placeInSignal, idColumn);
}

void TrackerPartialUpdater::watchClassForInserts(const QString& classShort,
                                                 const QStringList& predicates,
                                                 const QString& filterSnippet,
                                                 PlaceInSignal placeInSignal,
                                                 int idColumn)
{
    d->watchClass(classShort, predicates, Change::Insert,
                  filterSnippet, placeInSignal, idColumn);
}

void TrackerPartialUpdater::watchClassForDeletes(const QString& classShort,
                                                 const QStringList& predicates,
                                                 const QString& filterSnippet,
                                                 PlaceInSignal placeInSignal,
                                                 int idColumn)
{
    d->watchClass(classShort, predicates, Change::Delete,
                  filterSnippet, placeInSignal, idColumn);
}

//////////////////////////////////////////////////////////////////////

TrackerFullUpdater::TrackerFullUpdater()
    : d(new TrackerFullUpdaterPrivate())
{
}

TrackerFullUpdater::~TrackerFullUpdater()
{
    delete d;
}

TrackerFullUpdater::TrackerFullUpdater(const TrackerFullUpdater& other)
    : d(new TrackerFullUpdaterPrivate(*other.d))
{
}

TrackerFullUpdater& TrackerFullUpdater::operator=(const TrackerFullUpdater& other)
{
    TrackerFullUpdaterPrivate* newData = new TrackerFullUpdaterPrivate(*other.d);
    delete d;
    d = newData;
    return *this;
}

void TrackerFullUpdater::watchClass(const QString& classShort,
                                    const QStringList& predicates)
{
    d->watchClass(classShort, predicates, Change::Insert | Change::Delete);
}

void TrackerFullUpdater::watchClassForInserts(const QString& classShort,
                                              const QStringList& predicates)
{
    d->watchClass(classShort, predicates, Change::Insert);
}

void TrackerFullUpdater::watchClassForDeletes(const QString& classShort,
                                              const QStringList& predicates)
{
    d->watchClass(classShort, predicates, Change::Delete);
}

void TrackerFullUpdaterPrivate::watchClass(const QString& classShort,
                                           const QStringList& predicates,
                                           unsigned int changeType)
{
    if (classShort.isEmpty()) {
        LQ_WARNING << "class to watch is empty.";
        return;
    }
    changes << Change(classShort, predicates.toSet(), changeType);
}

void TrackerFullUpdaterPrivate::convertPredicates(const QHash<QString, int>& predicateToId)
{
    for (int i = 0; i < changes.size(); ++i) {
        Q_FOREACH(const QString& predicate, changes[i].predicates)
            changes[i].predicateIds.insert(predicateToId[predicate]);
        changes[i].predicates.clear();
    }
}

bool TrackerFullUpdaterPrivate::checkSignal(const QString& classShort,
                                            const QList<TrackerChangeNotifier::Quad>& deletes,
                                            const QList<TrackerChangeNotifier::Quad>& inserts)
{
    Q_FOREACH(const Change& change, changes) {
        if (change.klass != classShort)
            continue;
        if (change.predicateIds.isEmpty()) {
            // empty list means: any predicate
            if ((change.type & Change::Delete) &&
                deletes.isEmpty() == false) {
                // we watch for deletes and something was deleted
                return true;
            }
            if ((change.type & Change::Insert) &&
                inserts.isEmpty() == false) {
                // we watch for inserts and something was inserted
                return true;
            }
            // nothing we watched for happend (e.g., we only watch for deletes,
            // and something was inserted).
            return false;
        }

        // predicate list nonempty -> check if the predicate matches
        if (change.type & Change::Delete) {
            Q_FOREACH(const TrackerChangeNotifier::Quad& q, deletes) {
                if (change.predicateIds.contains(q.predicate)) {
                    return true;
                }
            }
        }

        if (change.type & Change::Insert) {
            Q_FOREACH(const TrackerChangeNotifier::Quad& q, inserts) {
                if (change.predicateIds.contains(q.predicate)) {
                    return true;
                }
            }
        }
    }
    return false;
}
