/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "usermodifiablemodel.h"
#include <QLinkedList>
#include "updaterule.h"

UserModifiableModel::UserModifiableModel()
: insertBatchLocation(-1)
{
}

UserModifiableModel::~UserModifiableModel()
{
}

void UserModifiableModel::setIdentityColumns(const QList<int>& identityColumns)
{
    this->identityColumns = identityColumns;
}

bool UserModifiableModel::addRowWithData(int row, const QVariantList& data)
{
    const bool added = MovableRowModel::addRowWithData(row, data);
    if (added) {
        const IdentityList identity = IdentityList::fromData(data, identityColumns);
        identityToRowCache.insert(identity, QPersistentModelIndex(index(row, 0)));
    }
    return added;
}

bool UserModifiableModel::hasRowWithIdentity(const IdentityList& identity) const
{
    return identityToRowCache.contains(identity);
}

QPersistentModelIndex UserModifiableModel::identityToRow(const IdentityList& identity) const
{
    return identityToRowCache[identity];
}

void UserModifiableModel::beginLazyRowRemovals()
{
}

bool UserModifiableModel::lazyRemoveRow(const IdentityList& identity)
{
    // If the row was user-removed, we don't need to remember that any more.
    if (userRemovedRows.contains(identity)) {
        userRemovedRows.remove(identity);
        return true;
    }

    int ix = identityToRowCache.take(identity).row();
    return deleteRow(ix); // movableRowModel
}

void UserModifiableModel::endLazyRowRemovals()
{
}

void UserModifiableModel::beginLazyRowInsertions()
{
}

// This function doesn't really add the data to the model (until a suitable
// batch is found); that needs to be taken into account when computing row
// indices.  TODO: don't copy.
void UserModifiableModel::lazyInsertRow(int row, const QVariantList& data)
{
    const IdentityList identity = IdentityList::fromData(data, identityColumns);
    // If the row is user-removed, don't add it back.
    if (userRemovedRows.contains(identity)) {
        return;
    }

    if (row == insertBatchLocation) {
        // row goes to the already started batch
        insertBatchData.append(data);
        insertBatchIdentities.append(identity);
        return;
    }
    // else, insert the current batch (if any) and start a new one
    int insertedBatchSize = insertCurrentBatchIfAny();
    // adjust the "row" index, since we might have just added some rows
    startNewInsertBatch(row + insertedBatchSize, identity, data);
    return;
}

void UserModifiableModel::endLazyRowInsertions()
{
    insertCurrentBatchIfAny();
}

bool matchesIdentity(const QList<UpdateRule>& rules, const IdentityList& identity, const QList<int>& identityColumns)
{
    for (int i = 0; i < rules.size(); ++i) {
        const int indexOfIdCol = identityColumns.indexOf(rules[i].idColumn());
        if (indexOfIdCol >= 0 && rules[i].containsIdValue(identity[indexOfIdCol]))
            return true;
    }
    return false;
}

QLinkedList<IdentityList> UserModifiableModel::affectedRows(const QList<UpdateRule>& rules) const
{
    QLinkedList<IdentityList> toReturn;
    // It's important that affected rows are stored in the same order than they
    // appear in the model.
    for (int r = 0; r < rowCount(); ++r) { // FIXME: slow!
        for (int i = 0; i < rules.size(); ++i) {
            const QModelIndex ix = index(r, rules[i].idColumn());
            if (rules[i].containsIdValue(data(ix).toInt())) {
                toReturn.append(IdentityList::fromRow(*this, r, identityColumns));
                break; // no need to check the rest of the rules
            }
        }
    }

    for (QSet<IdentityList>::const_iterator i = userRemovedRows.constBegin();
         i != userRemovedRows.constEnd(); ++i) {
        if (matchesIdentity(rules, *i, identityColumns)) {
            toReturn.append(*i);
        }
    }

    return toReturn;
}

bool UserModifiableModel::removeRows(int row, int count, const QModelIndex & /*parent*/)
{
    if (row < 0 || count < 0 || (row+count) > rowCount() )
        return false;

    for (int r = row; r < row+count; ++r) {
        const IdentityList rowId = IdentityList::fromRow(*this, r, identityColumns);
        identityToRowCache.remove(rowId);
        userRemovedRows.insert(rowId);
    }
    return MovableRowModel::deleteRows(row, count);
}

void UserModifiableModel::startNewInsertBatch(int row,
                                              const IdentityList& identity,
                                              const QVariantList& data)
{
    insertBatchLocation = row;
    insertBatchData.append(data);
    insertBatchIdentities.append(identity);
}

int UserModifiableModel::insertCurrentBatchIfAny()
{
    if (insertBatchLocation == -1)
        return 0;
    int toReturn = insertBatchData.size();

    beginAddRowsBatch(insertBatchLocation, insertBatchData.size());
    for (int offset = 0; offset < insertBatchData.size(); ++offset) {
        MovableRowModel::addRowWithData(insertBatchLocation + offset,
                                        insertBatchData[offset]);

    }
    endAddRowsBatch();

    // Get the QPersistentModelIndices when the data is really in the model.
    // TODO: this fails if we try to get the indices in the previous loop.  Find
    // out exactly why.
    for (int offset = 0; offset < insertBatchData.size(); ++offset) {
        identityToRowCache.insert(insertBatchIdentities[offset],
                                  QPersistentModelIndex(index(insertBatchLocation + offset, 0)));
    }

    insertBatchLocation = -1;
    insertBatchData.clear();
    insertBatchIdentities.clear();
    return toReturn;
}
