/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "TrackerLiveQuery"

#include "../utils/livetesthelpers.h"
#include "../utils/signalsrecorder.h"

#include <QtTest/QtTest>
#include <QVariant>
#include <QtSparql/QtSparql>

class tst_TrackerLiveBenchmark : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveBenchmark();
    virtual ~tst_TrackerLiveBenchmark();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void measureInitialModelCreation_data();
    void measureInitialModelCreation();

private:
    bool createTestData(int testDataAmount);
    void cleanupTestData();

private:
    QSparqlConnection* conn;
    TrackerLiveQuery* liveQuery;
    SignalsRecorder* signalsRecorder;
};

namespace {
    const char* const benchmarkEnableEnvName = "LIVEQUERY_BENCHMARK";
    const char* const qsparqlDriver = "QTRACKER_DIRECT";
    const char* const originalQueryTemplate =
            "select tracker:id(?musicPiece) ?title ?performer ?album ?duration ?created "
                "{ "
                "?musicPiece a nmm:MusicPiece; "
                "nie:isLogicalPartOf <tracker-live-tests> ;"
                "nie:title ?title; "
                "nmm:performer ?performer; "
                "nmm:musicAlbum ?album; "
                "nfo:duration ?duration; "
                "nie:contentCreated ?created ."
                "%1 } order by ?title ?created";
    const QString originalQuery = QString(originalQueryTemplate).arg("");
    const int columnCount = 6;
    const QList<int> identityColumns = QList<int>() << 0;
    const QList<TrackerLiveQuery::CollationColumn> collationColumns =
            QList<TrackerLiveQuery::CollationColumn>()
                << TrackerLiveQuery::CollationColumn(1, QVariant::String)
                << TrackerLiveQuery::CollationColumn(5, QVariant::DateTime);
    const int insertBatchSize   = 200;
    const int dataReadyInterval = 100;
    const QString updateQuery = QString(originalQueryTemplate).arg("%FILTER");
}

tst_TrackerLiveBenchmark::tst_TrackerLiveBenchmark()
    : conn(0), liveQuery(0), signalsRecorder(0)
{
}

tst_TrackerLiveBenchmark::~tst_TrackerLiveBenchmark()
{
}

void tst_TrackerLiveBenchmark::initTestCase()
{
    QSparqlConnectionOptions connOptions;
    connOptions.setDataReadyInterval(dataReadyInterval);
    conn = new QSparqlConnection(qsparqlDriver, connOptions);

    cleanupTestData();

    // Enable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "1");

    qDebug() << "QSparql driver:" << QString(qsparqlDriver);
    qDebug() << "Data ready interval:" << dataReadyInterval;
}

void tst_TrackerLiveBenchmark::cleanupTestCase()
{
    // Disable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "");

    cleanupTestData();

    delete conn;
    conn = 0;
}

void tst_TrackerLiveBenchmark::init()
{
    liveQuery = new TrackerLiveQuery(originalQuery, columnCount,
                                     identityColumns, *conn);
    liveQuery->setCollationColumns(collationColumns);
    TrackerPartialUpdater updater(updateQuery);
    updater.watchClass("nmm:MusicPiece", QStringList(),
                       "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(updater);

    signalsRecorder = new SignalsRecorder;
}

void tst_TrackerLiveBenchmark::cleanup()
{
    delete signalsRecorder;
    signalsRecorder = 0;

    delete liveQuery;
    liveQuery = 0;

    cleanupTestData();
}

void tst_TrackerLiveBenchmark::measureInitialModelCreation_data()
{
    QTest::addColumn<int>("testDataAmount");
    for (int testDataAmount = 1000; testDataAmount <= 5000; testDataAmount += 1000) {
        QTest::newRow(QString("%1 rows").arg(testDataAmount).toAscii().data())
                << testDataAmount;
    }
}

void tst_TrackerLiveBenchmark::measureInitialModelCreation()
{
    QFETCH(int, testDataAmount);
    createTestData(testDataAmount);

    SignalList sigs;
    sigs.append(liveQuery, SIGNAL(initialQueryFinished()));
    signalsRecorder->listen(sigs);

    liveQuery->start();

    signalsRecorder->waitForSignals(sigs, 5000);
    QCOMPARE( liveQuery->model()->rowCount(), testDataAmount );

    const qint64 liveQueryStartTime = liveQuery->property("test-livequerystart").toLongLong();

    // Sanity-check the performance counter values
    QVERIFY( liveQueryStartTime < liveQuery->property("test-initialdataready").toLongLong() );
    QVERIFY( liveQueryStartTime < liveQuery->property("test-ontologyquerycomplete").toLongLong() );
    QVERIFY( liveQuery->property("test-ontologyquerycomplete").toLongLong() <= liveQuery->property("test-originalquerystart").toLongLong() );
    QVERIFY( liveQuery->property("test-originalquerystart").toLongLong() < liveQuery->property("test-firstbatchoforiginalqueryavailable").toLongLong() );
    QVERIFY( liveQuery->property("test-firstbatchoforiginalqueryavailable").toLongLong() < liveQuery->property("test-originalqueryprocessingfinished").toLongLong() );
    QVERIFY( liveQuery->property("test-originalqueryprocessingfinished").toLongLong() <= liveQuery->property("test-initialdataready").toLongLong() );

    // Print out the performance counter values to debug out
    qDebug() << "Ontology query completed at:"
            << liveQuery->property("test-ontologyquerycomplete").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "Original query started at:"
            << liveQuery->property("test-originalquerystart").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "First batch of original query available at:"
            << liveQuery->property("test-firstbatchoforiginalqueryavailable").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "First block of original data written in to the model at:"
            << liveQuery->property("test-firstbatchoforiginaldatainmodel").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "Original query processing completed at:"
            << liveQuery->property("test-originalqueryprocessingfinished").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "Original query completion notified and processed by client at:"
            << liveQuery->property("test-initialdataready").toLongLong() - liveQueryStartTime
            << "ms";

    qDebug() << "Model building time in original query processing:"
            << liveQuery->property("test-modelbuildtime").toInt()
            << "ms";
    if (liveQuery->property("test-modelbuildtime").toInt() < testDataAmount/dataReadyInterval) {
        // Some model updates may be missed if they are shorter than 1 ms
        qDebug() << "Warning: Model building time below 1 ms measurement resolution";
    }

    qDebug() << "Worst case block of original query processing took:"
            << liveQuery->property("test-maxmodelblockbuildtime").toInt()
            << "ms";
}

bool tst_TrackerLiveBenchmark::createTestData(int testDataAmount)
{
    const QString insertQueryTemplate =
            "<urn:music:%1> a nmm:MusicPiece, nfo:FileDataObject, nfo:Audio;"
                "nie:isLogicalPartOf <tracker-live-tests> ;"
                "tracker:available          true ;"
                "nie:byteSize               \"0\" ;"
                "nie:url                    \"file://music/Song_%1.mp3\" ;"
                "nfo:belongsToContainer     <file://music/> ;"
                "nie:title                  \"Song %1\" ;"
                "nie:mimeType               \"audio/mpeg\" ;"
                "nie:contentCreated         \"2000-01-01T01:01:01Z\" ;"
                "nie:isLogicalPartOf        <urn:album:%2> ;"
                "nco:contributor            <urn:artist:%2> ;"
                "nfo:fileLastAccessed       \"2000-01-01T01:01:01Z\" ;"
                "nfo:fileSize               \"0\" ;"
                "nfo:fileName               \"Song_%1.mp3\" ;"
                "nfo:fileLastModified       \"2000-01-01T01:01:01Z\" ;"
                "nfo:codec                  \"MPEG\" ;"
                "nfo:averageBitrate         \"16\" ;"
                "nfo:genre                  \"Genre %2\" ;"
                "nfo:channels               \"2\" ;"
                "nfo:sampleRate             \"44100.0\" ;"
                "nmm:musicAlbum             <urn:album:%2> ;"
                "nmm:musicAlbumDisc         <urn:album-disk:%2> ;"
                "nmm:performer              <urn:artist:%2> ;"
                "nfo:duration               \"1\" ;"
                "nmm:trackNumber            \"1\" .";

    for (int item = 1; item <= testDataAmount; ) {
        QString insertQuery = "insert { ";
        int itemDozen = 10;
        const int batchEnd = item + insertBatchSize;
        for (; item < batchEnd && item <= testDataAmount; ++item) {
            insertQuery.append( insertQueryTemplate.arg(item).arg(itemDozen) );
            if (item % 10 == 0) itemDozen += 10;
        }
        insertQuery.append(" }");
        if (!executeQuery(QSparqlQuery(insertQuery, QSparqlQuery::InsertStatement), false))
            return false;
    }

    // Wait for tracker signaling to finish
    QTest::qWait(4000);

    return true;
}

void tst_TrackerLiveBenchmark::cleanupTestData()
{
    executeQuery(
            *conn,
            QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

QTEST_MAIN( tst_TrackerLiveBenchmark )
#include "tst_live_benchmark.moc"
