/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/signalsrecorder.h"

#include "TrackerLiveQuery"

#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

//const QString qtest(qTableName( "qtest", __FILE__ )); // FIXME: what's this

//TESTED_FILES=

namespace {
    const char* const benchmarkEnableEnvName = "LIVEQUERY_BENCHMARK";
}

class UpdateAndResult
{
public:
    virtual ~UpdateAndResult()
    {
    }
    virtual void doAction() const
    {
    }
    virtual bool isNoOperation() const
    {
        return false;
    }
    virtual bool waitForResults() const
    {
        // If true, we need to wait for the results of this action to appear in
        // the model.
        return false;
    }
    virtual QString description() const
    {
        return "";
    }

    int rowsInsertedCount, rowsRemovedCount, rowsMovedCount, dataChangedCount;
    QStringList dataAfter;
    int queryCount, requeryCount;

    UpdateAndResult()
    : rowsInsertedCount(0), rowsRemovedCount(0),
      rowsMovedCount(0), dataChangedCount(0), queryCount(0), requeryCount(0)
    { }

    UpdateAndResult(int ric, int rrc, int rmc, int dcc,
                    const QStringList& da,
                    int qc = 1, int rc = 0)
    : rowsInsertedCount(ric), rowsRemovedCount(rrc),
      rowsMovedCount(rmc), dataChangedCount(dcc),
      dataAfter(da),
      queryCount(qc),
      requeryCount(rc)
    { }
};

class NoOperationUpdate : public UpdateAndResult
{
public:
    virtual bool isNoOperation() const
    {
        return true;
    }
    virtual QString description() const
    {
        return "nothing";
    }
};

class UpdateQuery : public UpdateAndResult
{
public:
    QSparqlConnection* conn; // not owned
    QString updateQuery;
    UpdateQuery(QSparqlConnection* c, const QString& uq,
                int ric, int rrc, int rmc, int dcc,
                const QStringList& da,
                int qc = 1, int rc = 0)
    : UpdateAndResult(ric, rrc, rmc, dcc, da, qc, rc),
      conn(c), updateQuery(uq)
    { }

    virtual void doAction() const
    {
        QSparqlResult* res = conn->exec(QSparqlQuery(updateQuery,
                                                     QSparqlQuery::InsertStatement));
        res->waitForFinished();
        delete res;
    }
    virtual QString description() const
    {
        return updateQuery;
    }
    virtual bool waitForResults() const
    {
        return true; // we need to wait for the results to be in the model
    }
};

// TODO: extend this when other user-modifiablity operations than the removal are added
class UserRowRemoval : public UpdateAndResult
{
public:
    QAbstractItemModel* model;
    int start, count;
    UserRowRemoval(QAbstractItemModel* m, int s, int c,
                   int ric, int rrc, int rmc, int dcc,
                   const QStringList& da,
                   int qc = 1, int rc = 0)
    : UpdateAndResult(ric, rrc, rmc, dcc, da, qc, rc),
      model(m), start(s), count(c)
    {
    }
    virtual void doAction() const
    {
        model->removeRows(start, count);
    }
    virtual QString description() const
    {
        return QString("removeRows(%1, %2)").arg(start).arg(count);
    }
};

Q_DECLARE_METATYPE(UpdateAndResult*)
Q_DECLARE_METATYPE(QList<UpdateAndResult*>)
Q_DECLARE_METATYPE(QList<int>)
Q_DECLARE_METATYPE(QModelIndex)

class tst_TrackerLiveDataDriven : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveDataDriven();
    virtual ~tst_TrackerLiveDataDriven();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void test();
    void test_data();

private:
    void test_data_resourceAndSortedProperty();
    void test_data_primaryAndSecondaryResource();
    void test_data_dateTimeComparison();
    void test_data_unrelatedChange();
    void test_data_initialRowsInserted();
    void test_data_streamInitialRows();
    void test_data_fillAndEmpty();
    void test_data_watchOnlyInserts();
    void test_data_watchOnlyDeletes();
    void test_data_rerunForInserts();
    void test_data_rerunForDeletes();
    void test_data_groupedInserts();
    void test_data_userRowRemoval();
    void test_data_userRowRemovalWithoutAffectedRows();

private:
    QSparqlConnectionOptions defaultOptions;
};

tst_TrackerLiveDataDriven::tst_TrackerLiveDataDriven()
{
    qRegisterMetaType<UpdateAndResult*>();
    qRegisterMetaType<QList<UpdateAndResult*> >();
    qRegisterMetaType<QList<int> >();
    qRegisterMetaType<QModelIndex>();
}

tst_TrackerLiveDataDriven::~tst_TrackerLiveDataDriven()
{
}

void tst_TrackerLiveDataDriven::initTestCase()
{
    // Clean up leftover data from previous test run
    cleanup();
    executeQuery(QSparqlQuery("insert { "
                              "<tracker-live-tests> a nie:InformationElement . }",
                              QSparqlQuery::InsertStatement));
    // Enable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "1");

    // Set a large data ready interval so that the initial data will come in one
    // chunk for most of the tests. (StreamInitialRows test a case where it
    // doesn't.)
    defaultOptions.setDataReadyInterval(100);
}

void tst_TrackerLiveDataDriven::cleanupTestCase()
{
    // Disable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "");
}

void tst_TrackerLiveDataDriven::init()
{
}

void tst_TrackerLiveDataDriven::cleanup()
{
    // Clean up test data
    executeQuery(QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

void tst_TrackerLiveDataDriven::test()
{
    QFETCH(void*, connectionVoid);
    QFETCH(void*, liveQueryVoid);

    QSparqlConnection* conn = (QSparqlConnection*) connectionVoid;
    TrackerLiveQuery* liveQuery = (TrackerLiveQuery*) liveQueryVoid;

    QAbstractItemModel* model = liveQuery->model();
    QFETCH(QList<int>, columnsToCheck);
    QFETCH(QList<UpdateAndResult*>, updatesAndResults);

    bool started = false;

    int i = 0;
    Q_FOREACH(const UpdateAndResult* up, updatesAndResults) {
        qDebug() << "Update" << ++i << "/" << updatesAndResults.size();
        if (up->isNoOperation() && !started) {
            // No initial data wanted, just start the live model
            liveQuery->start();
            started = true;
            // Give the TrackerLiveQuery some time to execute its startup queries, so
            // that they won't be confused with the updates.
            QTest::qWait(2000);
            continue;
        }
        else if (up->isNoOperation()) {
            qWarning() << "test(): ignoring an empty update";
            continue;
        }

        // measure additional queries caused by this step
        liveQuery->setProperty("test-updatequerycount", QVariant(0));
        liveQuery->setProperty("test-requerycount", QVariant(0));

        SignalsRecorder rec;
        SignalList sigs;

        if (up->rowsInsertedCount > 0) // these checks should be unnecessary
            sigs.append(model,
                        SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                        up->rowsInsertedCount);
        if (up->rowsRemovedCount > 0)
            sigs.append(model,
                        SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
                        up->rowsRemovedCount);
        if (up->rowsMovedCount > 0)
            sigs.append(model,
                        SIGNAL(rowsMoved(const QModelIndex&, int, int,
                                         const QModelIndex&, int)),
                        up->rowsMovedCount);
        if (up->dataChangedCount > 0)
            sigs.append(model,
                        SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
                        up->dataChangedCount);
        rec.listen(sigs);

        // Create another signals recorder to listen to all signals to verify no unwanted
        // signals are emitted
        SignalList allSigs;
        allSigs.append(model,
                    SIGNAL(rowsInserted(const QModelIndex&, int, int)));
        allSigs.append(model,
                    SIGNAL(rowsRemoved(const QModelIndex&, int, int)));
        allSigs.append(model,
                    SIGNAL(rowsMoved(const QModelIndex&, int, int,
                                     const QModelIndex&, int)));
        allSigs.append(model,
                    SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
        SignalsRecorder allSigsRec;
        allSigsRec.listen(allSigs);

        // Do the actual action (update query or user-operation on the model)
        up->doAction();
        if (!started) {
            // We're inserting initial data.  Here we need to wait a bit to
            // drain the signals so that they won't make our model think it
            // needs to do update queries.
            QTest::qWait(3000);

            // We start the liveQuery this late, to be able to insert initial
            // data (index 0 of updatesAndResults)
            liveQuery->start();
            started = true;
        }

        if (up->waitForResults()) {
            // Wait until the signals have arrived
            bool signalsArrived = rec.waitForSignals(sigs, 5000);

            if (!signalsArrived) {
                // print out error information
                qDebug() << "Failure when executing update" << up->description();
                qDebug() << "Received signals:";
                Q_FOREACH(const SignalsRecorder::SignalData &sd, rec.allSignalData()) {
                    qDebug() << sd.signal() << sd.params();
                }
                qDebug() << "Data after the update:";
                printModel(model);
            }
            QVERIFY(signalsArrived);
        }

        // Verify the signal counts
        int actualRowsInserted =
            rec.signalData(model,
                           SIGNAL(rowsInserted(const QModelIndex&, int, int))).size();
        int actualRowsRemoved =
            rec.signalData(model,
                           SIGNAL(rowsRemoved(const QModelIndex&, int, int))).size();

        int actualRowsMoved =
            rec.signalData(model,
                           SIGNAL(rowsMoved(const QModelIndex&, int, int,
                                            const QModelIndex&, int))).size();

        if (actualRowsInserted != up->rowsInsertedCount ||
            actualRowsRemoved != up->rowsRemovedCount ||
            actualRowsMoved != up->rowsMovedCount) {
            qDebug() << "Failure when executing update" << up->description();
            qDebug() << "Data after the update:";
            printModel(model);
        }

        QCOMPARE(actualRowsInserted, up->rowsInsertedCount);
        QCOMPARE(actualRowsRemoved, up->rowsRemovedCount);
        QCOMPARE(actualRowsMoved, up->rowsMovedCount);

        // If we expect no signals (= we're testing that something didn't
        // happen), wait a bit here so that there are some changes when that
        // something could have happened.
        if (up->rowsInsertedCount == 0 && up->rowsRemovedCount  == 0 &&
            up->rowsMovedCount == 0 && up->dataChangedCount == 0) {
            QTest::qWait(4000);
            QCOMPARE( allSigsRec.allSignalData().count(), 0 );
        }

        // Verify the data
        QStringList actual = getRows(model, columnsToCheck);
        if (actual != up->dataAfter) {
            qDebug() << "Failure when executing update" << up->description();
            qDebug() << "Data after the update:";
            printModel(model);
        }
        QCOMPARE(actual, up->dataAfter);

        // When the delays in the system are long, we cannot ensure that the
        // several D-Bus signals will be delivered before TrackerLiveQuery decides to
        // run an update.  If some signals are still undelivered when the
        // decision to run an update is made, TrackerLiveQuery will run unnecessary
        // queries but the data should still be correct.  Thus, don't assert the
        // query count by default.
        if (liveQuery->property("test-updatequerycount").toInt() != up->queryCount) {
            qWarning() << "Update query count mismatch when executing update"
                       << up->description();
            qWarning() << "Actual:" << liveQuery->property("test-updatequerycount").toInt()
                       << "expected:" << up->queryCount;
        }

        if (liveQuery->property("test-requerycount").toInt() != up->requeryCount) {
            qWarning() << "Requery count mismatch when executing update"
                       << up->description();
            qWarning() << "Actual:" << liveQuery->property("test-requerycount").toInt()
                       << "expected:" << up->requeryCount;
        }

        // Run the event loop; Qt deletes some things in a delayed way.  We need
        // this for memory-related tests.
        while(QCoreApplication::hasPendingEvents())
        {
            QCoreApplication::sendPostedEvents();
            QCoreApplication::sendPostedEvents(0, QEvent::DeferredDelete);
            QCoreApplication::processEvents();
        }
    }
    qDeleteAll(updatesAndResults);

    delete liveQuery;
    delete conn;
}

void tst_TrackerLiveDataDriven::test_data()
{
    // Definition of what consitutes a test case
    QTest::addColumn<void*>("connectionVoid");
    QTest::addColumn<void*>("liveQueryVoid");
    QTest::addColumn<QList<int> >("columnsToCheck");
    QTest::addColumn<QList<UpdateAndResult*> >("updatesAndResults");

    // Test cases:
    test_data_resourceAndSortedProperty();
    test_data_primaryAndSecondaryResource();
    test_data_dateTimeComparison();
    test_data_unrelatedChange();
    test_data_initialRowsInserted();
    test_data_streamInitialRows();
    //test_data_fillAndEmpty(); // slow
    test_data_watchOnlyInserts();
    test_data_watchOnlyDeletes();
    test_data_rerunForInserts();
    test_data_rerunForDeletes();

    test_data_groupedInserts();

    test_data_userRowRemoval();
    test_data_userRowRemovalWithoutAffectedRows();
}

void tst_TrackerLiveDataDriven::test_data_resourceAndSortedProperty()
{
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    updates << new UpdateQuery(conn,
                               "insert { <mp1> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 100 .}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp1:100:");

    updates << new UpdateQuery(conn,
                               "insert { <mp2> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 80 .}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp2:80:"
                               << "mp1:100:");

    updates << new UpdateQuery(conn,
                               "insert { <mp3> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 90 .}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp2:80:"
                               << "mp3:90:"
                               << "mp1:100:");

    // Change a property which is relevant for sorting
    updates << new UpdateQuery(conn,
                               "delete { <mp2> nfo:duration 80 . }"
                               "insert { <mp2> nfo:duration 110 . }",
                               0, 0, 1, 0, // signals
                               QStringList()
                               << "mp3:90:"
                               << "mp1:100:"
                               << "mp2:110:");

    updates << new UpdateQuery(conn,
                               "delete { <mp1> nfo:duration 100 . }"
                               "insert { <mp1> nfo:duration 105 . }",
                               0, 0, 0, 1, // signals
                               QStringList()
                               << "mp3:90:"
                               << "mp1:105:"
                               << "mp2:110:");

    updates << new UpdateQuery(conn,
                               "delete { <mp2> nfo:duration 110 . }"
                               "insert { <mp2> nfo:duration 60 . }",
                               0, 0, 0, 1, // signals
                               QStringList()
                               << "mp2:60:"
                               << "mp3:90:"
                               << "mp1:105:");

    updates << new UpdateQuery(conn,
                               "insert { <mp4> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 91 .}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp2:60:"
                               << "mp3:90:"
                               << "mp4:91:"
                               << "mp1:105:");

    updates << new UpdateQuery(conn,
                               "delete { <mp3> a rdfs:Resource .}",
                               0, 1, 0, 0, // signals
                               QStringList()
                               << "mp2:60:"
                               << "mp4:91:"
                               << "mp1:105:");

    updates << new UpdateQuery(conn,
                               "delete { <mp2> a rdfs:Resource .}",
                               0, 1, 0, 0, // signals
                               QStringList()
                               << "mp4:91:"
                               << "mp1:105:");

    updates << new UpdateQuery(conn,
                               "delete { <mp1> a rdfs:Resource .}",
                               0, 1, 0, 0, // signals
                               QStringList()
                               << "mp4:91:");

    updates << new UpdateQuery(conn,
                               "delete { <mp4> a rdfs:Resource .}",
                               0, 1, 0, 0, // signals
                               QStringList());

    QTest::newRow("ResourceAndSortedProperty")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_primaryAndSecondaryResource()
{
    QString query = "select tracker:id(?ch) ?ch ?cht tracker:id(?m) ?m ?mt "
        "{ ?ch a mfo:FeedChannel ; nie:title ?cht ; "
        "nie:isLogicalPartOf <tracker-live-tests> ."
        "?m a mfo:FeedMessage ; nie:isLogicalPartOf ?ch ; nie:title ?mt ; "
        "nie:isLogicalPartOf <tracker-live-tests> . "
        "%1 }"
        "order by ?cht ?mt";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 6,
                                                       QList<int>() << 0 << 3,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("mfo:FeedChannel",
                  QStringList() << "nie:title" << "rdf:type",
                  "tracker:id(?ch) in %LIST", TrackerPartialUpdater::Subject, 0);
    up.watchClass("mfo:FeedMessage",
                  QStringList() << "nie:title" << "nie:isLogicalPartOf" << "rdf:type",
                  "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 3);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::String) <<
                                   TrackerLiveQuery::CollationColumn(5, QVariant::String));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data:
    updates << new UpdateQuery(conn,
                               "insert { <ch1> a mfo:FeedChannel ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nie:title \"foo\" . "
                               "<msg1> a mfo:FeedMessage ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nie:isLogicalPartOf <ch1> ;"
                               "nie:title \"b\" . }",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "ch1:foo:msg1:b:",
                               0); // no of update queries (initial query doesn't count)

    // Insert another channel with the same title...
    updates << new UpdateQuery(conn,
                               "insert { <ch2> a mfo:FeedChannel ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nie:title \"foo\" . "
                               "<msg2> a mfo:FeedMessage ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nie:isLogicalPartOf <ch2> ;"
                               "nie:title \"a\" . "
                               "<msg3> a mfo:FeedMessage ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nie:isLogicalPartOf <ch2> ;"
                               "nie:title \"c\" . }",
                               2, 0, 0, 0,
                               QStringList()
                               << "ch2:foo:msg2:a:"
                               << "ch1:foo:msg1:b:"
                               << "ch2:foo:msg3:c:");

    QTest::newRow("PrimaryAndSecondardyResource")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2 << 4 << 5) // columns to check
        << updates;
}


void tst_TrackerLiveDataDriven::test_data_dateTimeComparison()
{
    QString query = "select tracker:id(?m) ?m ?d "
        "{ ?m a nmo:Message ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nmo:sentDate ?d . %1 }"
        " order by ?d";
    // Use the D-Bus driver here, since we want to bypass QDateTime creation and
    // the mangling of the date time that it does (e.g., no public api for
    // offsets, string -> datetime -> string conversion doens't give back the
    // same string).
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER");
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmo:Message",
                  QStringList() << "nmo:sentDate" << "rdf:type",
                  "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::DateTime));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    // First: all UTC time
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T10:43:00Z\" . " // UTC
                               "<m2> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T12:00:00Z\" . " // UTC
                               "<m3> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T11:50:00Z\" . " // UTC
                               "}",
                               // note: only one rowsInserted signal, since the
                               // rows are consecutive
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:1998-12-10T10:43:00Z:"
                               << "m3:1998-12-10T11:50:00Z:"
                               << "m2:1998-12-10T12:00:00Z:");

    // Test that new UTC times are sorter properly
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m4> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T11:00:00Z\" ." // UTC
                               "<m5> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T11:55:00Z\" ." // UTC
                               "}",
                               2, 0, 0, 0, // signals
                               QStringList()
                               << "m1:1998-12-10T10:43:00Z:"
                               << "m4:1998-12-10T11:00:00Z:"
                               << "m3:1998-12-10T11:50:00Z:"
                               << "m5:1998-12-10T11:55:00Z:"
                               << "m2:1998-12-10T12:00:00Z:");

    // Passing these test cases is not very interesting, since string sorting is
    // also good for these UTC dates.


    // Test that new non-UTC times are sorter properly. Not that interesting,
    // since tracker converts the times to UTC.
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m6> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nmo:sentDate \"1998-12-10T11:30:00-02:00\" ."
                               "}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:1998-12-10T10:43:00Z:"
                               << "m4:1998-12-10T11:00:00Z:"
                               << "m3:1998-12-10T11:50:00Z:"
                               << "m5:1998-12-10T11:55:00Z:"
                               << "m2:1998-12-10T12:00:00Z:"
                               << "m6:1998-12-10T13:30:00Z:");

    // TODO: add more interesting tests cases.

    QTest::newRow("DateTimeComparison")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_unrelatedChange()
{
    QString query = "select tracker:id(?m) ?m "
        "{ ?m a nmo:Message ; "
        "  nie:isLogicalPartOf <tracker-live-tests> . "
        "  %1 } order by tracker:id(?m)";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 2,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmo:Message",
                  QStringList() << "rdf:type",
                  "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(0, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nie:title \"title\" . " // title is not watched
                               "<m2> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> ; "
                               "  nie:title \"another\" . "
                               "}",
                               // note: only one rowsInserted signal, since the
                               // rows are consecutive
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:"
                               << "m2:");

    QTest::newRow("UnrelatedChange")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_initialRowsInserted()
{
    // This tests that we get only one rowsInserted signal for the initial data
    // set, even if it contains multiple rows.
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data set contains 2 rows
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<mp1> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 100 . "
                               "<mp2> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 200 . "
                               "}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp1:100:"
                               << "mp2:200:",
                               0); // update query count: initial query doesn't count

    QTest::newRow("InitialRowsInserted")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_streamInitialRows()
{
    // This tests that we get only one rowsInserted signal for the initial data
    // set, even if it contains multiple rows.
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnectionOptions opts;
    opts.setDataReadyInterval(5);
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", opts);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data set contains 10 rows
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<mp01> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 100 . "
                               "<mp02> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 200 . "
                               "<mp03> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 300 . "
                               "<mp04> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 400 . "
                               "<mp05> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 500 . "
                               "<mp06> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 600 . "
                               "<mp07> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 700 . "
                               "<mp08> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 800 . "
                               "<mp09> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 900 . "
                               "<mp10> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 1000 . "
                               "}",
                               // note: 2 rowsinserted signals since dataReadyInterval == 5
                               2, 0, 0, 0, // signals
                               QStringList()
                               << "mp01:100:"
                               << "mp02:200:"
                               << "mp03:300:"
                               << "mp04:400:"
                               << "mp05:500:"
                               << "mp06:600:"
                               << "mp07:700:"
                               << "mp08:800:"
                               << "mp09:900:"
                               << "mp10:1000:",
                               0); // update query count: initial query doesn't count

    QTest::newRow("StreamInitialRows")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_fillAndEmpty()
{
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    QStringList dataNow;
    for (int i = 0; i < 10; ++i) {
        QString query = QString("insert { <mp%1> a nmm:MusicPiece ; "
                                "nie:isLogicalPartOf <tracker-live-tests> ; "
                                "nfo:duration %1 .}").arg(i);
        dataNow << QString("mp%1:%1:").arg(i);
        updates << new UpdateQuery(conn, query,
                                   1, 0, 0, 0, // signals
                                   dataNow);
    }

    for (int i = 0; i < 10; ++i) {
        QString query = QString("delete { <mp%1> a rdfs:Resource . } ").arg(i);
        dataNow.removeFirst();
        updates << new UpdateQuery(conn, query,
                                   0, 1, 0, 0, // signals
                                   dataNow);
    }

    QTest::newRow("FillAndEmpty")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_watchOnlyInserts()
{
    QString query = "select tracker:id(?m) ?m ?t"
        "{ ?m a nmo:Message ; "
        "  nie:title ?t ; "
        "  nie:isLogicalPartOf <tracker-live-tests> . "
        "  %1 } order by tracker:id(?m)";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClassForInserts("nmo:Message",
                            QStringList() << "rdf:type" << "nie:title",
                            "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(0, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:title \"t1\" ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "<m2> a nmo:Message ; "
                               "  nie:title \"t2\" ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               // note: only one rowsInserted signal, since the
                               // rows are consecutive
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:t1:"
                               << "m2:t2:");

    // and delete a property (but deletes are not watched)
    updates << new UpdateQuery(conn,
                               "delete { "
                               "<m1> nie:title \"t1\" . "
                               "}",
                               0, 0, 0, 0, // signals
                               QStringList()
                               << "m1:t1:"
                               << "m2:t2:",
                               0); // no updates
    // Note: now nothing should happen (no row removals or insertions, so it's
    // essential that the test function waits a bit in case all signals are 0.

    QTest::newRow("WatchOnlyInserts")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_watchOnlyDeletes()
{
    QString query = "select tracker:id(?m) ?m ?t"
        "{ ?m a nmo:Message ; "
        "  nie:title ?t ; "
        "  nie:isLogicalPartOf <tracker-live-tests> . "
        "  %1 } order by tracker:id(?m)";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClassForDeletes("nmo:Message",
                            QStringList() << "rdf:type",
                            "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);

    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(0, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:title \"t1\" ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "<m2> a nmo:Message ; "
                               "  nie:title \"t2\" ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               // note: only 1 rowsInserted signal since the
                               // initial data is inserted in one big bunch
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:t1:"
                               << "m2:t2:",
                               0); // no of update queries (initial query doesn't count)

    // delete a resource, this should be caught normally.
    updates << new UpdateQuery(conn,
                               "delete { "
                               "<m1> a rdfs:Resource . "
                               "}",
                               0, 1, 0, 0, // signals
                               QStringList()
                               << "m2:t2:");

    // insert a new resource, but insert is not watched
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m3> a nmo:Message ; "
                               "  nie:title \"t3\" ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               0, 0, 0, 0, // signals
                               QStringList()
                               << "m2:t2:",
                               0); // no updates
    // Note: now nothing should happen (no row removals or insertions), so it's
    // essential that the test function waits a bit in case all signals are 0.

    QTest::newRow("WatchOnlyDeletes")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_rerunForInserts()
{
    QString query = "select tracker:id(?m) ?m "
        "{ ?m a nmo:Message ; "
        "  nie:isLogicalPartOf <tracker-live-tests> . "
        "} order by tracker:id(?m)";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query, 2,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerFullUpdater up;
    up.watchClassForInserts("nmo:Message", QStringList());
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(0, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    updates << new NoOperationUpdate(); // empty initial data set

    // Insert some resources...
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "<m2> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               // note: only one rowsInserted signal, since the
                               // rows are consecutive
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:"
                               << "m2:",
                               0, // no updates
                               1); // but 1 rerun

    // and delete a resource, but delete is not watched
    updates << new UpdateQuery(conn, "delete { "
                               "<m1> a rdfs:Resource . "
                               "}",
                               0, 0, 0, 0, // signals
                               QStringList()
                               << "m1:"
                               << "m2:",
                               0); // no updates or reruns
    // Note: now nothing should happen (no row removals or insertions), so it's
    // essential that the test function waits a bit in case all signals are 0.

    QTest::newRow("RerunForInserts")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_rerunForDeletes()
{
    QString query = "select tracker:id(?m) ?m"
        "{ ?m a nmo:Message ; "
        "  nie:isLogicalPartOf <tracker-live-tests> . "
        "} order by tracker:id(?m)";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query, 2,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerFullUpdater up;
    up.watchClassForDeletes("nmo:Message", QStringList());

    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(0, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m1> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "<m2> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               // note: only 1 rowsInserted signal since the
                               // initial data is inserted in one big bunch
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "m1:"
                               << "m2:",
                               0); // no of update queries (initial query doesn't count)

    // delete a resource, this should be caught normally.
    updates << new UpdateQuery(conn,
                               "delete { "
                               "<m1> a rdfs:Resource . "
                               "}",
                               0, 1, 0, 0, // signals
                               QStringList()
                               << "m2:",
                               0, // no updates
                               1); // but 1 rerun

    // insert a new resource, but insert is not watched
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<m3> a nmo:Message ; "
                               "  nie:isLogicalPartOf <tracker-live-tests> . "
                               "}",
                               0, 0, 0, 0, // signals
                               QStringList()
                               << "m2:",
                               0); // no updates or reruns
    // Note: now nothing should happen (no row removals or insertions), so it's
    // essential that the test function waits a bit in case all signals are 0.

    QTest::newRow("RerunForDeletes")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_groupedInserts()
{
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", defaultOptions);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    // Initial data set
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<mp1> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 100 . "
                               "<mp2> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 200 . "
                               "<mp3> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 300 . "
                               "}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp3:300:",
                               0); // no update queries for initial data

    // Insert some resources between mp1 and mp2 -> we should get only one
    // rowsInserted
    updates << new UpdateQuery(conn,
                               "insert { "
                               "<mpadded1> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 120 ."
                               "<mpadded2> a nmm:MusicPiece ; "
                               "nie:isLogicalPartOf <tracker-live-tests> ; "
                               "nfo:duration 140 ."
                               "}",
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp1:100:"
                               << "mpadded1:120:"
                               << "mpadded2:140:"
                               << "mp2:200:"
                               << "mp3:300:");

    QTest::newRow("GroupedInserts")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_userRowRemoval()
{
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnectionOptions opts;
    opts.setDataReadyInterval(100);
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", opts);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    QString updateQuery = "insert {";
    QString songTemplate = "<mp%1> a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration %2 .";
    for (int i = 0; i < 10; ++i) {
        updateQuery.append(songTemplate.arg(i).arg(i*100));
    }
    updateQuery.append(" }");

    // Initial data set contains 10 rows
    updates << new UpdateQuery(conn,
                               updateQuery,
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp3:300:"
                               << "mp4:400:"
                               << "mp5:500:"
                               << "mp6:600:"
                               << "mp7:700:"
                               << "mp8:800:"
                               << "mp9:900:",
                               0); // update query count: initial query doesn't count

    // Then the user removes 3 rows in the middle
    updates << new UserRowRemoval(liveQuery->model(),
                                  4, 3, // which rows to remove
                                  0, 1, 0, 0, // signals
                                  QStringList()
                                  << "mp0:0:"
                                  << "mp1:100:"
                                  << "mp2:200:"
                                  << "mp3:300:"
                                  << "mp7:700:"
                                  << "mp8:800:"
                                  << "mp9:900:",
                                  0); // no update queries ran

    // A change to an existing row
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp7> nfo:duration 700 . "
                               "} "
                               "insert {"
                               "<mp7> nfo:duration 250 ."
                               "}",
                               0, 0, 1, 1,
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp8:800:"
                               << "mp9:900:");

    // A change to a user-removed row
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp4> nfo:duration 400 . "
                               "} "
                               "insert {"
                               "<mp4> nfo:duration 950 ."
                               "}",
                               0, 0, 0, 0, // nothing happens
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp8:800:"
                               << "mp9:900:",
                               1); // note: an update query is ran; we don't
                                   // care whether the resource is in the model
                                   // or not when deciding whether to run an
                                   // update query.

    // User-removed row really removed
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp5> a rdfs:Resource . "
                               "} ",
                               0, 0, 0, 0, // nothing happens
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp8:800:"
                               << "mp9:900:",
                               1); // note: an update query is ran; we don't
                                   // care whether the resource is in the model
                                   // or not when deciding whether to run an
                                   // update query.

    // Existing row really removed
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp8> a rdfs:Resource . "
                               "} ",
                               0, 1, 0, 0,
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp9:900:");

    // A new rows is inserted in the middle of the user-removed range
    updates << new UpdateQuery(conn,
                               songTemplate.arg(10).arg(400).prepend("insert {").append("}"),
                               1, 0, 0, 0,
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp10:400:"
                               << "mp9:900:");

    // All rows get user-removed
    updates << new UserRowRemoval(liveQuery->model(),
                                  0, 7, // start = 0, count = 7
                                  0, 1, 0, 0,
                                  QStringList(),
                                  0); // no update queries

    // And an update to a user-removed row comes
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp2> nfo:duration 200 . "
                               "} "
                               "insert {"
                               "<mp2> nfo:duration 550 ."
                               "}",
                               0, 0, 0, 0, // nothing happens
                               QStringList(),
                               1); // note: an update query is ran; we don't
                                   // care whether the resource is in the model
                                   // or not when deciding whether to run an
                                   // update query.

    // A new row is inserted
    updates << new UpdateQuery(conn,
                               songTemplate.arg(11).arg(330).prepend("insert {").append("}"),
                               1, 0, 0, 0,
                               QStringList()
                               << "mp11:330:");

    QTest::newRow("UserRowRemoval")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

void tst_TrackerLiveDataDriven::test_data_userRowRemovalWithoutAffectedRows()
{
    // This test simulates a situation where we cannot compute the affected rows
    // properly.  In real life, this would happen e.g., in aggregate cases.
    // Because the affected rows cannot be computed, nothing is ever removed in
    // the model, but updates should target the correct rows and no duplicate
    // rows should be inserted.
    QString query = "select tracker:id(?mp) ?mp ?l "
        "{ ?mp a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration ?l . %1 }"
        " order by ?l";
    QSparqlConnectionOptions opts;
    opts.setDataReadyInterval(100);
    QSparqlConnection* conn = new QSparqlConnection("QTRACKER_DIRECT", opts);
    TrackerLiveQuery* liveQuery = new TrackerLiveQuery(query.arg(""), 3,
                                                       QList<int>() << 0,
                                                       *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));

    // We simulate the failing affected rows computation by giving column index
    // -1 here.
    up.watchClass("nmm:MusicPiece",
                  QStringList() << "nfo:duration" << "rdf:type",
                  "tracker:id(?mp) in %LIST",
                  TrackerPartialUpdater::Subject, -1);
    liveQuery->addUpdater(up);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::Int));

    // The updates we run, and the expected results
    QList<UpdateAndResult*> updates;

    QString updateQuery = "insert {";
    QString songTemplate = "<mp%1> a nmm:MusicPiece ; "
        "nie:isLogicalPartOf <tracker-live-tests> ; "
        "nfo:duration %2 .";
    for (int i = 0; i < 10; ++i) {
        updateQuery.append(songTemplate.arg(i).arg(i*100));
    }
    updateQuery.append(" }");

    // Initial data set contains 10 rows
    updates << new UpdateQuery(conn,
                               updateQuery,
                               1, 0, 0, 0, // signals
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp3:300:"
                               << "mp4:400:"
                               << "mp5:500:"
                               << "mp6:600:"
                               << "mp7:700:"
                               << "mp8:800:"
                               << "mp9:900:",
                               0); // update query count: initial query doesn't count

    // Then the user removes 3 rows in the middle
    updates << new UserRowRemoval(liveQuery->model(),
                                  4, 3, // which rows to remove
                                  0, 1, 0, 0, // signals
                                  QStringList()
                                  << "mp0:0:"
                                  << "mp1:100:"
                                  << "mp2:200:"
                                  << "mp3:300:"
                                  << "mp7:700:"
                                  << "mp8:800:"
                                  << "mp9:900:",
                                  0); // no update queries ran

    // A change to an existing row; this looks like a new row since affected
    // rows cannot be computed. But the "don't insert duplicate rows" safeguard
    // makes the update target the correct row.
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp7> nfo:duration 700 . "
                               "} "
                               "insert {"
                               "<mp7> nfo:duration 250 ."
                               "}",
                               0, 0, 1, 1,
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp8:800:"
                               << "mp9:900:");

    // A change to a user-removed row; again, the safeguard kicks in.
    updates << new UpdateQuery(conn,
                               "delete {"
                               "<mp4> nfo:duration 400 . "
                               "} "
                               "insert {"
                               "<mp4> nfo:duration 950 ."
                               "}",
                               0, 0, 0, 0, // nothing happens
                               QStringList()
                               << "mp0:0:"
                               << "mp1:100:"
                               << "mp2:200:"
                               << "mp7:250:"
                               << "mp3:300:"
                               << "mp8:800:"
                               << "mp9:900:",
                               1); // note: an update query is ran; we don't
                                   // care whether the resource is in the model
                                   // or not when deciding whether to run an
                                   // update query.

    QTest::newRow("UserRowRemovalWithoutAffectedRows")
        <<(void*)conn
        <<(void*)liveQuery
        << (QList<int>() << 1 << 2) // columns to check
        << updates;
}

QTEST_MAIN( tst_TrackerLiveDataDriven )
#include "tst_live_datadriven.moc"
