/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "merger.h"
#include "usermodifiablemodel.h"

#include <QtTest/QtTest>
#include <QStandardItemModel>

Q_DECLARE_METATYPE(QModelIndex)

class tst_Merger : public QObject
{
    Q_OBJECT

public:
    tst_Merger();
    virtual ~tst_Merger();

public slots:
    void init();
    void cleanup();

private slots:
    void addEmptyToEmpty();
    void addOneRowToEmpty();
    void addTwoRowsToEmpty();

    void addEmptyToOneRow();
    void addOneRowToOneRow();

    void updateOneRowWithEmpty();
    void updateOneRowWithOneRow();
    void updateOneRowWithTwoRows();

    void updateAndRelocateForward();
    void updateAndRelocateBackward();

    void orderingInt();
    void orderingIntDesc();

private:
    void fillModel(UserModifiableModel* model,
                   int idStart,
                   const QVariantList& data);

    bool verifyModel(const QAbstractItemModel* model,
                     int column,
                     const QVariantList& data);

    QList<int> idCols;
    QList<TrackerLiveQuery::CollationColumn> sortCols;

    QLinkedList<IdentityList> emptyAffectedRows;
};

tst_Merger::tst_Merger()
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
    idCols << 0;
    sortCols << TrackerLiveQuery::CollationColumn(1, QVariant::String);
}

tst_Merger::~tst_Merger()
{
    cleanup();
}

void tst_Merger::init()
{
}

void tst_Merger::cleanup()
{
}

void tst_Merger::fillModel(UserModifiableModel* model,
                           int idStart,
                           const QVariantList& data)
{
    if (model->rowCount() != 0) {
        qFatal("Test error: trying to fill a non-empty model");
    }
    // Construct a model where the first column contains id's (starting from
    // idStart) and the second column contains the data.
    model->addColumns(0, 2);
    model->setIdentityColumns(QList<int>() << 0);
    model->beginLazyRowInsertions();
    for (int r = 0; r < data.size(); ++r) {
        model->lazyInsertRow(0,
                              QVariantList()
                              << QVariant(idStart + r)
                              << data.at(r));
    }
    model->endLazyRowInsertions();
}

bool tst_Merger::verifyModel(const QAbstractItemModel* model, int col,
                             const QVariantList& data)
{
    bool res = true;
    if (model->rowCount() != data.size())
        res = false;
    for (int r = 0; res && r < model->rowCount(); ++r) {
        if (model->data(model->index(r, col)) != data.at(r))
            res = false;
    }
    if (!res) {
        qDebug() << "--- Expected:";
        qDebug() << data;
        qDebug() << "--- Got:";
        for (int r = 0; r < model->rowCount(); ++r)
            for (int c = 0; c < model->columnCount(); ++c)
                qDebug() << "(" << r << "," << c << ")" <<
                    model->data(model->index(r, c)).toString();
    }

    return res;
}

void tst_Merger::addEmptyToEmpty()
{
    UserModifiableModel to;
    to.addColumns(0, 2);
    UserModifiableModel from;
    from.addColumns(0, 2);

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()));
}

void tst_Merger::addOneRowToEmpty()
{
    UserModifiableModel to;
    to.addColumns(0, 2);
    UserModifiableModel from;
    fillModel(&from, 100, QVariantList() << QVariant("aaa"));

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList() << QVariant("aaa")));
}

void tst_Merger::addTwoRowsToEmpty()
{
    UserModifiableModel to;
    to.addColumns(0, 2);
    UserModifiableModel from;
    // A new constraint: the order of the rows must be the same in the "new
    // data" and in the merged result.
    fillModel(&from, 100, QVariantList() << QVariant("aaa") << QVariant("bbb"));

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("aaa")
                        << QVariant("bbb")));
}

void tst_Merger::addEmptyToOneRow()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList() << QVariant("bbb"));

    UserModifiableModel from;
    from.addColumns(0, 2);

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("bbb")));
}

void tst_Merger::addOneRowToOneRow()
{
    UserModifiableModel to1;
    fillModel(&to1, 0, QVariantList() << QVariant("aaa"));

    UserModifiableModel to2;
    fillModel(&to2, 0, QVariantList() << QVariant("bbb"));

    UserModifiableModel to3;
    fillModel(&to3, 0, QVariantList() << QVariant("ccc"));

    UserModifiableModel from;
    fillModel(&from, 100, QVariantList() << QVariant("bbb"));

    mergeResultsToModel(to1,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    mergeResultsToModel(to2,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    mergeResultsToModel(to3,
                        emptyAffectedRows,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to1.columnCount(), 2);
    QVERIFY(verifyModel(&to1, 1, QVariantList()
                        << QVariant("aaa")
                        << QVariant("bbb")));

    QCOMPARE(to2.columnCount(), 2);
    QVERIFY(verifyModel(&to2, 1, QVariantList()
                        << QVariant("bbb")
                        << QVariant("bbb")));

    QCOMPARE(to3.columnCount(), 2);
    QVERIFY(verifyModel(&to3, 1, QVariantList()
                        << QVariant("bbb")
                        << QVariant("ccc")));
}

void tst_Merger::updateOneRowWithEmpty()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList() << QVariant("aaa"));

    UserModifiableModel from;
    from.addColumns(0, 2);

    QLinkedList<IdentityList> affected;
    affected.append(IdentityList::fromRow(to, 0, idCols));

    mergeResultsToModel(to, affected,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()));
}

void tst_Merger::updateOneRowWithOneRow()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList() << QVariant("aaa"));

    UserModifiableModel from;
    // Note: filling the model with a row with the same identity
    fillModel(&from, 0, QVariantList() << QVariant("bbb"));

    QLinkedList<IdentityList> affected;
    affected.append(IdentityList::fromRow(to, 0, idCols));

    mergeResultsToModel(to, affected,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("bbb")));
}

void tst_Merger::updateOneRowWithTwoRows()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList() << QVariant("aaa"));

    UserModifiableModel from;
    // Note: the row with ccc has the same identity as the row in the model
    fillModel(&from, -1, QVariantList()
              << QVariant("ddd")
              << QVariant("ccc"));

    QLinkedList<IdentityList> affected;
    affected.append(IdentityList::fromRow(to, 0, idCols));

    mergeResultsToModel(to, affected,
                        from, idCols, sortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("ccc")
                        << QVariant("ddd")));
}

void tst_Merger::updateAndRelocateForward()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList()
              << QVariant("bbb")
              << QVariant("ccc")
              << QVariant("ddd"));

    UserModifiableModel from;
    // Note: filling the model with a row with the same identity
    fillModel(&from, 1, QVariantList() << QVariant("eee"));

    QLinkedList<IdentityList> affected;
    affected.append(IdentityList::fromRow(to, 1, idCols));

    mergeResultsToModel(to, affected,
                        from, idCols, sortCols);

    // Note: check here that we got the "moved" signal instead of "removed" +
    // "inserted"

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("bbb")
                        << QVariant("ddd")
                        << QVariant("eee")));
}

void tst_Merger::updateAndRelocateBackward()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList()
              << QVariant("bbb")
              << QVariant("ccc")
              << QVariant("ddd"));

    UserModifiableModel from;
    // Note: filling the model with a row with the same identity
    fillModel(&from, 1, QVariantList() << QVariant("aaa"));

    QLinkedList<IdentityList> affected;
    affected.append(IdentityList::fromRow(to, 1, idCols));

    mergeResultsToModel(to, affected,
                        from, idCols, sortCols);

    // Note: check here that we got the "moved" signal instead of "removed" +
    // "inserted"

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant("aaa")
                        << QVariant("bbb")
                        << QVariant("ddd")));
}

void tst_Merger::orderingInt()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList()
              << QVariant(1)
              << QVariant(10));

    UserModifiableModel from;
    // The rows in from must be in a correct order among themselves
    fillModel(&from, 100, QVariantList()
              << QVariant(2)
              << QVariant(100));

    QList<TrackerLiveQuery::CollationColumn> intSortCols;
    intSortCols << TrackerLiveQuery::CollationColumn(1, QVariant::Int);

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, intSortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant(1)
                        << QVariant(2)
                        << QVariant(10)
                        << QVariant(100)));
}

void tst_Merger::orderingIntDesc()
{
    UserModifiableModel to;
    fillModel(&to, 0, QVariantList()
              << QVariant(10)
              << QVariant(1));
    UserModifiableModel from;
    // The rows in from must be in a correct order among themselves
    fillModel(&from, 100, QVariantList()
              << QVariant(100)
              << QVariant(2));

    QList<TrackerLiveQuery::CollationColumn> intSortCols;
    intSortCols << TrackerLiveQuery::CollationColumn(1, QVariant::Int, Qt::DescendingOrder);

    mergeResultsToModel(to,
                        emptyAffectedRows,
                        from, idCols, intSortCols);

    // Verify the data
    QCOMPARE(to.columnCount(), 2);
    QVERIFY(verifyModel(&to, 1, QVariantList()
                        << QVariant(100)
                        << QVariant(10)
                        << QVariant(2)
                        << QVariant(1)));
}

QTEST_MAIN( tst_Merger )
#include "tst_live_merger.moc"
