/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "movablerowmodel.h"

#include <QtTest/QtTest>
#include <QPersistentModelIndex>
#include <QStandardItemModel>
#include "../utils/signalsrecorder.h"
#include "../utils/modelsignalrecorder.h"

class ModelSignalReceiver;
Q_DECLARE_METATYPE(QModelIndex)

class tst_MovableRowModel : public QObject
{
    Q_OBJECT

public:
    tst_MovableRowModel();
    virtual ~tst_MovableRowModel();

public slots:
    void init();
    void cleanup();

private slots:
    void construct();
    void addColumn();
    void addColumn_invalid_column_index();
    void addColumns();
    void addColumns_invalid_column_index();
    void addColumns_invalid_amount();
    void addRow();
    void addRowWithData();
    void addRow_invalid_row_index();
    void addRows();
    void addRows_invalid_row_index();
    void addRows_invalid_amount();
    void index();
    void setValue();
    void setValue_invalid_index();
    void deleteRow();
    void deleteRows();
    void moveRow_empty();
    void moveRow_single();
    void moveRow_two();
    void moveRow_persistentIndex();
    void persistentIndex();
    void checkModelIsReadOnlyThroughQaim();
    void batchAddRows();
    void batchAddRowsWithData();
    void setColumnHeaders();
    void setRoleNames();
    void verifyQmlData();

private:
    // Object under test
    MovableRowModel* model;
    // Test helpers
    ModelSignalRecorder* expectedSignalRecorder;
    ModelSignalRecorder* unexpectedSignalRecorder;
};

tst_MovableRowModel::tst_MovableRowModel()
    : model(0), expectedSignalRecorder(0), unexpectedSignalRecorder(0)
{
}

tst_MovableRowModel::~tst_MovableRowModel()
{
    cleanup();
}

void tst_MovableRowModel::init()
{
   model = new MovableRowModel;
   expectedSignalRecorder = new ModelSignalRecorder(model);
   unexpectedSignalRecorder = new ModelSignalRecorder(model);
}

void tst_MovableRowModel::cleanup()
{
    delete unexpectedSignalRecorder;
    unexpectedSignalRecorder = 0;
    delete expectedSignalRecorder;
    expectedSignalRecorder = 0;
    delete model;
    model = 0;
}

void tst_MovableRowModel::construct()
{
    QCOMPARE( model->rowCount(), 0 );
    QCOMPARE( model->columnCount(), 0 );
}

void tst_MovableRowModel::addColumn()
{
    expectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( model->addColumn(0) );
    QCOMPARE( model->columnCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    expectedSignalRecorder->clearAllSignalData();
    QVERIFY( model->addColumn(1) );
    QCOMPARE( model->columnCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 1, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    expectedSignalRecorder->clearAllSignalData();
    QVERIFY( model->addColumn(0) );
    QCOMPARE( model->columnCount(), 3 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addColumn_invalid_column_index()
{
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addColumn(-1) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addColumn(1) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->addColumn(0) );
    QCOMPARE( model->columnCount(), 1 );
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->addColumn(-1) );
    QCOMPARE( model->columnCount(), 1 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addColumn(2) );
    QCOMPARE( model->columnCount(), 1 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addColumns()
{
    expectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( model->addColumns(0, 2) );
    QCOMPARE( model->columnCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 0, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    expectedSignalRecorder->clearAllSignalData();
    QVERIFY( model->addColumns(1, 3) );
    QCOMPARE( model->columnCount(), 5 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 1, 3) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->addColumns(5, 2) );
    QCOMPARE( model->columnCount(), 7 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 5, 6) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    expectedSignalRecorder->clearAllSignalData();
    QVERIFY( model->addColumns(0, 2) );
    QCOMPARE( model->columnCount(), 9 );
    QVERIFY( expectedSignalRecorder->verifyInsertColumnSignalPairReceived(QModelIndex(), 0, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addColumns_invalid_column_index()
{
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addColumns(-1, 2) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addColumns(1, 2) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->addColumns(0, 2) );
    QCOMPARE( model->columnCount(), 2 );
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->addColumns(-1, 2) );
    QCOMPARE( model->columnCount(), 2 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addColumns(3, 2) );
    QCOMPARE( model->columnCount(), 2 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addColumns_invalid_amount()
{
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addColumns(0, 0) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addColumns(1, -1) );
    QCOMPARE( model->columnCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addRow()
{
    expectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( model->addRow(0) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->addRow(1) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 1, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->addRow(0) );
    QCOMPARE( model->rowCount(), 3 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addRowWithData()
{
    QVERIFY(model->addColumns(0, 2));

    expectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QStandardItemModel sourceModel;
    sourceModel.insertColumns(0, 2);
    sourceModel.insertRows(0, 3);
    sourceModel.setData(sourceModel.index(0, 0), 0);
    sourceModel.setData(sourceModel.index(0, 1), 1);
    sourceModel.setData(sourceModel.index(1, 0), 2);
    sourceModel.setData(sourceModel.index(1, 1), 3);
    sourceModel.setData(sourceModel.index(2, 0), 4);
    sourceModel.setData(sourceModel.index(2, 1), 5);

    // note: adding the rows in a different order
    QVERIFY(model->addRowWithData(0, sourceModel, 1));
    QCOMPARE(model->rowCount(), 1);
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE(model->data(model->index(0, 0)), QVariant(2));
    QCOMPARE(model->data(model->index(0, 1)), QVariant(3));
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY(model->addRowWithData(1, sourceModel, 2));
    QCOMPARE(model->rowCount(), 2);
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 1, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE(model->data(model->index(1, 0)), QVariant(4));
    QCOMPARE(model->data(model->index(1, 1)), QVariant(5));
    expectedSignalRecorder->clearAllSignalData();

    // insert between the existing rows
    QVERIFY(model->addRowWithData(1, sourceModel, 0));
    QCOMPARE(model->rowCount(), 3);
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 1, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE(model->data(model->index(1, 0)), QVariant(0));
    QCOMPARE(model->data(model->index(1, 1)), QVariant(1));
    QCOMPARE(model->data(model->index(2, 0)), QVariant(4));
    QCOMPARE(model->data(model->index(2, 1)), QVariant(5));
    expectedSignalRecorder->clearAllSignalData();

    // insert again between the existing rows
    // note: using the different addRowWithData here.
    model->addRowWithData(2, QVariantList() << QVariant(6) << QVariant(7));
    QCOMPARE(model->rowCount(), 4);
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 2, 2) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE(model->data(model->index(2, 0)), QVariant(6));
    QCOMPARE(model->data(model->index(2, 1)), QVariant(7));
    QCOMPARE(model->data(model->index(3, 0)), QVariant(4));
    QCOMPARE(model->data(model->index(3, 1)), QVariant(5));
}

void tst_MovableRowModel::addRow_invalid_row_index()
{
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addRow(-1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addRow(1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->addRow(0) );
    QCOMPARE( model->rowCount(), 1 );
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->addRow(-1) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addRow(2) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addRows()
{
    expectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( model->addRows(0, 2) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 0, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->addRows(1, 3) );
    QCOMPARE( model->rowCount(), 5 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 1, 3) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->addRows(0, 4) );
    QCOMPARE( model->rowCount(), 9 );
    QVERIFY( expectedSignalRecorder->verifyInsertRowSignalPairReceived(QModelIndex(), 0, 3) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addRows_invalid_row_index()
{
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addRows(1, 2) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addRows(-1, 3) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::addRows_invalid_amount()
{
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->addRows(0, 0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !model->addRows(0, -1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::index()
{
    // Invalid index coordinates
    QVERIFY( !model->index(-1, 0).isValid() );
    QVERIFY( !model->index(0, -1).isValid() );
    QVERIFY( !model->index(-1, -1).isValid() );

    // Empty model does not provide indeces
    QVERIFY( !model->index(0, 0).isValid() );

    // Model without columns does not provide indeces
    model->addRow(0);
    QVERIFY( !model->index(0, 0).isValid() );

    // A 1x1 model
    model->addColumn(0);
    QModelIndex index = model->index(0, 0);
    QVERIFY( index.isValid() );
    QCOMPARE( index.row(), 0 );
    QCOMPARE( index.column(), 0 );
    QVERIFY( !model->index(1, 0).isValid() );
    QVERIFY( !model->index(0, 1).isValid() );

    // Add rows and test indeces for all
    model->addRows(1, 3);
    for(int row = 0; row < model->rowCount(); ++row)
    {
        QModelIndex index = model->index(row, 0);
        QVERIFY( index.isValid() );
        QCOMPARE( index.row(), row );
        QCOMPARE( index.column(), 0 );
    }
    QVERIFY( !model->index(model->rowCount(), 0).isValid() );

    // Add more columns and test indeces for all
    model->addColumn(1);
    for(int row = 0; row < model->rowCount(); ++row)
    {
        for(int col=0; col < model->columnCount(); ++col)
        {
            QModelIndex index = model->index(row, col);
            QVERIFY( index.isValid() );
            QCOMPARE( index.row(), row );
            QCOMPARE( index.column(), col );
        }
    }

    // Test out of bounds indeces
    QVERIFY( !model->index(-1, 0).isValid() );
    QVERIFY( !model->index(0, -1).isValid() );
    QVERIFY( !model->index(-1, -1).isValid() );
    QVERIFY( !model->index(model->rowCount()  , model->columnCount()-1).isValid() );
    QVERIFY( !model->index(model->rowCount()-1, model->columnCount()  ).isValid() );
    QVERIFY( !model->index(model->rowCount(),   model->columnCount()  ).isValid() );
}

void tst_MovableRowModel::setValue()
{
    expectedSignalRecorder->connectDataChangedSignal();

    model->addColumn(0);
    model->addRow(0);

    const QModelIndex index = model->index(0, 0);
    const QVariant testValue(QPoint(index.row(), index.column()));
    QVERIFY( model->setValue(index, testValue) );
    QVERIFY( expectedSignalRecorder->verifyDataChangedSignalReceived(index, index) );
    expectedSignalRecorder->clearAllSignalData();

    QCOMPARE( model->data(index), testValue );
    // Accessing the data should not emit dataChanged
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->setValue(index, testValue) );
    // Re-setting the same value shoud not emit dataChanged signal
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE( model->data(index), testValue );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::setValue_invalid_index()
{
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->setValue(model->index(0,0), QVariant()) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::deleteRow()
{
    expectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->deleteRow(-1) );
    QVERIFY( !model->deleteRow(0) );
    QVERIFY( !model->deleteRow(1) );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addRow(0);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRow(1) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addRows(0, 2);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRow(2) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->deleteRow(1) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 1, 1) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addRows(0, 2);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRow(0) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::deleteRows()
{
    expectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->deleteRows(-1, 1) );
    QVERIFY( !model->deleteRows(0, 0) );
    QVERIFY( !model->deleteRows(0, 1) );
    QVERIFY( !model->deleteRows(1, 1) );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addRow(0);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRows(1, 1) );
    QVERIFY( !model->deleteRows(0, 2) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->deleteRows(0, 1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRows(0, 1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    model->addRows(0, 3);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRows(2, 2) );
    QCOMPARE( model->rowCount(), 3 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( model->deleteRows(1, 2) );
    QCOMPARE( model->rowCount(), 1 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 1, 2) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->deleteRows(0, 1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 0) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->deleteRows(0, 1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addRows(0, 5);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( model->deleteRows(0, 5) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyRemoveRowSignalPairReceived(QModelIndex(), 0, 4) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::moveRow_empty()
{
    expectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    // Cannot move rows in an empty model
    QVERIFY( !model->moveRow( 0,  0) );
    QVERIFY( !model->moveRow( 0,  1) );
    QCOMPARE( model->rowCount(), 0 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::moveRow_single()
{
    // Cannot move a single row
    model->addColumn(0);
    model->addRow(0);
    model->setValue(model->index(0,0), QVariant(QPoint(0,0)));

    expectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->moveRow( 0,  0) );
    QVERIFY( !model->moveRow( 0,  1) );
    QVERIFY( !model->moveRow( 1,  0) );
    QCOMPARE( model->rowCount(), 1 );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(0,0)) );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::moveRow_two()
{
    model->addColumn(0);
    model->addRows(0, 2);
    model->setValue(model->index(0,0), QVariant(QPoint(0,0)));
    model->setValue(model->index(1,0), QVariant(QPoint(1,0)));

    expectedSignalRecorder->connectMoveRowSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !model->moveRow( 0,  0) );
    QVERIFY( !model->moveRow( 0,  1) );
    QCOMPARE( model->rowCount(), 2 );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(0,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(1,0)) );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY(  model->moveRow( 0,  2) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyMoveRowSignalPairReceived(QModelIndex(), 0, 0, QModelIndex(), 2) );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(1,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(0,0)) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->moveRow( 0,  3) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(1,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(0,0)) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY(  model->moveRow( 1,  0) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyMoveRowSignalPairReceived(QModelIndex(), 1, 1, QModelIndex(), 0) );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(0,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(1,0)) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
    expectedSignalRecorder->clearAllSignalData();

    QVERIFY( !model->moveRow( 1,  1) );
    QVERIFY( !model->moveRow( 1,  2) );
    QVERIFY( !model->moveRow( 2,  0) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(0,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(1,0)) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    // Move rejects negative source or target row
    QVERIFY( !model->moveRow(-1,  0) );
    QVERIFY( !model->moveRow( 0, -1) );
    QVERIFY( !model->moveRow(-1, -1) );
    QCOMPARE( model->rowCount(), 2 );
    QVERIFY( expectedSignalRecorder->verifyNoSignalsReceived() );
    QCOMPARE( model->data(model->index(0,0)), QVariant(QPoint(0,0)) );
    QCOMPARE( model->data(model->index(1,0)), QVariant(QPoint(1,0)) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::moveRow_persistentIndex()
{
    model->addColumn(0);
    model->addRows(0, 2);
    const QVariant val00 = QVariant(QPoint(0,0));
    const QVariant val10 = QVariant(QPoint(1,0));
    model->setValue(model->index(0,0), val00);
    model->setValue(model->index(1,0), val10);
    QPersistentModelIndex pidx00 = model->index(0,0);
    QPersistentModelIndex pidx10 = model->index(1,0);

    QVERIFY( !model->moveRow( 0,  0) );
    QVERIFY( !model->moveRow( 0,  1) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );

    QVERIFY(  model->moveRow( 0,  2) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );

    QVERIFY( !model->moveRow( 0,  3) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );

    QVERIFY(  model->moveRow( 1,  0) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );

    QVERIFY( !model->moveRow( 1,  1) );
    QVERIFY( !model->moveRow( 1,  2) );
    QVERIFY( !model->moveRow( 2,  0) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );

    QVERIFY( !model->moveRow(-1,  0) );
    QVERIFY( !model->moveRow( 0, -1) );
    QVERIFY( !model->moveRow(-1, -1) );
    QVERIFY( pidx00.isValid() );
    QVERIFY( pidx10.isValid() );
    QCOMPARE( model->data(pidx00), val00 );
    QCOMPARE( model->data(pidx10), val10 );
}

void tst_MovableRowModel::persistentIndex()
{
    QModelIndex idx = model->index(0, 0);
    QPersistentModelIndex pidx(idx);
    QVERIFY( !idx.isValid() );
    QVERIFY( !pidx.isValid() );
    QVERIFY( idx == pidx );

    model->addRow(0);
    QVERIFY( !pidx.isValid() );
    model->addColumn(0);
    // Persistent index to previously invalid item does not become valid when the item is added
    QVERIFY( !pidx.isValid() );

    idx = model->index(0, 0);
    const QVariant testVal(QString("pidx"));
    model->setValue(idx, testVal);
    pidx = idx;
    QVERIFY( pidx.isValid() );
    QVERIFY( idx == pidx );
    QCOMPARE( model->data(pidx), testVal );

    model->addRow(1);
    QVERIFY( pidx == model->index(0,0) );
    QCOMPARE( model->data(pidx), testVal );

    model->addRow(0);
    QVERIFY( pidx == model->index(1, 0) );
    QCOMPARE( model->data(pidx), testVal );

    model->addColumn(1);
    QVERIFY( pidx == model->index(1, 0) );
    QCOMPARE( model->data(pidx), testVal );

    model->addColumn(0);
    QVERIFY( pidx == model->index(1, 1) );
    QCOMPARE( model->data(pidx), testVal );

    model->deleteRow(1);
    QVERIFY( !pidx.isValid() );
    model->addRow(1);
    pidx = model->index(1, 1);
    model->setValue(pidx, testVal);

    model->deleteRow(2);
    QVERIFY( pidx == model->index(1, 1) );
    QCOMPARE( model->data(pidx), testVal );

    model->deleteRow(0);
    QVERIFY( pidx == model->index(0, 1) );
    QCOMPARE( model->data(pidx), testVal );

    model->deleteRow(0);
    QVERIFY( !pidx.isValid() );
}

void tst_MovableRowModel::checkModelIsReadOnlyThroughQaim()
{
    QAbstractItemModel* modelAsQaim = model;
    unexpectedSignalRecorder->connectInsertColumnSignals();
    unexpectedSignalRecorder->connectInsertRowSignals();
    unexpectedSignalRecorder->connectRemoveColumnSignals();
    unexpectedSignalRecorder->connectRemoveRowSignals();
    unexpectedSignalRecorder->connectDataChangedSignal();

    QVERIFY( !modelAsQaim->insertColumn(0) );
    QVERIFY( !modelAsQaim->insertColumns(0, 2) );

    QVERIFY( !modelAsQaim->insertRow(0) );
    QVERIFY( !modelAsQaim->insertRows(0, 2) );

    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    model->addColumns(0, 2);
    model->addRows(0, 2);
    unexpectedSignalRecorder->clearAllSignalData();

    QVERIFY( !modelAsQaim->setData(model->index(0,0), QVariant("test")) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !modelAsQaim->removeRow(0) );
    QVERIFY( !modelAsQaim->removeRows(0, 2) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );

    QVERIFY( !modelAsQaim->removeColumn(0) );
    QVERIFY( !modelAsQaim->removeColumns(0, 2) );
    QVERIFY( unexpectedSignalRecorder->verifyNoSignalsReceived() );
}

void tst_MovableRowModel::batchAddRows()
{
    QSignalSpy aboutSpy(model,
                        SIGNAL(rowsAboutToBeInserted(const QModelIndex&,
                                                     int, int)));

    QSignalSpy insertSpy(model,
                         SIGNAL(rowsInserted(const QModelIndex&,
                                             int, int)));

    // Add 2 rows in the batch mode
    model->beginAddRowsBatch(0, 2);

    QVERIFY(model->addRow(0));
    QCOMPARE(model->rowCount(), 1);

    // We have only received rowsAboutToBeInserted
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 0);

    QVERIFY(model->addRow(1));
    QCOMPARE(model->rowCount(), 2);

    // Still only rowsAboutToBeInserted
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 0);

    model->endAddRowsBatch();

    // Now we have received rowsInserted with both rows
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 1);

    // Do another batch operation which adds a row between the existing rows
    model->beginAddRowsBatch(1, 1);

    QCOMPARE(aboutSpy.count(), 2);
    QCOMPARE(insertSpy.count(), 1);

    QVERIFY(model->addRow(1));
    QCOMPARE(model->rowCount(), 3);

    model->endAddRowsBatch();

    QCOMPARE(aboutSpy.count(), 2);
    QCOMPARE(insertSpy.count(), 2);
}

void tst_MovableRowModel::batchAddRowsWithData()
{
    model->addColumns(0, 1);
    QSignalSpy aboutSpy(model,
                        SIGNAL(rowsAboutToBeInserted(const QModelIndex&,
                                                     int, int)));

    QSignalSpy insertSpy(model,
                         SIGNAL(rowsInserted(const QModelIndex&,
                                             int, int)));

    // Add 2 rows in the batch mode
    model->beginAddRowsBatch(0, 2);

    QVERIFY(model->addRowWithData(0, QVariantList() << "first"));
    QCOMPARE(model->rowCount(), 1);

    // We have only received rowsAboutToBeInserted
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 0);

    QVERIFY(model->addRowWithData(1, QVariantList() << "second"));
    QCOMPARE(model->rowCount(), 2);

    // Still only rowsAboutToBeInserted
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 0);

    model->endAddRowsBatch();

    // Now we have received rowsInserted with both rows
    QCOMPARE(aboutSpy.count(), 1);
    QCOMPARE(insertSpy.count(), 1);

    // Verify the data
    QCOMPARE(model->data(model->index(0, 0)), QVariant("first"));
    QCOMPARE(model->data(model->index(1, 0)), QVariant("second"));

    // Do another batch operation which adds a row between the existing rows
    model->beginAddRowsBatch(1, 1);

    QCOMPARE(aboutSpy.count(), 2);
    QCOMPARE(insertSpy.count(), 1);

    QVERIFY(model->addRowWithData(1, QVariantList() << "added"));
    QCOMPARE(model->rowCount(), 3);

    model->endAddRowsBatch();

    QCOMPARE(aboutSpy.count(), 2);
    QCOMPARE(insertSpy.count(), 2);

    // Verify the data
    QCOMPARE(model->data(model->index(0, 0)), QVariant("first"));
    QCOMPARE(model->data(model->index(1, 0)), QVariant("added"));
    QCOMPARE(model->data(model->index(2, 0)), QVariant("second"));
}

void tst_MovableRowModel::setColumnHeaders()
{
    model->addColumns(0, 2);
    model->addRows(0, 2);

    QVERIFY(model->setHeaderData(0, Qt::Horizontal, "firstcol"));
    QVERIFY(model->setHeaderData(1, Qt::Horizontal, "secondcol"));

    QCOMPARE(model->headerData(0, Qt::Horizontal), QVariant("firstcol"));
    QCOMPARE(model->headerData(1, Qt::Horizontal), QVariant("secondcol"));

    QVERIFY(model->setHeaderData(0, Qt::Vertical, "firstrow"));
    QVERIFY(model->setHeaderData(1, Qt::Vertical, "secondrow"));

    QCOMPARE(model->headerData(0, Qt::Vertical), QVariant("firstrow"));
    QCOMPARE(model->headerData(1, Qt::Vertical), QVariant("secondrow"));
}

void tst_MovableRowModel::setRoleNames()
{
    model->addColumns(0,2);
    model->addRows(0,2);
    QVERIFY(model->setHeaderData(0, Qt::Horizontal, "firstrole"));
    QVERIFY(model->setHeaderData(1, Qt::Horizontal, "secondrole"));

    QHash<int, QByteArray> roleNames = model->roleNames();

    QCOMPARE(roleNames.value(Qt::UserRole+1), QByteArray("firstrole"));
    QCOMPARE(roleNames.value(Qt::UserRole+2), QByteArray("secondrole"));

}

void tst_MovableRowModel::verifyQmlData()
{
    model->addColumns(0,2);

    QVERIFY(model->setHeaderData(0, Qt::Horizontal, "firstrole"));
    QVERIFY(model->setHeaderData(1, Qt::Horizontal, "secondrole"));

    QVERIFY(model->addRowWithData(0, QVariantList() << "firstRowData" << "secondFirstRowData"));
    QCOMPARE(model->rowCount(), 1);
    QVERIFY(model->addRowWithData(1, QVariantList() << "secondRowData" << "secondSecondRowData"));
    QCOMPARE(model->rowCount(), 2);

    // Verify the data

    QCOMPARE(model->data(model->index(0, 0)), QVariant("firstRowData"));
    QCOMPARE(model->data(model->index(0, 1)), QVariant("secondFirstRowData"));

    QCOMPARE(model->data(model->index(1, 0)), QVariant("secondRowData"));
    QCOMPARE(model->data(model->index(1, 1)), QVariant("secondSecondRowData"));

    // qml will call the abstractitemmodel we return without any regard to columns
    // so check this behaviour and make sure we get the right data based on the
    // role we pass

    QCOMPARE(model->data(model->index(0,0), Qt::UserRole+1), QVariant("firstRowData"));
    QCOMPARE(model->data(model->index(0,0), Qt::UserRole+2), QVariant("secondFirstRowData"));

    QCOMPARE(model->data(model->index(1,0), Qt::UserRole+1), QVariant("secondRowData"));
    QCOMPARE(model->data(model->index(1,0), Qt::UserRole+2), QVariant("secondSecondRowData"));

}

QTEST_MAIN( tst_MovableRowModel )
#include "tst_movablerowmodel.moc"
