/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "ontology.h"

#include "../utils/signalsrecorder.h"
#include "../utils/messagehandler.h"

#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

class tst_TrackerLiveOntology : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveOntology();
    virtual ~tst_TrackerLiveOntology();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void queryPredicates();
    void queryClasses();
    void queryBoth();

    void queryFailsImmediately();
    void queryFails();

    void invalidClass();

private:
    // object under test
    QSparqlConnection* conn;
    OntologyQuerier* ontologyQuerier;
    SignalList* signalList;
    SignalsRecorder* signalRecorder;
};

tst_TrackerLiveOntology::tst_TrackerLiveOntology()
    : conn(0), ontologyQuerier(0), signalList(0), signalRecorder(0)
{

}

tst_TrackerLiveOntology::~tst_TrackerLiveOntology()
{
}

void tst_TrackerLiveOntology::initTestCase()
{
    conn = new QSparqlConnection("QTRACKER_DIRECT");
}

void tst_TrackerLiveOntology::cleanupTestCase()
{
    delete conn;
    conn = 0;
}

void tst_TrackerLiveOntology::init()
{
    ontologyQuerier = new OntologyQuerier(*conn);
    signalList = new SignalList();
    signalList->append(ontologyQuerier, SIGNAL(finished()));
    signalRecorder = new SignalsRecorder();
    signalRecorder->listen(*signalList);
}

void tst_TrackerLiveOntology::cleanup()
{
    delete signalRecorder;
    signalRecorder = 0;
    delete signalList;
    signalList = 0;

    delete ontologyQuerier;
    ontologyQuerier = 0;
}

void tst_TrackerLiveOntology::queryPredicates()
{
    ontologyQuerier->addPredicate("nie:title");
    ontologyQuerier->addPredicate("nmo:messageSubject");
    ontologyQuerier->start();

    QVERIFY(signalRecorder->waitForSignals(*signalList, 6000));
    QCOMPARE(signalRecorder->signalData(*signalList).count(), 1);

    const QHash<QString, int>& predicateToId = ontologyQuerier->shortPredicateToIdMap();
    QCOMPARE(predicateToId.size(), 2);
    // 0 would also be the default-constructed value if the value is not in the
    // hash
    QVERIFY(predicateToId["nie:title"] > 0);
    QVERIFY(predicateToId["nmo:messageSubject"] > 0);
}

void tst_TrackerLiveOntology::queryClasses()
{
    QStringList shortClasses;
    shortClasses << "nie:InformationElement" << "nmo:Email";
    Q_FOREACH(const QString& shortClass, shortClasses) {
        ontologyQuerier->addClass(shortClass);
    }
    ontologyQuerier->start();

    QVERIFY(signalRecorder->waitForSignals(*signalList, 3000));
    QCOMPARE(signalRecorder->signalData(*signalList).count(), 1);

    QCOMPARE(ontologyQuerier->longClasses().size(), shortClasses.size());

    for (int i = 0; i < 2; ++i) {
        QString longClass = ontologyQuerier->longClasses()[i];
        QString shortClass = ontologyQuerier->longClassToShort(longClass);
        QVERIFY(shortClass.size());
        QVERIFY(shortClasses.contains(shortClass));
        shortClasses.removeAll(shortClass);
    }
    QCOMPARE(shortClasses, QStringList());
}

void tst_TrackerLiveOntology::queryBoth()
{
    // Query both predicates and classes at the same time.
    ontologyQuerier->addPredicate("nie:title");
    ontologyQuerier->addPredicate("nmo:messageSubject");

    QStringList shortClasses;
    shortClasses << "nie:InformationElement" << "nmo:Email";
    Q_FOREACH(const QString& shortClass, shortClasses) {
        ontologyQuerier->addClass(shortClass);
    }
    ontologyQuerier->start();

    QVERIFY(signalRecorder->waitForSignals(*signalList, 3000));
    QCOMPARE(signalRecorder->signalData(*signalList).count(), 1);

    QCOMPARE(ontologyQuerier->longClasses().size(), shortClasses.size());

    for (int i = 0; i < 2; ++i) {
        QString longClass = ontologyQuerier->longClasses()[i];
        QString shortClass = ontologyQuerier->longClassToShort(longClass);
        QVERIFY(shortClasses.contains(shortClass));
        shortClasses.removeAll(shortClass);
    }
    QCOMPARE(shortClasses, QStringList());

    const QHash<QString, int>& predicateToId = ontologyQuerier->shortPredicateToIdMap();
    QCOMPARE(predicateToId.size(), 2);
    // 0 would also be the default-constructed value if the value is not in the
    // hash
    QVERIFY(predicateToId["nie:title"] > 0);
    QVERIFY(predicateToId["nmo:messageSubject"] > 0);
}

void tst_TrackerLiveOntology::queryFailsImmediately()
{
    // Supress warnings
    QtMsgHandler defaultMsgHandler = qInstallMsgHandler(outputWithoutWarnings);

    QSparqlConnection bogusConn("BOGUS");
    // don't use the one created in init()
    OntologyQuerier brokenQuerier(bogusConn);
    brokenQuerier.addClass("nmo:Email");

    // test that we get the finished() signal
    SignalsRecorder signalRecorder2;
    SignalList signalList2;
    signalList2.append(&brokenQuerier, SIGNAL(finished()));
    signalRecorder2.listen(signalList2);

    brokenQuerier.start();

    QCOMPARE(signalRecorder2.signalData(signalList2).count(), 1);

    qInstallMsgHandler(defaultMsgHandler);
}

void tst_TrackerLiveOntology::queryFails()
{
    // Supress warnings
    QtMsgHandler defaultMsgHandler = qInstallMsgHandler(outputWithoutWarnings);

    // give so broken predicate that the query will be an error
    ontologyQuerier->addPredicate("\"FAIL");

    ontologyQuerier->start();

    QVERIFY(signalRecorder->waitForSignals(*signalList, 3000));
    QCOMPARE(signalRecorder->signalData(*signalList).count(), 1);

    qInstallMsgHandler(defaultMsgHandler);
}

void tst_TrackerLiveOntology::invalidClass()
{
    // Supress warnings
    QtMsgHandler defaultMsgHandler = qInstallMsgHandler(outputWithoutWarnings);

    ontologyQuerier->addClass("invalid");
    ontologyQuerier->start();

    QVERIFY(signalRecorder->waitForSignals(*signalList, 3000));
    QCOMPARE(signalRecorder->signalData(*signalList).count(), 1);

    QCOMPARE(ontologyQuerier->longClasses().size(), 0);
    QCOMPARE(ontologyQuerier->shortPredicateToIdMap().size(), 0);

    qInstallMsgHandler(defaultMsgHandler);
}

QTEST_MAIN(tst_TrackerLiveOntology)
#include "tst_live_ontology.moc"
