/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/updatewaiter.h"

#include "TrackerLiveQuery"

#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

Q_DECLARE_METATYPE(QModelIndex)

namespace {
    const char* const benchmarkEnableEnvName = "LIVEQUERY_BENCHMARK";
}

class tst_TrackerLivePausing : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLivePausing();
    virtual ~tst_TrackerLivePausing();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();
    void cleanupTestData();

private slots:
    void initialQueryAndUpdatesSimultaneously();
    void pauseUpdates();
    void tooManyIds();

private:
    QSparqlConnection* conn;
    // object under test
    TrackerLiveQuery* liveQuery;
};

tst_TrackerLivePausing::tst_TrackerLivePausing()
    : conn(0), liveQuery(0)
{
    qRegisterMetaType<QModelIndex>();
}

tst_TrackerLivePausing::~tst_TrackerLivePausing()
{
}

void tst_TrackerLivePausing::initTestCase()
{
    conn = new QSparqlConnection("QTRACKER_DIRECT");
    // Clean up leftover data from previous test run
    cleanup();
    executeQuery(QSparqlQuery("insert { "
                              "<tracker-live-tests> a nie:InformationElement . }",
                              QSparqlQuery::InsertStatement));

    // Enable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "1");
}

void tst_TrackerLivePausing::cleanupTestCase()
{
    // Disable TrackerLiveQuery performance counters
    qputenv(benchmarkEnableEnvName, "");

    delete conn;
    conn = 0;
}

void tst_TrackerLivePausing::init()
{
    QString query = "select tracker:id(?m) ?m ?t"
        "{ ?m a nmo:Message ; "
        "  nie:isLogicalPartOf <tracker-live-tests> ; "
        "  nie:title ?t ."
        "  %1 }"
        " order by ?t";
    // Note: Using the deprecated TrackerLiveQuery constructor here.  Start
    // using the new constructor when the deprecated one is removed.
    liveQuery = new TrackerLiveQuery(QSparqlQuery(query.arg("")), 3, *conn);
    TrackerPartialUpdater up(query.arg("%FILTER"));
    up.watchClass("nmo:Message",
                  QStringList() << "rdf:type" << "nie:title",
                  "tracker:id(?m) in %LIST", TrackerPartialUpdater::Subject, 0);
    liveQuery->addUpdater(up);
    liveQuery->setIdentityColumns(QList<int>() << 0);
    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(2, QVariant::String));
}

void tst_TrackerLivePausing::cleanup()
{
    delete liveQuery;
    liveQuery = 0;
    cleanupTestData();
}

void tst_TrackerLivePausing::cleanupTestData()
{
    executeQuery(QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

namespace {

class Inserter : public QObject
{
    Q_OBJECT
public:
    Inserter() : done(false)
    {
    }
public Q_SLOTS:
    void doInserts()
    {
        QSet<QSparqlResult*> results;
        QSparqlConnection conn("QTRACKER_DIRECT");

        int ix = 0;
        for (int q = 0; q < 100; ++q) {
            QString query = QString("insert { ");
            for (int i = 0; i < 20; ++i) {
                query.append(QString("<m%1> a nmo:Message ; "
                                     "nie:isLogicalPartOf <tracker-live-tests> ; "
                                     "  nie:title \"title%1\" .").arg(ix));
                ++ix;
            }
            query.append(" }");
            results.insert(conn.exec(QSparqlQuery(query, QSparqlQuery::InsertStatement)));

        }
        // make sure we don't delete active queries
        Q_FOREACH(QSparqlResult* r, results) {
            while (!r->isFinished())
                QTest::qWait(100);
            delete r;
        }
        done = true;
    }
public:
    bool done;
};

} // end of unnamed namespace

void tst_TrackerLivePausing::initialQueryAndUpdatesSimultaneously()
{
    // Note: This test is useful for manual debugging but not very useful as an
    // automated test.
    QSKIP("The test case is not for automated testing", SkipSingle);
    // Start a thread that inserts data repeatedly, try to ensure that we get
    // some GraphChanged signals while our initial query is still running. Note:
    // we cannot really verify that it really happened (automatically). It can
    // be verified by e.g., debug prints (LIVEQUERY_DEBUG=1). Also, the test
    // case is racy, so even with buggy code, it doesn't always fail.
    Inserter inserter;
    QThread inserterThread;
    inserter.moveToThread(&inserterThread);
    QObject::connect(&inserterThread, SIGNAL(started()),
                     &inserter, SLOT(doInserts()));

    inserterThread.start();
    QTest::qWait(600); // correct delay depends on the environment
    liveQuery->start();

    while (!inserter.done)
        QTest::qWait(100);

    QAbstractItemModel* model = liveQuery->model();
    // wait until we receive the rest of the signals
    QTime timer;
    timer.start();
    while (model->rowCount() < 2000 && timer.elapsed() < 10000)
        QTest::qWait(100);

    //printModel(liveQuery->model());
    QCOMPARE(model->rowCount(), 2000);

    inserterThread.quit();
    inserterThread.wait();

    UpdateWaiter waiterMsg(QString("http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#Message"), 0, 2000);
    cleanupTestData();
    QVERIFY( waiterMsg.waitForAllUpdated() );
}

void tst_TrackerLivePausing::pauseUpdates()
{
    // Insert some data
    UpdateWaiter waiterMsg(QString("http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#Message"), 12);
    executeQuery(QSparqlQuery("insert { "
                              "<m01> a nmo:Message ; "
                              "nie:isLogicalPartOf <tracker-live-tests> ; "
                              "  nie:title \"title01\" ."
                              "<m02> a nmo:Message ; "
                              "nie:isLogicalPartOf <tracker-live-tests> ; "
                              "  nie:title \"title02\" ."
                              "<m03> a nmo:Message ; "
                              "nie:isLogicalPartOf <tracker-live-tests> ; "
                              "  nie:title \"title03\" ."
                              "<m04> a nmo:Message ; "
                              "nie:isLogicalPartOf <tracker-live-tests> ; "
                              "  nie:title \"title04\" ."
                              " }", QSparqlQuery::InsertStatement), false);
    // Wait until all inserted data is in tracker
    QVERIFY(waiterMsg.waitForAllUpdated());
    // Run the original query
    liveQuery->start();
    QVERIFY(waitUntilModelFilled(liveQuery, 4));

    // Pause the updates
    liveQuery->setUpdatesEnabled(false);

    // Do 2 bunches of changes, wait enough between them so that there will be 2
    // separate GraphChanged signals.
    waiterMsg.setUpdatesAmount(1, 1);
    executeQuery(QSparqlQuery("delete { "
                              "<m01> nie:title \"title01\" . "
                              "} "
                              "insert { "
                              "<m01> nie:title \"title01changed\" . "
                              " }", QSparqlQuery::InsertStatement), false);
    // Wait until all inserted data is in tracker
    QVERIFY(waiterMsg.waitForAllUpdated());
    waiterMsg.setUpdatesAmount(1, 1);
    executeQuery(QSparqlQuery("delete { "
                              "<m03> nie:title \"title03\" . "
                              "} "
                              "insert { "
                              "<m03> nie:title \"title03changed\" . "
                              " }", QSparqlQuery::InsertStatement), false);
    // Wait until all inserted data is in tracker
    QVERIFY(waiterMsg.waitForAllUpdated());

    // We shouldn't run any update queries
    QCOMPARE(liveQuery->property("test-updatequerycount"), QVariant(0));

    QSignalSpy spy(liveQuery->model(), SIGNAL(dataChanged(const QModelIndex&,
                                                          const QModelIndex&)));
    // Resume the updates
    liveQuery->setUpdatesEnabled(true);

    // Wait until the model has the data
    QTime timer;
    timer.start();
    while (spy.count() < 2 && timer.elapsed() < 5000)
        QTest::qWait(100);

    QCOMPARE(spy.count(), 2);

    // Only one update query is ran (data is gathered from both signals).
    QCOMPARE(liveQuery->property("test-updatequerycount"), QVariant(1));

    // And the data is correct
    QStringList data = getRows(liveQuery->model(), QList<int>() << 1 << 2);

    QStringList expected;
    expected << "m01:title01changed:"
             << "m02:title02:"
             << "m03:title03changed:"
             << "m04:title04:";
    QCOMPARE(data, expected);
}

void tst_TrackerLivePausing::tooManyIds()
{
    // Note: checking the update query counts is normally not robust.  But in
    // this test it should be, since we pause the updates, and after the
    // pausing, there should only one update query ran.  If this test keeps
    // failing randomly, something might be wrong.

    // Run the original query
    liveQuery->start();
    QVERIFY(waitUntilModelFilled(liveQuery, 0));

    // Pause the updates to ensure that TrackerLiveQuery collects all the id's
    // it gets, without running any queries.
    liveQuery->setUpdatesEnabled(false);

    // Insert many objects
    QSparqlConnection conn("QTRACKER_DIRECT");
    for (int i = 0; i < 1500; ++i) {
        QString query = QString("insert { "
                                "<m%1> a nmo:Message ; "
                                "nie:isLogicalPartOf <tracker-live-tests> ; "
                                "  nie:title \"title%1\" ."
                                " }").arg(i);
        executeQuery(QSparqlQuery(query, QSparqlQuery::InsertStatement));
    }
    // wait enough so that all the signals are received
    QTest::qWait(4000);

    // Resume the updates
    liveQuery->setUpdatesEnabled(true);

    // Wait until the model has the data
    QTime timer;
    timer.start();
    while (liveQuery->model()->rowCount() < 1500 && timer.elapsed() < 5000)
        QTest::qWait(100);

    QCOMPARE(liveQuery->model()->rowCount(), 1500);

    // Only one update query is ran (data is gathered from all signals).
    QCOMPARE(liveQuery->property("test-updatequerycount"), QVariant(1));
}

QTEST_MAIN( tst_TrackerLivePausing )
#include "tst_live_pausing.moc"
