/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/updatewaiter.h"

#include "TrackerLiveQuery"

#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

//const QString qtest(qTableName( "qtest", __FILE__ )); // FIXME: what's this

//TESTED_FILES=

/*

This test contains the simplest possible live model:
column 0: tracker:id of an e-mail
column 1: subject of the e-mail

*/

class tst_TrackerLiveSimpleOrdered : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveSimpleOrdered();
    virtual ~tst_TrackerLiveSimpleOrdered();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void initialModel();
    void resourceAdded_data();
    void resourceAdded();

private:
    void initLiveQuery(Qt::SortOrder order);

private:
    // Object under test
    TrackerLiveQuery* liveQuery;
    QSparqlConnection* conn;
    QAbstractItemModel* model;
};

Q_DECLARE_METATYPE(QModelIndex)

tst_TrackerLiveSimpleOrdered::tst_TrackerLiveSimpleOrdered()
: liveQuery(0), conn(0), model(0)
{
}

tst_TrackerLiveSimpleOrdered::~tst_TrackerLiveSimpleOrdered()
{
}

void tst_TrackerLiveSimpleOrdered::initTestCase()
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
    // Clean up leftover data from previous test run
    cleanup();
}

void tst_TrackerLiveSimpleOrdered::cleanupTestCase()
{
}

void tst_TrackerLiveSimpleOrdered::init()
{
    // Insert initial test data
    UpdateWaiter waiterEmail(QString("http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#Message"), 12);
    executeQuery(QSparqlQuery(
                     "insert { "
                     "<tracker-live-tests> a nie:InformationElement . "
                     "<uri001> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"fff\" ."
                     "<uri002> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"hhh\" ."
                     "<uri003> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"ddd\" ."
                     "<uri004> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"bbb\" ."
                     "}",
                     QSparqlQuery::InsertStatement),
                 false);
    // Wait until all inserted data is in tracker
    QVERIFY(waiterEmail.waitForAllUpdated());
}

void tst_TrackerLiveSimpleOrdered::initLiveQuery(Qt::SortOrder order)
{
    QString orderBy = "?s";
    if (order == Qt::DescendingOrder) {
        orderBy = "DESC(?s)";
    }
    // Create the object under test
    QString origQuery = QString("select tracker:id(?e) ?s "
                                "{ ?e a nmo:Email ; "
                                "nie:isLogicalPartOf <tracker-live-tests> ;"
                                "nmo:messageSubject ?s . "
                                "} "
                                "order by %1").arg(orderBy);
    conn = new QSparqlConnection("QTRACKER_DIRECT");
    // use the ctor taking QSparqlQuery just for fun (and coverage)
    liveQuery = new TrackerLiveQuery(QSparqlQuery(origQuery), 2,
                                     QList<int>() << 0, *conn);

    // This creates an update schema: run this query...
    QString updateQuery = QString("select tracker:id(?e) ?s "
                                  "{ ?e a nmo:Email ; "
                                  "nie:isLogicalPartOf <tracker-live-tests> ;"
                                  "nmo:messageSubject ?s . "
                                  "%FILTER "
                                  "}"
                                  "order by %1").arg(orderBy);

    TrackerPartialUpdater up(updateQuery);
    // When these changes come
    up.watchClass("nmo:Message", QStringList(),
                  "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);

    liveQuery->addUpdater(up);

    liveQuery->setCollationColumns(QList<TrackerLiveQuery::CollationColumn>() <<
                                   TrackerLiveQuery::CollationColumn(1, QVariant::String, order));
    liveQuery->start();

    // store this for convenience
    model = liveQuery->model();
}

void tst_TrackerLiveSimpleOrdered::cleanup()
{
    // Delete the tested object
    delete liveQuery;
    liveQuery = 0;
    model = 0;
    delete conn;
    conn = 0;

    // Clean up test data
    executeQuery(QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

void tst_TrackerLiveSimpleOrdered::initialModel()
{
    initLiveQuery(Qt::AscendingOrder);

    QSignalSpy spy(model, SIGNAL(rowsInserted(const QModelIndex&, int, int)));

    // ignore the return value; we'll do the checks here
    waitUntilModelFilled(liveQuery, 2);

    // Check that the initial rowsInserted signals are correct.  The default
    // dataReadyInterval is 1, so we should get 4 different signals.
    QCOMPARE(spy.count(), 4);

    for (int r = 0; r < 4; ++ r) {
        QList<QVariant> parameters = spy.takeFirst();
        QModelIndex ix = parameters.at(0).value<QModelIndex>();
        int start = parameters.at(1).value<int>();
        int end = parameters.at(2).value<int>();

        QCOMPARE(ix, QModelIndex());
        QCOMPARE(start, r);
        QCOMPARE(end, r);
    }

    QCOMPARE(model->rowCount(), 4);
    QCOMPARE(model->columnCount(), 2);

    // Verify the data and the ordering
    QStringList data = getRows(model, QList<int>() << 1);
    QCOMPARE(data.at(0), QString("bbb:"));
    QCOMPARE(data.at(1), QString("ddd:"));
    QCOMPARE(data.at(2), QString("fff:"));
    QCOMPARE(data.at(3), QString("hhh:"));
}

void tst_TrackerLiveSimpleOrdered::resourceAdded_data()
{
    QTest::addColumn<bool>("asc");
    QTest::addColumn<QString>("subject");
    QTest::addColumn<int>("row");

    QTest::newRow("beginning") << true << QString("aaa") << 0;
    QTest::newRow("middle") << true << QString("ccc") << 1;
    QTest::newRow("end") << true << QString("iii") << 4;

    QTest::newRow("end_desc") << false << QString("aaa") << 4;
    QTest::newRow("middle_desc") << false << QString("ccc") << 3;
    QTest::newRow("beginning_desc") << false << QString("iii") << 0;
}

void tst_TrackerLiveSimpleOrdered::resourceAdded()
{
    QFETCH(bool, asc);
    QFETCH(QString, subject);
    QFETCH(int, row);

    if (asc)
        initLiveQuery(Qt::AscendingOrder);
    else
        initLiveQuery(Qt::DescendingOrder);


    QVERIFY(waitUntilModelFilled(liveQuery, 4));

    QStringList oldData = getRows(model, QList<int>() << 0 << 1);

    // Insert a resource. Depending on the subject the row will appear at a
    // different location in the model.
    QSignalSpy spy(model, SIGNAL(rowsInserted(const QModelIndex&, int, int)));
    executeQuery(QSparqlQuery(QString(
                                  "insert {"
                                  "<addeduri001> a nmo:Email ; "
                                  "nie:isLogicalPartOf <tracker-live-tests> ; "
                                  "nmo:messageSubject \"%1\" . }").arg(subject),
                              QSparqlQuery::InsertStatement));
    QTime timer;
    timer.start();
    while(timer.elapsed() < 5000 && spy.count() < 1)
        QTest::qWait(100);

    // Verify the inserted row
    QCOMPARE(spy.count(), 1);
    QList<QVariant> parameters = spy.takeFirst();
    QModelIndex ix = parameters.at(0).value<QModelIndex>();
    int start = parameters.at(1).value<int>();
    int end = parameters.at(2).value<int>();

    QCOMPARE(ix, QModelIndex());
    QCOMPARE(start, row);
    QCOMPARE(end, row);

    QCOMPARE(model->data(model->index(start, 1)).toString(), subject);

    // Verify the data of the whole model
    QStringList newData = getRows(model, QList<int>() << 0 << 1);
    for (int i = 0; i < row; ++i) {
        QCOMPARE(newData.at(i), oldData.at(i));
    }
    QVERIFY(newData.at(row).endsWith(subject + ":"));
    for (int i = row + 1; i < newData.size(); ++i) {
        QCOMPARE(newData.at(i), oldData.at(i-1));
    }
}

QTEST_MAIN( tst_TrackerLiveSimpleOrdered )
#include "tst_live_simple_ordered.moc"
