/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/updatewaiter.h"

#include "TrackerLiveQuery"

#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

//const QString qtest(qTableName( "qtest", __FILE__ )); // FIXME: what's this

//TESTED_FILES=

/*

This test contains the simplest possible live model:
column 0: tracker:id of an e-mail
column 1: subject of the e-mail

*/

class tst_TrackerLiveSimple : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveSimple();
    virtual ~tst_TrackerLiveSimple();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void initialModel();
    void resourceAdded();
    void resourceRemoved();
    void resourceChanged();
    void setColumnHeaders();

private:
    // Object under test
    TrackerLiveQuery* liveQuery;
    QSparqlConnection* conn;
    QAbstractItemModel* model;
};

Q_DECLARE_METATYPE(QModelIndex)

tst_TrackerLiveSimple::tst_TrackerLiveSimple()
: liveQuery(0), conn(0), model(0)
{
}

tst_TrackerLiveSimple::~tst_TrackerLiveSimple()
{
}

void tst_TrackerLiveSimple::initTestCase()
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
    // Clean up leftover data from previous test run
    cleanup();
}

void tst_TrackerLiveSimple::cleanupTestCase()
{
}

void tst_TrackerLiveSimple::init()
{
    // Insert initial test data
    UpdateWaiter waiterEmail(QString("http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#Message"), 6);
    executeQuery(QSparqlQuery(
                     "insert { "
                     "<tracker-live-tests> a nie:InformationElement . "
                     "<uri001> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"subject 1\" ."
                     "<uri002> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ;"
                     "nmo:messageSubject \"subject 2\" ."
                     "}",
                     QSparqlQuery::InsertStatement),
                 false);
    // Wait until all inserted data is in tracker
    QVERIFY(waiterEmail.waitForAllUpdated());

    // Create the object under test
    QString origQuery("select tracker:id(?e) ?s "
                      "{ ?e a nmo:Email ; "
                      "nie:isLogicalPartOf <tracker-live-tests> ;"
                      "nmo:messageSubject ?s . "
                      "}");
    conn = new QSparqlConnection("QTRACKER");
    liveQuery = new TrackerLiveQuery(origQuery, 2, QList<int>() << 0, *conn);

    // This creates an update schema: run this query...
    QString updateQuery("select tracker:id(?e) ?s "
                        "{ ?e a nmo:Email ; "
                        "nie:isLogicalPartOf <tracker-live-tests> ;"
                        "nmo:messageSubject ?s . "
                        "%FILTER "
                        "}");
    TrackerPartialUpdater up(updateQuery);
    // When these changes come
    up.watchClass("nmo:Message", QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);

    liveQuery->addUpdater(up);

    liveQuery->start();

    // store this for convenience
    model = liveQuery->model();
}

void tst_TrackerLiveSimple::cleanup()
{
    // Delete the tested object
    delete liveQuery;
    liveQuery = 0;
    model = 0;
    delete conn;
    conn = 0;

    // Clean up test data
    executeQuery(QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

void tst_TrackerLiveSimple::initialModel()
{
    // ignore the return value; we'll do the checks here
    waitUntilModelFilled(liveQuery, 2);

    QCOMPARE(model->rowCount(), 2);
    QCOMPARE(model->columnCount(), 2);

    // Verify the data but don't care about ordering.
    QSet<QString> subjects;
    subjects.insert(model->data(model->index(0, 1)).toString());
    subjects.insert(model->data(model->index(1, 1)).toString());

    QVERIFY(subjects.contains(QString("subject 1")));
    QVERIFY(subjects.contains(QString("subject 2")));
}

void tst_TrackerLiveSimple::resourceAdded()
{
    QVERIFY(waitUntilModelFilled(liveQuery, 2));

    // Insert a resource
    QSignalSpy spy(model, SIGNAL(rowsInserted(const QModelIndex&, int, int)));
    executeQuery(QSparqlQuery(
                     "insert {"
                     "<addeduri001> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ; "
                     "nmo:messageSubject \"added subject\" . }",
                     QSparqlQuery::InsertStatement));
    QTime timer;
    timer.start();
    while(timer.elapsed() < 5000 && spy.count() < 1)
        QTest::qWait(100);

    // Verify the inserted row but don't care about ordering
    QCOMPARE(spy.count(), 1);
    QList<QVariant> parameters = spy.takeFirst();
    QModelIndex ix = parameters.at(0).value<QModelIndex>();
    int start = parameters.at(1).value<int>();
    int end = parameters.at(2).value<int>();

    QCOMPARE(ix, QModelIndex());
    QCOMPARE(start, end); // only 1 row inserted

    QCOMPARE(model->data(model->index(start, 1)).toString(),
             QString("added subject"));
}

void tst_TrackerLiveSimple::resourceRemoved()
{
    QVERIFY(waitUntilModelFilled(liveQuery, 2));

    // Find the resource we're going to delete from the model (don't care about
    // ordering)
    int row = 0;
    while (row < model->rowCount() &&
           model->data(model->index(row, 1)).toString() != "subject 1")
        ++row;

    // Delete a resource
    QSignalSpy spy(model, SIGNAL(rowsRemoved(const QModelIndex&, int, int)));
    executeQuery(QSparqlQuery(
                     "delete {"
                     "<uri001> a rdfs:Resource . }",
                     QSparqlQuery::DeleteStatement));
    QTime timer;
    timer.start();
    while(timer.elapsed() < 5000 && spy.count() < 1)
        QTest::qWait(100);

    // Verify the row got removed
    QCOMPARE(spy.count(), 1);
    QList<QVariant> parameters = spy.takeFirst();
    QModelIndex ix = parameters.at(0).value<QModelIndex>();
    int start = parameters.at(1).value<int>();
    int end = parameters.at(2).value<int>();

    QCOMPARE(ix, QModelIndex());
    QCOMPARE(start, end); // only 1 row removed
    QCOMPARE(start, row); // and it was the correct one
}

void tst_TrackerLiveSimple::resourceChanged()
{
    QVERIFY(waitUntilModelFilled(liveQuery, 2));

    // Find the resource we're going to modify (don't care about ordering)
    int row = 0;
    while (row < model->rowCount() &&
           model->data(model->index(row, 1)).toString() != "subject 2")
        ++row;

    // Modify a property of a resource (the subject of an email)
    QSignalSpy spy(model,
                   SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
    executeQuery(QSparqlQuery(
                     "delete { "
                     "<uri002> nmo:messageSubject \"subject 2\" . "
                     "}"
                     "insert { "
                     "<uri002> nmo:messageSubject \"subject changed\" . "
                     "}",
                     QSparqlQuery::DeleteStatement));
    QTime timer;
    timer.start();
    while(timer.elapsed() < 5000 && spy.count() < 1)
        QTest::qWait(100);

    // Verify the row got modified
    QCOMPARE(spy.count(), 1);
    QList<QVariant> parameters = spy.takeFirst();
    QModelIndex topLeft = parameters.at(0).value<QModelIndex>();
    QModelIndex bottomRight = parameters.at(1).value<QModelIndex>();

    QCOMPARE(topLeft.row(), row);
    QCOMPARE(bottomRight.row(), row);

    QCOMPARE(topLeft.column(), 1);
    QCOMPARE(bottomRight.column(), 1);

    QCOMPARE(model->data(model->index(row, 1)).toString(),
             QString("subject changed"));

}

void tst_TrackerLiveSimple::setColumnHeaders()
{
    // ignore the return value; we'll do the checks here
    waitUntilModelFilled(liveQuery, 2);

    QCOMPARE(model->columnCount(), 2);
    QVERIFY(model->setHeaderData(0, Qt::Horizontal, "Email"));
    QVERIFY(model->setHeaderData(1, Qt::Horizontal, "Subject"));

    QCOMPARE(model->headerData(0, Qt::Horizontal), QVariant("Email"));
    QCOMPARE(model->headerData(1, Qt::Horizontal), QVariant("Subject"));
}


QTEST_MAIN( tst_TrackerLiveSimple )
#include "tst_live_simple.moc"
