/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/modeltesthelpers.h"
#include "../utils/signalsrecorder.h"
#include "TrackerLiveQuery"

#include <QScopedPointer>
#include <QtTest/QtTest>
#include <QtSparql/QtSparql>

class tst_TrackerLiveUpdaters : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveUpdaters();
    virtual ~tst_TrackerLiveUpdaters();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void TrackerPartialUpdater_copy_constructed_should_produce_identical_result_as_original();
    void TrackerPartialUpdater_assigned_to_empty_should_produce_identical_result_as_original();
    void TrackerPartialUpdater_assigned_to_nonempty_should_produce_identical_result_as_original();

    void TrackerFullUpdater_copy_constructed_should_produce_identical_result_as_original();
    void TrackerFullUpdater_assigned_to_empty_should_produce_identical_result_as_original();
    void TrackerFullUpdater_assigned_to_nonempty_should_produce_identical_result_as_original();

private:
    void initRefPartialUpdater();
    void initRefFullUpdater();
    template<class Updater>
        void createParallelLiveQuery_impl
                (QSharedPointer<TrackerLiveQuery>& result, const Updater& withUpdater);
    template<class Updater>
        QSharedPointer<TrackerLiveQuery> createParallelLiveQuery(const Updater& withUpdater);
    TrackerLiveQuery* createLiveQuery();
    void verifyRowInsertionToBothModels_impl(bool& result, QAbstractItemModel* model1, QAbstractItemModel* model2);
    bool verifyRowInsertionToBothModels(QAbstractItemModel* model1, QAbstractItemModel* model2);
    bool addResource();

private:
    QSparqlConnection* conn;
    TrackerLiveQuery* refLiveQuery;
    TrackerPartialUpdater* refPartialUpdater;
    TrackerFullUpdater* refFullUpdater;
};

Q_DECLARE_METATYPE(QModelIndex);

namespace {
    const char* const origQuery =
            "select tracker:id(?e) ?s "
            "{ ?e a nmo:Email ; "
            "nie:isLogicalPartOf <tracker-live-tests> ;"
            "nmo:messageSubject ?s . "
            "}";

    const int origQueryColCount = 2;
    const QList<int> origIdentityColumns(QList<int>() << 0);

    const char* const initialDataInsertQuery =
            "insert { "
            "<tracker-live-tests> a nie:InformationElement . "
            "<uri001> a nmo:Email ; "
            "nie:isLogicalPartOf <tracker-live-tests> ;"
            "nmo:messageSubject \"subject 1\" ."
            "<uri002> a nmo:Email ; "
            "nie:isLogicalPartOf <tracker-live-tests> ;"
            "nmo:messageSubject \"subject 2\" ."
            "}";

    const int initialDataRowCount = 2;

    const char* const updateQuery =
            "select tracker:id(?e) ?s "
            "{ ?e a nmo:Email ; "
            "nie:isLogicalPartOf <tracker-live-tests> ;"
            "nmo:messageSubject ?s . "
            "%FILTER "
            "}";
}

tst_TrackerLiveUpdaters::tst_TrackerLiveUpdaters()
    : conn(0)
{
}

tst_TrackerLiveUpdaters::~tst_TrackerLiveUpdaters()
{
}

void tst_TrackerLiveUpdaters::initTestCase()
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
}

void tst_TrackerLiveUpdaters::cleanupTestCase()
{
}

void tst_TrackerLiveUpdaters::init()
{
    // Insert initial test data
    QVERIFY( executeQuery(QSparqlQuery(
                     initialDataInsertQuery,
                     QSparqlQuery::InsertStatement),
                 true) );

    conn = new QSparqlConnection("QTRACKER_DIRECT");

    refLiveQuery = createLiveQuery();
    refPartialUpdater = 0;
    refFullUpdater = 0;
}

void tst_TrackerLiveUpdaters::initRefPartialUpdater()
{
    refPartialUpdater = new TrackerPartialUpdater(updateQuery);
    refPartialUpdater->watchClass("nmo:Message", QStringList(),
            "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    refLiveQuery->addUpdater(*refPartialUpdater);
}

void tst_TrackerLiveUpdaters::initRefFullUpdater()
{
    refFullUpdater = new TrackerFullUpdater;
    refFullUpdater->watchClass("nmo:Message", QStringList());
    refLiveQuery->addUpdater(*refFullUpdater);
}

void tst_TrackerLiveUpdaters::cleanup()
{
    delete refPartialUpdater;
    refPartialUpdater = 0;

    delete refFullUpdater;
    refFullUpdater = 0;

    delete refLiveQuery;
    refLiveQuery = 0;

    delete conn;
    conn = 0;

    // Clean up test data
    executeQuery(QSparqlQuery(
                     "delete { "
                     "?u a rdfs:Resource . } "
                     "where { "
                     "?u nie:isLogicalPartOf <tracker-live-tests> . "
                     "} "
                     "delete {"
                     "<tracker-live-tests> a nie:InformationElement . "
                     "}",
                     QSparqlQuery::DeleteStatement));
}

template<class Updater>
void tst_TrackerLiveUpdaters::createParallelLiveQuery_impl
        (QSharedPointer<TrackerLiveQuery>& result, const Updater& withUpdater)
{
    result.clear();
    QSharedPointer<TrackerLiveQuery> liveQuery(createLiveQuery());
    liveQuery->addUpdater(withUpdater);

    refLiveQuery->start();
    QVERIFY( waitUntilModelFilled(refLiveQuery, initialDataRowCount) );
    liveQuery->start();
    QVERIFY( waitUntilModelFilled(liveQuery.data(), initialDataRowCount) );
    QVERIFY( equalModels(liveQuery->model(), refLiveQuery->model()) );

    result = liveQuery;
}

template<class Updater>
QSharedPointer<TrackerLiveQuery> tst_TrackerLiveUpdaters::createParallelLiveQuery
        (const Updater& withUpdater)
{
    QSharedPointer<TrackerLiveQuery> result;
    createParallelLiveQuery_impl(result, withUpdater);
    return result;
}

void tst_TrackerLiveUpdaters::verifyRowInsertionToBothModels_impl
        (bool& result, QAbstractItemModel* model1, QAbstractItemModel* model2)
{
    result = false;

    QVERIFY( equalModels(model1, model2) );

    SignalsRecorder signalsRecorder;
    SignalList sl;
    sl.append(model1, SIGNAL(rowsInserted(QModelIndex, int, int)));
    sl.append(model2, SIGNAL(rowsInserted(QModelIndex, int, int)));
    signalsRecorder.listen(sl);

    QVERIFY( addResource() );

    QVERIFY( signalsRecorder.waitForSignals(sl, 5000) );
    const QList<SignalsRecorder::SignalData> model1SignalData(
            signalsRecorder.signalData(model1,SIGNAL(rowsInserted(QModelIndex, int, int))) );
    const QList<SignalsRecorder::SignalData> model2SignalData(
            signalsRecorder.signalData(model2,SIGNAL(rowsInserted(QModelIndex, int, int))) );
    QCOMPARE( model1SignalData.back().params()[0].value<QModelIndex>(),
              model2SignalData.back().params()[0].value<QModelIndex>() );
    QCOMPARE( model1SignalData.back().params()[1], model2SignalData.back().params()[1] );
    QCOMPARE( model1SignalData.back().params()[2], model2SignalData.back().params()[2] );
    QVERIFY( equalModels(model1, model2) );

    result = true;
}

bool tst_TrackerLiveUpdaters::verifyRowInsertionToBothModels
        (QAbstractItemModel* model1, QAbstractItemModel* model2)
{
    bool result = false;
    verifyRowInsertionToBothModels_impl(result, model1, model2);
    return result;
}

void tst_TrackerLiveUpdaters::TrackerPartialUpdater_copy_constructed_should_produce_identical_result_as_original()
{
    initRefPartialUpdater();
    TrackerPartialUpdater updaterCopy(*refPartialUpdater);
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

void tst_TrackerLiveUpdaters::TrackerPartialUpdater_assigned_to_empty_should_produce_identical_result_as_original()
{
    initRefPartialUpdater();
    TrackerPartialUpdater updaterCopy("<<<INVALID QUERY { %FILTER }>>>");
    updaterCopy = *refPartialUpdater;
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

void tst_TrackerLiveUpdaters::TrackerPartialUpdater_assigned_to_nonempty_should_produce_identical_result_as_original()
{
    initRefPartialUpdater();
    TrackerPartialUpdater updaterCopy("<<<INVALID QUERY { %FILTER }>>>");
    updaterCopy.watchClass(
            "<<<invalid_class>>>", QStringList(),
            "?<<<invalid_var>>> in %LIST", TrackerPartialUpdater::Subject, 0);
    updaterCopy = *refPartialUpdater;
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

void tst_TrackerLiveUpdaters::TrackerFullUpdater_copy_constructed_should_produce_identical_result_as_original()
{
    initRefFullUpdater();
    TrackerFullUpdater updaterCopy(*refFullUpdater);
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

void tst_TrackerLiveUpdaters::TrackerFullUpdater_assigned_to_empty_should_produce_identical_result_as_original()
{
    initRefFullUpdater();
    TrackerFullUpdater updaterCopy;
    updaterCopy = *refFullUpdater;
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

void tst_TrackerLiveUpdaters::TrackerFullUpdater_assigned_to_nonempty_should_produce_identical_result_as_original()
{
    initRefFullUpdater();
    TrackerFullUpdater updaterCopy;
    updaterCopy.watchClass("<<<invalid_class>>>", QStringList());
    updaterCopy = *refFullUpdater;
    QSharedPointer<TrackerLiveQuery> parallelLiveQuery( createParallelLiveQuery(updaterCopy) );
    QVERIFY( parallelLiveQuery );

    QVERIFY( verifyRowInsertionToBothModels(parallelLiveQuery->model(), refLiveQuery->model()) );
}

TrackerLiveQuery* tst_TrackerLiveUpdaters::createLiveQuery()
{
    TrackerLiveQuery* lq = new TrackerLiveQuery(origQuery, origQueryColCount,
                                                origIdentityColumns,
                                                *conn);
    return lq;
}

bool tst_TrackerLiveUpdaters::addResource()
{
    return executeQuery(QSparqlQuery(
                     "insert {"
                     "<addeduri001> a nmo:Email ; "
                     "nie:isLogicalPartOf <tracker-live-tests> ; "
                     "nmo:messageSubject \"added subject\" . }",
                     QSparqlQuery::InsertStatement));
}


QTEST_MAIN( tst_TrackerLiveUpdaters )
#include "tst_live_updaters.moc"
