/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "usermodifiablemodel.h"

#include <QtTest/QtTest>
#include <QStandardItemModel>

#include "../utils/modelequator.h"
#include "../utils/modeltesthelpers.h"
#include "updaterule.h"

Q_DECLARE_METATYPE(QModelIndex)

class tst_UserModifiableModel : public QObject
{
    Q_OBJECT

public:
    tst_UserModifiableModel();
    virtual ~tst_UserModifiableModel();

public Q_SLOTS:
    void init();
    void cleanup();

private Q_SLOTS:
    void identities();
    void identities2();

    void lazyInsertRow();
    void combinesInserts();

    void addRowWithData();
    void addRowsBatch();

    void userRemoveRow();
    void userRemovedRowReallyRemoved();

    void userRemovalDoesntAffectOtherRows();
    void userRemoveRowFailsWithIllegalIndices();

    void allRowsUserRemoved();

    void affectedRowsEmpty();
    void affectedRowsNonEmpty();

    void affectedRowsWithMultipleRules();
};

void equalModelRow_impl(bool& result, const UserModifiableModel& model, int row,
                        const QVariantList& rowData, const QList<int>& idColumns)
{
    result = false;
    QVERIFY( equalRowData(&model, row, rowData) );
    const IdentityList rowIdentity(IdentityList::fromData(rowData, idColumns));
    QVERIFY( model.hasRowWithIdentity(rowIdentity) );
    QCOMPARE( model.identityToRow(rowIdentity).row(), row );
    result = true;
}

bool equalModelRow(const UserModifiableModel& model, int row,
                   const QVariantList& rowData, const QList<int>& idColumns)
{
    bool result = false;
    equalModelRow_impl(result, model, row, rowData, idColumns);
    return result;
}

tst_UserModifiableModel::tst_UserModifiableModel()
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
}

tst_UserModifiableModel::~tst_UserModifiableModel()
{
}

void tst_UserModifiableModel::init()
{
}

void tst_UserModifiableModel::cleanup()
{
}

void tst_UserModifiableModel::identities()
{
    // Test that UserModifiableModel keeps its identity cache up to date.
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0 << 1;
    m.setIdentityColumns(idColumns);

    m.beginLazyRowInsertions();
    QVariantList dataList1, dataList2;
    dataList1 << 3 << 0;
    dataList2 << 6 << 8;
    m.lazyInsertRow(0, dataList1);
    m.lazyInsertRow(0, dataList2);
    m.endLazyRowInsertions();

    QVERIFY(m.hasRowWithIdentity(IdentityList::fromRow(m, 0, idColumns)));
    QVERIFY(m.hasRowWithIdentity(IdentityList::fromRow(m, 1, idColumns)));

    QCOMPARE(m.identityToRow(IdentityList::fromRow(m, 0, idColumns)).row(), 0);
    QCOMPARE(m.identityToRow(IdentityList::fromRow(m, 1, idColumns)).row(), 1);

    IdentityList bogus;
    bogus << 8 << 9;
    QVERIFY(!m.hasRowWithIdentity(bogus));

    IdentityList toBeRemoved;
    toBeRemoved << 3 << 0;
    QVERIFY(m.hasRowWithIdentity(toBeRemoved));

    m.beginLazyRowRemovals();
    m.lazyRemoveRow(toBeRemoved);
    m.endLazyRowRemovals();

    QVERIFY(m.hasRowWithIdentity(IdentityList::fromRow(m, 0, idColumns)));
    QVERIFY(!m.hasRowWithIdentity(toBeRemoved));

}

void  tst_UserModifiableModel::identities2()
{
    // This is a regression test

    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);
    QList<QVariantList> modelData;

    m.beginLazyRowInsertions();
    modelData << (QVariantList() << 1 << "data1");
    m.lazyInsertRow(0, modelData.back());
    m.endLazyRowInsertions();

    for (int i = 0; i < 1; ++i) {
        QVERIFY(equalModelRow(m, i, modelData[i], idColumns));
    }

    m.beginLazyRowInsertions();
    // Insert before the previous row
    modelData.insert(0, QVariantList() <<  2 << "data2");
    m.lazyInsertRow(0, modelData[0]);
    m.endLazyRowInsertions();

    for (int i = 0; i < 2; ++i) {
        QVERIFY(equalModelRow(m, i, modelData[i], idColumns));
    }
}

void tst_UserModifiableModel::lazyInsertRow()
{
    // Test that lazyInsertRow commits data to the model only after it has sent signals to indicate
    // the data is available

    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0 << 1;
    m.setIdentityColumns(idColumns);
    QList<QVariantList> modelData;
    QSignalSpy beginSpy(&m, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)));
    QSignalSpy changedSpy(&m, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
    QSignalSpy endSpy(&m, SIGNAL(rowsInserted(const QModelIndex&, int, int)));
    int modelRowCountBefore = m.rowCount();

    m.beginLazyRowInsertions();
    for (int row = 0; row < 10; ++row) {
        QVariantList dataList;
        dataList << row << QString("data %1").arg(row);
        m.lazyInsertRow(0, dataList);
        modelData << dataList;

        // Check that nothing is added to the model yet
        QCOMPARE(m.rowCount(), modelRowCountBefore);
        const IdentityList rowIdentity(IdentityList::fromData(dataList, idColumns));
        QVERIFY(!m.hasRowWithIdentity(rowIdentity));
        QCOMPARE(m.identityToRow(rowIdentity), QPersistentModelIndex());
    }
    m.endLazyRowInsertions();

    QCOMPARE(beginSpy.count(), 1);
    QVERIFY(verifySignalParameters(beginSpy.back(), QModelIndex(), 0, 9));
    QCOMPARE(changedSpy.count(), 0);
    QCOMPARE(endSpy.count(), 1);
    QVERIFY(verifySignalParameters(endSpy.back(), QModelIndex(), 0, 9));

    QCOMPARE(m.rowCount(), 10);
    for (int row = 0; row < 10; ++row) {
        QVERIFY(equalModelRow(m, row, modelData[row], idColumns));
    }
}

void tst_UserModifiableModel::combinesInserts()
{
    // Test that UserModifiableModel combines multiple row insertions into the
    // same rowsInserted signal.
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0 << 1;
    m.setIdentityColumns(idColumns);
    QList<QVariantList> modelData;

    m.beginLazyRowInsertions();
    for (int i = 0; i < 10; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);
        modelData << dataList;
    }
    m.endLazyRowInsertions();

    // Add rows to 2 different places.  We make sure we add to the same
    // place always, by using a QPersistentModelIndex. (When the first batch is
    // inserted, the row indices change.)

    QSignalSpy beginSpy(&m, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)));
    QSignalSpy changedSpy(&m, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
    QSignalSpy endSpy(&m, SIGNAL(rowsInserted(const QModelIndex&, int, int)));

    m.beginLazyRowInsertions();
    QPersistentModelIndex ix(m.index(3, 0));
    for (int i = 0; i < 5; ++i) {
        QVariantList dataList;
        dataList << 10 + i << QString("data %1").arg(10 + i);
        m.lazyInsertRow(ix.row(), dataList);
        modelData.insert(i+3, dataList);
    }


    QPersistentModelIndex ix2(m.index(7, 0));
    for (int i = 0; i < 5; ++i) {
        QVariantList dataList;
        dataList << 20 + i << QString("data %1").arg(20 + i);
        m.lazyInsertRow(ix2.row(),dataList);
        modelData.insert(i+5+7, dataList);
    }
    m.endLazyRowInsertions();

    QCOMPARE(m.rowCount(), 10+5+5);
    for (int row = 0; row < m.rowCount(); ++row) {
        QVERIFY(equalModelRow(m, row, modelData[row], idColumns));
    }

    // we got 2 different signals
    QCOMPARE(beginSpy.count(), 2);
    QCOMPARE(changedSpy.count(), 0);
    QCOMPARE(endSpy.count(), 2);
    QVERIFY(verifySignalParameters(beginSpy[0], QModelIndex(), 3, 3 + 5-1));
    QVERIFY(verifySignalParameters(endSpy[0], QModelIndex(), 3, 3 + 5-1));
    QVERIFY(verifySignalParameters(beginSpy[1], QModelIndex(), 5+7, 5+7 + 5-1));
    QVERIFY(verifySignalParameters(endSpy[1], QModelIndex(), 5+7, 5+7 + 5-1));
}

void tst_UserModifiableModel::addRowWithData()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0 << 1;
    m.setIdentityColumns(idColumns);

    QSignalSpy beginSpy(&m, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)));
    QSignalSpy changedSpy(&m, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
    QSignalSpy endSpy(&m, SIGNAL(rowsInserted(const QModelIndex&, int, int)));

    for (int row = 0; row < 10; ++row) {
        QVariantList dataList;
        dataList << row << QString("data %1").arg(row);

        QVERIFY( m.addRowWithData(row, dataList) );

        QVERIFY(equalModelRow(m, row, dataList, idColumns));

        QCOMPARE( beginSpy.count(), row+1 );
        QVERIFY( verifySignalParameters(beginSpy.back(), QModelIndex(), row, row) );

        QCOMPARE( changedSpy.count(), 0 );

        QCOMPARE( endSpy.count(), row+1 );
        QVERIFY( verifySignalParameters(endSpy.back(), QModelIndex(), row, row) );
    }
}

void tst_UserModifiableModel::addRowsBatch()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0 << 1;
    m.setIdentityColumns(idColumns);

    QSignalSpy beginSpy(&m, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)));
    QSignalSpy changedSpy(&m, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
    QSignalSpy endSpy(&m, SIGNAL(rowsInserted(const QModelIndex&, int, int)));

    m.beginAddRowsBatch(0, 10);
    QCOMPARE( beginSpy.count(), 1 );
    QVERIFY( verifySignalParameters(beginSpy.back(), QModelIndex(), 0, 9) );
    QCOMPARE( changedSpy.count(), 0 );
    QCOMPARE( endSpy.count(), 0 );

    for (int row = 0; row < 10; ++row) {
        QVariantList dataList;
        dataList << row << QString("data %1").arg(row);

        QVERIFY( m.addRowWithData(row, dataList) );

        QVERIFY(equalModelRow(m, row, dataList, idColumns));

        QCOMPARE( beginSpy.count(), 1 );
        QCOMPARE( changedSpy.count(), 0 );
        QCOMPARE( endSpy.count(), 0 );
    }

    m.endAddRowsBatch();
    QCOMPARE( beginSpy.count(), 1 );
    QCOMPARE( changedSpy.count(), 0 );
    QCOMPARE( endSpy.count(), 1 );
    QVERIFY( verifySignalParameters(endSpy.back(), QModelIndex(), 0, 9) );
}

void tst_UserModifiableModel::userRemoveRow()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    // Check the data with this reference model
    QStandardItemModel r;
    r.insertColumns(0, 2);
    r.insertRows(0, 10);

    m.beginLazyRowInsertions();
    for (int i = 0; i < 10; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);

        r.setData(r.index(i, 0), dataList[0]);
        r.setData(r.index(i, 1), dataList[1]);
    }
    m.endLazyRowInsertions();

    // Start equating after the initial data has been inserted; QSIM will emit
    // dataChanged(), UMM won't.
    ModelEquator equator(&m, &r);

    const IdentityList removedId = IdentityList::fromRow(m, 5, idColumns);
    QVERIFY(m.removeRow(5));
    QVERIFY(r.removeRow(5));

    QVERIFY(equator.good());

    QCOMPARE(m.hasRowWithIdentity(removedId), false);

    // If we try to add a row with the same identity, it won't be added (since
    // it's in the user-removed list).

    m.beginLazyRowInsertions();
    QVariantList dataList;
    dataList << 5 << QString("new data");
    m.lazyInsertRow(0, dataList);
    m.endLazyRowInsertions();

    QVERIFY(equator.good());
}

void tst_UserModifiableModel::userRemovedRowReallyRemoved()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    // Check the data with this reference model
    MovableRowModel r;
    r.addColumns(0, 2);

    ModelEquator equator(&m, &r);

    m.beginLazyRowInsertions();
    r.beginAddRowsBatch(0, 10);
    for (int i = 0; i < 10; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);

        r.addRowWithData(i, dataList);
    }
    m.endLazyRowInsertions();
    r.endAddRowsBatch();

    // User removes a row
    const IdentityList removedId = IdentityList::fromRow(m, 5, idColumns);
    QVERIFY(m.deleteRow(5));
    QVERIFY(r.deleteRow(5));
    QVERIFY(equator.good());

    // Then we tell UMM to remove that row (the row is removed by the backend)
    m.beginLazyRowRemovals();
    m.lazyRemoveRow(removedId);
    m.endLazyRowRemovals();

    // Nothing visible happens...
    QVERIFY(equator.good());

    // But if we try to add a row with the identity back again, then it's added
    QVariantList dataList;
    dataList << 5 << "added data";
    m.beginLazyRowInsertions();
    m.lazyInsertRow(0, dataList);
    m.endLazyRowInsertions();

    r.addRowWithData(0, dataList);

    QVERIFY(equator.good());
}

void tst_UserModifiableModel::userRemovalDoesntAffectOtherRows()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    // Check the data with this reference model
    MovableRowModel r;
    r.addColumns(0, 2);

    ModelEquator equator(&m, &r);

    m.beginLazyRowInsertions();
    r.beginAddRowsBatch(0, 10);
    for (int i = 0; i < 10; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);

        r.addRowWithData(i, dataList);
    }
    m.endLazyRowInsertions();
    r.endAddRowsBatch();

    // User removes a row
    const IdentityList removedId = IdentityList::fromRow(m, 6, idColumns);
    QVERIFY(m.deleteRow(6));
    QVERIFY(r.deleteRow(6));

    QVERIFY(equator.good());

    // Inserting a row
    QVariantList dataList;
    dataList << 20 << "new";

    m.beginLazyRowInsertions();
    m.lazyInsertRow(7, dataList);
    m.endLazyRowInsertions();

    r.addRowWithData(7, dataList);

    QVERIFY(equator.good());

    // Moving a row
    m.moveRow(9, 8);
    r.moveRow(9, 8);

    QVERIFY(equator.good());

    // Deleting a row
    m.beginLazyRowRemovals();
    m.lazyRemoveRow(IdentityList::fromRow(m, 7, idColumns));
    m.endLazyRowRemovals();

    r.deleteRow(7);

    QVERIFY(equator.good());
}

void tst_UserModifiableModel::userRemoveRowFailsWithIllegalIndices()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    // Check the data with this reference model
    MovableRowModel r;
    r.addColumns(0, 2);

    ModelEquator equator(&m, &r);
    QVERIFY(equator.good());

    // Removing rows from empty model should fail
    QVERIFY(!m.removeRow(0));
    QVERIFY(!m.removeRows(0, 1));
    QVERIFY(equator.good());

    const int numTestRows = 10;
    m.beginLazyRowInsertions();
    r.beginAddRowsBatch(0, numTestRows);
    for (int i = 0; i < numTestRows; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);

        r.addRowWithData(i, dataList);
    }
    m.endLazyRowInsertions();
    r.endAddRowsBatch();
    QVERIFY(equator.good());

    // Negative indices should fail
    QVERIFY(!m.removeRow(-1));
    QVERIFY(!m.removeRows(-1, 1));
    QVERIFY(equator.good());

    // Too large index should fail
    QVERIFY(!m.removeRow(numTestRows));
    QVERIFY(!m.removeRows(numTestRows, 1));
    QVERIFY(equator.good());

    // Invalid amount of rows shold fail
    QVERIFY(!m.removeRows(0, 0));
    QVERIFY(!m.removeRows(1, -1));
    QVERIFY(equator.good());

    // Too large amount of rows should fail
    QVERIFY(!m.removeRows(0, numTestRows+1));
    QVERIFY(!m.removeRows(5, numTestRows-5+1));
    QVERIFY(!m.removeRows(numTestRows-1, 2));
    QVERIFY(equator.good());
}

void tst_UserModifiableModel::allRowsUserRemoved()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    // Check the data with this reference model
    MovableRowModel r;
    r.addColumns(0, 2);

    ModelEquator equator(&m, &r);

    m.beginLazyRowInsertions();
    r.beginAddRowsBatch(0, 10);
    for (int i = 0; i < 10; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);

        r.addRowWithData(i, dataList);
    }
    m.endLazyRowInsertions();
    r.endAddRowsBatch();

    QList<IdentityList> removedIds;
    for (int i = 0; i < 10; ++i) {
        IdentityList id = IdentityList::fromRow(m, i, idColumns);
        removedIds.append(id);
        QCOMPARE(m.hasRowWithIdentity(id), true);
    }

    // User removes all rows
    QVERIFY(m.removeRows(0, 10));

    // And this is how the removal should look like
    QVERIFY(r.deleteRows(0, 10));
    QVERIFY(equator.good());

    // Check that none of the rows can be found
    Q_FOREACH(const IdentityList& id, removedIds) {
        QCOMPARE(m.hasRowWithIdentity(id), false);
    }

    // If we try to add rows with the same identities as user-removed rows, they
    // won't be added.

    m.beginLazyRowInsertions();
    Q_FOREACH(const IdentityList& id, removedIds) {
        QVariantList dataList;
        dataList << id[0] << QString("new data");
        m.lazyInsertRow(0, dataList);
    }
    m.endLazyRowInsertions();

    // Nothing was added
    QVERIFY(equator.good());

    // But if we insert a row with a new identity, it is added
    m.beginLazyRowInsertions();
    QVariantList dataList;
    dataList << 1000 << QString("really new data");
    m.lazyInsertRow(0, dataList);
    m.endLazyRowInsertions();

    r.addRowWithData(0, dataList);
    QVERIFY(equator.good());
}

void tst_UserModifiableModel::affectedRowsEmpty()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);
    QList<UpdateRule> updateRules;
    QCOMPARE( m.affectedRows(updateRules).count(), 0 );

    updateRules << UpdateRule(0, QSet<int>() << 42);
    QCOMPARE( m.affectedRows(updateRules).count(), 0 );
}

void tst_UserModifiableModel::affectedRowsNonEmpty()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    const int numTestRows = 10;
    m.beginLazyRowInsertions();
    for (int i = 0; i < numTestRows; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);
    }
    m.endLazyRowInsertions();

    {
        // Test rule with no matches
        QList<UpdateRule> updateRules;
        QCOMPARE( m.affectedRows(updateRules).count(), 0 );

        updateRules << UpdateRule(0, QSet<int>() << -1 << numTestRows);
        QCOMPARE( m.affectedRows(updateRules).count(), 0 );
    }

    {
        // Test rule with one match
        QList<UpdateRule> updateRules;
        updateRules << UpdateRule(0, QSet<int>() << 0);
        QCOMPARE( m.affectedRows(updateRules).count(), 1 );
        QVERIFY( m.affectedRows(updateRules).contains(IdentityList::fromRow(m, 0, idColumns)) );
    }

    {
        // Test rule with 2 matches
        QList<UpdateRule> updateRules;
        updateRules << UpdateRule(0, QSet<int>() << 1 << numTestRows-1);
        QCOMPARE( m.affectedRows(updateRules).count(), 2 );
        QVERIFY( m.affectedRows(updateRules).contains(IdentityList::fromRow(m, 1, idColumns)) );
        QVERIFY( m.affectedRows(updateRules).contains(IdentityList::fromRow(m, numTestRows-1, idColumns)) );
    }

    {
        // Test rule with all matches
        QList<UpdateRule> updateRules;
        for( int i = 0; i < numTestRows; ++i)
            updateRules << UpdateRule(0, QSet<int>() << i);
        QCOMPARE( m.affectedRows(updateRules).count(), numTestRows );
        for( int i = 0; i < numTestRows; ++i)
            QVERIFY( m.affectedRows(updateRules).contains(IdentityList::fromRow(m, i, idColumns)) );
    }
}

void tst_UserModifiableModel::affectedRowsWithMultipleRules()
{
    UserModifiableModel m;
    m.addColumns(0, 2);
    QList<int> idColumns;
    idColumns << 0;
    m.setIdentityColumns(idColumns);

    const int numTestRows = 10;
    m.beginLazyRowInsertions();
    for (int i = 0; i < numTestRows; ++i) {
        QVariantList dataList;
        dataList << i << QString("data %1").arg(i);
        m.lazyInsertRow(0, dataList);
    }
    m.endLazyRowInsertions();

    {
        // Test rules with no matches
        QList<UpdateRule> updateRules;
        updateRules << UpdateRule(0, QSet<int>() << -1 << numTestRows);
        updateRules << UpdateRule(0, QSet<int>() << numTestRows + 10);

        QCOMPARE( m.affectedRows(updateRules).count(), 0 );
    }

    {
        // Test 2 rules with one match each
        QList<UpdateRule> updateRules;
        updateRules << UpdateRule(0, QSet<int>() << 7);
        updateRules << UpdateRule(0, QSet<int>() << 0);
        QLinkedList<IdentityList> affected = m.affectedRows(updateRules);
        QCOMPARE( affected.count(), 2 );
        // also verify the correct order
        QCOMPARE( affected.takeFirst(), IdentityList::fromRow(m, 0, idColumns));
        QCOMPARE( affected.takeFirst(), IdentityList::fromRow(m, 7, idColumns));
    }

    {
        // Test 2 rules with 2 matches, partially overlapping
        QList<UpdateRule> updateRules;
        updateRules << UpdateRule(0, QSet<int>() << 1 << 3);
        updateRules << UpdateRule(0, QSet<int>() << 2 << 3);
        QLinkedList<IdentityList> affected = m.affectedRows(updateRules);
        QCOMPARE( affected.count(), 3 );

        QCOMPARE( affected.takeFirst(), IdentityList::fromRow(m, 1, idColumns));
        QCOMPARE( affected.takeFirst(), IdentityList::fromRow(m, 2, idColumns));
        QCOMPARE( affected.takeFirst(), IdentityList::fromRow(m, 3, idColumns));
    }
}

QTEST_MAIN( tst_UserModifiableModel )
#include "tst_live_usermodifiablemodel.moc"
