/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "../utils/livetesthelpers.h"
#include "../utils/messagehandler.h"

#include "TrackerLiveQuery"

#include <QtTest/QtTest>
#include <QVariant>
#include <QtSparql/QtSparql>

class tst_TrackerLiveWarnings : public QObject
{
    Q_OBJECT

public:
    tst_TrackerLiveWarnings();
    virtual ~tst_TrackerLiveWarnings();

public slots:
    void initTestCase();
    void cleanupTestCase();
    void init();
    void cleanup();

private slots:
    void LiveQuery_constructor_should_not_emit_warnings_for_incorrect_column_count();

    void LiveQuery_constructor_should_not_emit_warning_for_identity_columns_with_indexes_in_range();
    void LiveQuery_constructor_should_not_emit_warning_for_empty_identity_columns();
    void LiveQuery_constructor_should_emit_warning_for_identity_columns_with_indexes_out_of_range();

    void LiveQuery_setIdentityColumns_should_not_emit_warning_for_empty_list();
    void LiveQuery_setIdentityColumns_should_not_emit_warning_for_list_with_indexes_in_range();
    void LiveQuery_setIdentityColumns_should_emit_warning_for_list_with_indexes_out_of_range();

    void LiveQuery_setCollationColumns_should_not_emit_warning_for_empty_list();
    void LiveQuery_setCollationColumns_should_not_emit_warning_for_list_with_indexes_in_range();
    void LiveQuery_setCollationColumns_should_emit_warning_for_list_with_indexes_out_of_range();

    void LiveQuery_start_should_not_emit_warning_when_at_least_one_updater_has_been_added();
    void LiveQuery_start_should_emit_warning_when_no_updaters_have_been_added();
    void LiveQuery_start_should_not_emit_warning_when_at_least_one_identity_column_has_been_defined();
    void LiveQuery_start_should_emit_warning_when_no_identity_columns_have_been_defined();

    void TrackerPartialUpdater_constructor_should_not_emit_warning_when_one_filter_tag_is_present();
    void TrackerPartialUpdater_constructor_should_emit_warning_when_filter_tag_is_missing();
    void TrackerPartialUpdater_constructor_should_emit_warning_when_more_than_one_filter_tag();

    void TrackerPartialUpdater_watchClass_should_not_emit_warning_for_nonempty_class();
    void TrackerPartialUpdater_watchClass_should_emit_warning_for_empty_class();

    void TrackerPartialUpdater_watchClass_should_not_emit_warning_when_one_list_tag_is_present();
    void TrackerPartialUpdater_watchClass_should_emit_warning_when_list_tag_is_missing();
    void TrackerPartialUpdater_watchClass_should_emit_warning_when_more_than_one_list_tag();

    void TrackerLiveQuery_addUpdater_should_emit_warning_for_partial_updater_without_watchers();
    void TrackerLiveQuery_addUpdater_should_emit_warning_for_full_updater_without_watchers();

    void TrackerLiveQuery_addUpdater_should_not_emit_warning_when_id_column_is_in_range();
    void TrackerLiveQuery_addUpdater_should_emit_warning_when_id_column_is_out_of_range();

    void TrackerFullUpdater_watchClass_should_not_emit_warning_for_nonempty_class();
    void TrackerFullUpdater_watchClass_should_emit_warning_for_empty_class();

private:
    QSparqlConnection* conn;
    MessageRecorder* msgRecorder;
};

namespace {

    const char* const validUpdateQuery =
        "select tracker:id(?e) ?s "
        "{ ?e a nmo:Email ; "
        "nie:isLogicalPartOf <tracker-live-tests> ;"
        "nmo:messageSubject ?s . "
        "%FILTER "
        "}";

    const char* const validOriginalQuery =
        "select tracker:id(?e) ?s "
        "{ ?e a nmo:Email ; "
        "nie:isLogicalPartOf <tracker-live-tests> ;"
        "nmo:messageSubject ?s . "
        "}";
    const int validOriginalQueryColCount = 2;
    const QList<int> validOriginalIdentityColumns = QList<int>() << 0;
}

tst_TrackerLiveWarnings::tst_TrackerLiveWarnings()
    : conn(0), msgRecorder(0)
{
}

tst_TrackerLiveWarnings::~tst_TrackerLiveWarnings()
{
}

void tst_TrackerLiveWarnings::initTestCase()
{
}

void tst_TrackerLiveWarnings::cleanupTestCase()
{
}

void tst_TrackerLiveWarnings::init()
{
    conn = new QSparqlConnection("QTRACKER_DIRECT");
    msgRecorder = new MessageRecorder;
    msgRecorder->addMsgTypeToRecord(QtWarningMsg);
}

void tst_TrackerLiveWarnings::cleanup()
{
    delete msgRecorder;
    msgRecorder = 0;

    delete conn;
    conn = 0;
}

void tst_TrackerLiveWarnings::LiveQuery_constructor_should_not_emit_warnings_for_incorrect_column_count()
{
    const int incorrectQueryColCount = validOriginalQueryColCount + 1;

    TrackerLiveQuery lq(validOriginalQuery, incorrectQueryColCount, *conn);

    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_constructor_should_not_emit_warning_for_empty_identity_columns()
{
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, QList<int>(), *conn);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_constructor_should_not_emit_warning_for_identity_columns_with_indexes_in_range()
{
    QList<int> idCols;
    for (int col = 0; col < validOriginalQueryColCount; ++col) {
        idCols << col;
    }
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, idCols, *conn);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_constructor_should_emit_warning_for_identity_columns_with_indexes_out_of_range()
{
    QString expectedMsgFormat("TrackerLiveQuery: identity column index %0 is out of range for live query: \"%1\"");
    QList<int> idCols;
    idCols << 0 << -1;
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount,
                        idCols, *conn);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(-1).arg(validOriginalQuery).simplified() );

    idCols.clear();
    idCols << 0 << validOriginalQueryColCount << -1;
    TrackerLiveQuery lq2(validOriginalQuery, validOriginalQueryColCount,
                        idCols, *conn);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQueryColCount).arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::LiveQuery_setIdentityColumns_should_not_emit_warning_for_empty_list()
{
    // Note: this test tests the deprecated API's; remove the test when the
    // deprecated functions are removed.
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QList<int> idCols;
    lq.setIdentityColumns(idCols);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_setIdentityColumns_should_not_emit_warning_for_list_with_indexes_in_range()
{
    // Note: this test tests the deprecated API's; remove the test when the
    // deprecated functions are removed.
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QList<int> idCols;
    for (int col = 0; col < validOriginalQueryColCount; ++col) {
        idCols << col;
    }
    lq.setIdentityColumns(idCols);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_setIdentityColumns_should_emit_warning_for_list_with_indexes_out_of_range()
{
    // Note: this test tests the deprecated API's; remove the test when the
    // deprecated functions are removed.
    QString expectedMsgFormat("TrackerLiveQuery: identity column index %0 is out of range for live query: \"%1\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    QList<int> idCols;
    idCols << 0 << -1;
    lq.setIdentityColumns(idCols);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(-1).arg(validOriginalQuery).simplified() );

    idCols.clear();
    idCols << 0 << validOriginalQueryColCount << -1;
    lq.setIdentityColumns(idCols);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQueryColCount).arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::LiveQuery_setCollationColumns_should_not_emit_warning_for_empty_list()
{
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QList<TrackerLiveQuery::CollationColumn> collationColumns;
    lq.setCollationColumns(collationColumns);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_setCollationColumns_should_not_emit_warning_for_list_with_indexes_in_range()
{
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QList<TrackerLiveQuery::CollationColumn> collationColumns;
    for (int col = 0; col < validOriginalQueryColCount; ++col) {
        collationColumns << TrackerLiveQuery::CollationColumn(col, QVariant::Int);
    }
    lq.setCollationColumns(collationColumns);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_setCollationColumns_should_emit_warning_for_list_with_indexes_out_of_range()
{
    QString expectedMsgFormat("TrackerLiveQuery: collation column index %0 is out of range for live query: \"%1\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    QList<TrackerLiveQuery::CollationColumn> collationColumns;
    collationColumns
            << TrackerLiveQuery::CollationColumn( 0, QVariant::Int)
            << TrackerLiveQuery::CollationColumn(-1, QVariant::Int);
    lq.setCollationColumns(collationColumns);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(-1).arg(validOriginalQuery).simplified() );

    collationColumns.clear();
    collationColumns
            << TrackerLiveQuery::CollationColumn( 0, QVariant::Int)
            << TrackerLiveQuery::CollationColumn( validOriginalQueryColCount, QVariant::Int)
            << TrackerLiveQuery::CollationColumn(-1, QVariant::Int);
    lq.setCollationColumns(collationColumns);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQueryColCount).arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::LiveQuery_start_should_not_emit_warning_when_at_least_one_updater_has_been_added()
{
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    TrackerPartialUpdater updater(validUpdateQuery);
    updater.watchClass("nmo:Message", QStringList(),
                       "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    lq.addUpdater(updater);
    lq.setIdentityColumns(validOriginalIdentityColumns);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    lq.start();
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_start_should_emit_warning_when_no_updaters_have_been_added()
{
    const QString expectedMsgFormat("TrackerLiveQuery: Cannot start because updaters not set for live query: \"%0\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    lq.start();
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::LiveQuery_start_should_not_emit_warning_when_at_least_one_identity_column_has_been_defined()
{
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    TrackerPartialUpdater updater(validUpdateQuery);
    updater.watchClass("nmo:Message", QStringList(),
                       "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    lq.addUpdater(updater);
    lq.setIdentityColumns(validOriginalIdentityColumns);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    lq.start();
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::LiveQuery_start_should_emit_warning_when_no_identity_columns_have_been_defined()
{
    const QString expectedMsgFormat("TrackerLiveQuery: Cannot start because identity columns not set for live query: \"%0\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    TrackerPartialUpdater updater(validUpdateQuery);
    updater.watchClass("nmo:Message", QStringList(),
                       "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    lq.addUpdater(updater);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    lq.start();
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_constructor_should_not_emit_warning_when_one_filter_tag_is_present()
{
    QString updateQuery("%FILTER%");
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    TrackerPartialUpdater up1(updateQuery);

    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    updateQuery = ("select tracker:id(?e) ?s "
                   "{ ?e a nmo:Email ; "
                   "nie:isLogicalPartOf <tracker-live-tests> ;"
                   "nmo:messageSubject ?s . "
                   "%FILTER "
                   "}");

    TrackerPartialUpdater up2(updateQuery);

    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_constructor_should_emit_warning_when_filter_tag_is_missing()
{
    QString updateQuery;
    const QString expectedMsgFormat("TrackerLiveQuery: %FILTER missing in update query: \"%0\"");
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    TrackerPartialUpdater up1(updateQuery);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(updateQuery).simplified() );

    updateQuery = ("select tracker:id(?e) ?s "
                   "{ ?e a nmo:Email ; "
                   "nie:isLogicalPartOf <tracker-live-tests> ;"
                   "nmo:messageSubject ?s . "
                   "}");

    TrackerPartialUpdater up2(updateQuery);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(updateQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_constructor_should_emit_warning_when_more_than_one_filter_tag()
{
    const QString expectedMsgFormat("TrackerLiveQuery: more than one %FILTER found in update query: \"%0\"");
    QString updateQuery("%FILTER%FILTER");
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    TrackerPartialUpdater up1(updateQuery);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(updateQuery).simplified() );

    updateQuery = ("select tracker:id(?e) ?s "
                   "{ ?e a nmo:Email ; "
                   "nie:isLogicalPartOf <tracker-live-tests> ;"
                   "%FILTER"
                   "nmo:messageSubject ?s . "
                   "%FILTER "
                   "}");

    TrackerPartialUpdater up2(updateQuery);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(updateQuery).simplified()  );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_watchClass_should_not_emit_warning_for_nonempty_class()
{
    TrackerPartialUpdater up(validUpdateQuery);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("nmo:Message", QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_watchClass_should_emit_warning_for_empty_class()
{
    const QString expectedMsgFormat("TrackerLiveQuery: class to watch is empty for update query: \"%0\"");
    TrackerPartialUpdater up(validUpdateQuery);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("", QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validUpdateQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_watchClass_should_not_emit_warning_when_one_list_tag_is_present()
{
    TrackerPartialUpdater up(validUpdateQuery);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("nmo:Message", QStringList(),
                   "%LIST", TrackerPartialUpdater::Subject, 0);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("nmo:Message", QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, 0);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_watchClass_should_emit_warning_when_list_tag_is_missing()
{
    const QString expectedMsgFormat("TrackerLiveQuery: %LIST missing in filter snippet: \"%0\"");
    TrackerPartialUpdater up(validUpdateQuery);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    QString filterSnippet;
    up.watchClass("nmo:Message", QStringList(),
                   filterSnippet, TrackerPartialUpdater::Subject, 0);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(filterSnippet).simplified() );

    filterSnippet = "tracker:id(?e) in $LIST";
    up.watchClass("nmo:Message", QStringList(),
                   filterSnippet, TrackerPartialUpdater::Subject, 0);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(filterSnippet).simplified() );
}

void tst_TrackerLiveWarnings::TrackerPartialUpdater_watchClass_should_emit_warning_when_more_than_one_list_tag()
{
    const QString expectedMsgFormat("TrackerLiveQuery: more than one %LIST found in filter snippet: \"%0\"");
    TrackerPartialUpdater up(validUpdateQuery);
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    QString filterSnippet = "%LIST%LIST";
    up.watchClass("nmo:Message", QStringList(),
                   filterSnippet, TrackerPartialUpdater::Subject, 0);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(filterSnippet).simplified() );

    filterSnippet = "tracker:id(?e) in %LIST or %LIST";
    up.watchClass("nmo:Message", QStringList(),
                   filterSnippet, TrackerPartialUpdater::Subject, 0);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(filterSnippet).simplified() );
}

void tst_TrackerLiveWarnings::TrackerLiveQuery_addUpdater_should_emit_warning_for_partial_updater_without_watchers()
{
    const QString expectedMsgFormat("TrackerLiveQuery: No class watchers found for update query: \"%0\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    TrackerPartialUpdater partialUpdater(validUpdateQuery);

    lq.addUpdater(partialUpdater);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validUpdateQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerLiveQuery_addUpdater_should_emit_warning_for_full_updater_without_watchers()
{
    const QString expectedMsgFormat("TrackerLiveQuery: No class watchers found for full updater of live query: \"%0\"");
    TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
    TrackerFullUpdater fullUpdater;

    lq.addUpdater(fullUpdater);

    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsgFormat.arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerLiveQuery_addUpdater_should_not_emit_warning_when_id_column_is_in_range()
{
    for (int validIdCol = -1; validIdCol < validOriginalQueryColCount; ++validIdCol) {
        TrackerPartialUpdater up(validUpdateQuery);
        up.watchClass("nmo:Message", QStringList(),
                       "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, validIdCol);

        TrackerLiveQuery lq(validOriginalQuery, validOriginalQueryColCount, *conn);
        lq.addUpdater(up);
        QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
    }
}

void tst_TrackerLiveWarnings::TrackerLiveQuery_addUpdater_should_emit_warning_when_id_column_is_out_of_range()
{
    const QString expectedMsgFormat(
           "TrackerLiveQuery: updater id column %0 for class \"%1\" is out of range for live query: \"%2\"");
    const QString klass("nmo:Message");

    TrackerPartialUpdater up1(validUpdateQuery);
    int idColumn = -2;
    up1.watchClass(klass, QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, idColumn);
    TrackerLiveQuery lq1(validOriginalQuery, validOriginalQueryColCount, *conn);
    lq1.addUpdater(up1);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(),
              expectedMsgFormat.arg(idColumn).arg(klass).arg(validOriginalQuery).simplified() );

    TrackerPartialUpdater up2(validUpdateQuery);
    idColumn = validOriginalQueryColCount;
    up2.watchClass(klass, QStringList(),
                   "tracker:id(?e) in %LIST", TrackerPartialUpdater::Subject, idColumn);
    TrackerLiveQuery lq2(validOriginalQuery, validOriginalQueryColCount, *conn);
    lq2.addUpdater(up2);
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 2 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(),
              expectedMsgFormat.arg(idColumn).arg(klass).arg(validOriginalQuery).simplified() );
}

void tst_TrackerLiveWarnings::TrackerFullUpdater_watchClass_should_not_emit_warning_for_nonempty_class()
{
    TrackerFullUpdater up;
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("nmo:Message", QStringList());
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );
}

void tst_TrackerLiveWarnings::TrackerFullUpdater_watchClass_should_emit_warning_for_empty_class()
{
    const QString expectedMsg("TrackerLiveQuery: class to watch is empty.");
    TrackerFullUpdater up;
    QVERIFY( (*msgRecorder)[QtWarningMsg].isEmpty() );

    up.watchClass("", QStringList());
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].back().simplified(), expectedMsg.simplified() );
    QCOMPARE( (*msgRecorder)[QtWarningMsg].count(), 1 );
}


QTEST_MAIN( tst_TrackerLiveWarnings )
#include "tst_live_warnings.moc"

