/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QString>
#include <QtTest>
#include "signalsrecorder.h"
#include <QStandardItemModel>

Q_DECLARE_METATYPE(QModelIndex)

namespace
{

    template<typename C, typename P>
    C qSorted(const C& c, P p)
    {
        C result(c);
        qSort(result.begin(), result.end(), p);
        return result;
    }

    int nextTestArg(int testArg, int sigNum) { return 31*testArg + sigNum; }

}

class TestSignalEmitter : public QObject
{
    Q_OBJECT

public:
    explicit TestSignalEmitter(QObject* parent=0) : QObject(parent)
    {
        connect(this, SIGNAL(postSignal0()), this, SIGNAL(testSignal0()), Qt::QueuedConnection);
    }

    void emitTestSignal0() { emit testSignal0(); }
    void postTestSignal0() { emit postSignal0(); }

Q_SIGNALS:
    void testSignal0();
    void postSignal0();
};

class tst_SignalsRecorder : public QObject
{
    Q_OBJECT

    SignalsRecorder* recorder;
    TestSignalEmitter* testSignalEmitter;

    static void testMsgHandler(QtMsgType, const char *);

public:
    tst_SignalsRecorder();

Q_SIGNALS:
    void testSignal0();
    void postSignal0();
    void testSignal1(int);

private Q_SLOTS:
    void init();
    void cleanup();
    void testInitialState();
    void testOneSignal();
    void testOneSignalWithParam();
    void testTwoSignals();
    void testTwoSenders();
    void testMultipleConnectsToSameSignal();
    void testConnectNonExistentSignal();
    void testWaitSignalOne();
    void testWaitSignalMany();
    void testWaitMultipleSignals();
    void testWaitEmpty();
    void testWaitTimeout();
    void testClearSignalData();
    void testCompareNullSignalData();
    void testCompareNonNullSignalData();
    void testSignalDataDebugOutput();
    };

void tst_SignalsRecorder::testMsgHandler(QtMsgType, const char *)
{ }

tst_SignalsRecorder::tst_SignalsRecorder() :
    recorder(0), testSignalEmitter(0)
{
    connect(this, SIGNAL(postSignal0()), this, SIGNAL(testSignal0()), Qt::QueuedConnection);
    qRegisterMetaType<QModelIndex>("QModelIndex");
}

void tst_SignalsRecorder::init()
{
    recorder = new SignalsRecorder;
    testSignalEmitter = new TestSignalEmitter;
}

void tst_SignalsRecorder::cleanup()
{
    delete testSignalEmitter;
    testSignalEmitter = 0;
    delete recorder;
    recorder = 0;
}

void tst_SignalsRecorder::testInitialState()
{
    QCOMPARE( recorder->allSignalData().count(), 0 );
}

void tst_SignalsRecorder::testOneSignal()
{
    QVERIFY( recorder->listen(this, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );

    const int sigCount = 10;
    for( int sig=1; sig <= sigCount; ++sig ) {
        emit testSignal0();
        QCOMPARE( recorder->allSignalData().count(), sig );
        const QList<SignalsRecorder::SignalData> sd = recorder->signalData(this, SIGNAL(testSignal0()));
        QCOMPARE( sd.count(), sig );
    }

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), sigCount );

    const QList<SignalsRecorder::SignalData> sd = recorder->signalData(this, SIGNAL(testSignal0()));
    QCOMPARE( sd.count(), sigCount );

    for( int i=0; i < sigCount; ++i ) {
        QVERIFY( sd[i].matches(this, SIGNAL(testSignal0())) );
        QCOMPARE( sd[i].params().count(), 0 );
        if (i > 0) QVERIFY( sd[i-1].seqId() < sd[i].seqId() );
    }
    QCOMPARE( ad, sd );
    QCOMPARE( sd, recorder->signalData(SignalList().append(this, SIGNAL(testSignal0()))) );
}

void tst_SignalsRecorder::testOneSignalWithParam()
{
    QVERIFY( recorder->listen(this, SIGNAL(testSignal1(int))) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal1(int))).count(), 0 );

    const int sigCount = 10;
    for( int sig=1, testArg = 1; sig <= sigCount; ++sig, testArg = nextTestArg(testArg, sig) ) {
        emit testSignal1(testArg);
        QCOMPARE( recorder->allSignalData().count(), sig );
        const QList<SignalsRecorder::SignalData> sd = recorder->signalData(this, SIGNAL(testSignal1(int)));
        QCOMPARE( sd.count(), sig );
    }

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), sigCount );

    const QList<SignalsRecorder::SignalData> sd = recorder->signalData(this, SIGNAL(testSignal1(int)));
    QCOMPARE( sd.count(), sigCount );

    for( int i=0, testArg = 1; i < sigCount; ++i, testArg = nextTestArg(testArg, i+1) ) {
        QVERIFY( sd[i].matches(this, SIGNAL(testSignal1(int))) );
        QCOMPARE( sd[i].params().first().toInt(), testArg );
        if (i > 0) QVERIFY( sd[i-1].seqId() < sd[i].seqId() );
    }
    QCOMPARE( ad, sd );
    QCOMPARE( sd, recorder->signalData(SignalList().append(this, SIGNAL(testSignal1(int)))) );
}

void tst_SignalsRecorder::testTwoSignals()
{
    QVERIFY( recorder->listen(this, SIGNAL(testSignal0())) );
    QVERIFY( recorder->listen(this, SIGNAL(testSignal1(int))) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal1(int))).count(), 0 );

    const int sigCount = 10;
    for( int sig=1, testArg = 1; sig <= sigCount; ++sig, testArg = nextTestArg(testArg, sig) ) {
        emit testSignal0();
        const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
        QCOMPARE( sd0.count(), sig );
        emit testSignal1(testArg);
        const QList<SignalsRecorder::SignalData> sd1 = recorder->signalData(this, SIGNAL(testSignal1(int)));
        QCOMPARE( sd1.count(), sig );
        QCOMPARE( recorder->allSignalData().count(), 2*sig );
    }

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), 2*sigCount );

    const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
    QCOMPARE( sd0.count(), sigCount );

    const QList<SignalsRecorder::SignalData> sd1 = recorder->signalData(this, SIGNAL(testSignal1(int)));
    QCOMPARE( sd1.count(), sigCount );

    for( int i=0, testArg = 1; i < sigCount; ++i, testArg = nextTestArg(testArg, i+1) ) {
        QVERIFY( sd0[i].matches(this, SIGNAL(testSignal0())) );
        QCOMPARE( sd0[i].params().count(), 0 );
        QVERIFY( sd1[i].matches(this, SIGNAL(testSignal1(int))) );
        QCOMPARE( sd1[i].params().first().toInt(), testArg );
        if (i > 0) {
            QVERIFY( sd0[i-1].seqId() < sd0[i].seqId() );
            QVERIFY( sd1[i-1].seqId() < sd1[i].seqId() );
            QVERIFY( sd0[i-1].seqId() < sd1[i].seqId() );
        }
    }
    QCOMPARE( ad, qSorted(sd0 + sd1, sequenceOrdering) );
    SignalList expectedSignals;
    expectedSignals
            .append(this, SIGNAL(testSignal0()))
            .append(this, SIGNAL(testSignal1(int)));
    QCOMPARE( ad, recorder->signalData(expectedSignals) );
}

void tst_SignalsRecorder::testTwoSenders()
{
    QVERIFY( recorder->listen(this, SIGNAL(testSignal0())) );
    QVERIFY( recorder->listen(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );
    QCOMPARE( recorder->signalData(testSignalEmitter, SIGNAL(testSignal0())).count(), 0 );

    const int sigCount = 10;
    for( int sig=1; sig <= sigCount; ++sig ) {
        emit testSignal0();
        const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
        QCOMPARE( sd0.count(), sig );
        emit testSignalEmitter->emitTestSignal0();
        const QList<SignalsRecorder::SignalData> sd1 = recorder->signalData(testSignalEmitter, SIGNAL(testSignal0()));
        QCOMPARE( sd1.count(), sig );
        QCOMPARE( recorder->allSignalData().count(), 2*sig );
    }

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), 2*sigCount );

    const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
    QCOMPARE( sd0.count(), sigCount );

    const QList<SignalsRecorder::SignalData> sd1 = recorder->signalData(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( sd1.count(), sigCount );

    for( int i=0; i < sigCount; ++i ) {
        QVERIFY( sd0[i].matches(this, SIGNAL(testSignal0())) );
        QCOMPARE( sd0[i].params().count(), 0 );
        QVERIFY( sd1[i].matches(testSignalEmitter, SIGNAL(testSignal0())) );
        QCOMPARE( sd1[i].params().count(), 0 );
        if (i > 0) {
            QVERIFY( sd0[i-1].seqId() < sd0[i].seqId() );
            QVERIFY( sd1[i-1].seqId() < sd1[i].seqId() );
            QVERIFY( sd0[i-1].seqId() < sd1[i].seqId() );
        }
    }
    QCOMPARE( ad, qSorted(sd0 + sd1, sequenceOrdering) );
    SignalList expectedSignals;
    expectedSignals
            .append(this, SIGNAL(testSignal0()))
            .append(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( ad, recorder->signalData(expectedSignals) );
}

void tst_SignalsRecorder::testMultipleConnectsToSameSignal()
{
    QVERIFY( recorder->listen(this, SIGNAL(testSignal0())) );
    QVERIFY( recorder->listen(this, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );

    emit testSignal0();

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), 2 );

    const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
    QCOMPARE( sd0.count(), 2 );

    for( int i=0; i < 2; ++i) {
        QVERIFY( sd0[i].matches(this, SIGNAL(testSignal0())) );
        QCOMPARE( sd0[i].params().count(), 0 );
        if (i > 0) {
            QVERIFY( sd0[i-1].seqId() < sd0[i].seqId() );
        }
    }
    QCOMPARE( ad, sd0 );
    QCOMPARE( sd0, recorder->signalData(SignalList().append(this, SIGNAL(testSignal0()))) );
}

void tst_SignalsRecorder::testConnectNonExistentSignal()
{
    // Suppress warning messages from test log
    struct ScopedTestMsgHandler {
        QtMsgHandler prevHandler;
        ScopedTestMsgHandler() : prevHandler(qInstallMsgHandler(&testMsgHandler)) { }
        ~ScopedTestMsgHandler() { qInstallMsgHandler(prevHandler); }
    } scopedMsgHandler;

    QVERIFY( !recorder->listen(this, SIGNAL(nonExistentSignal())) );
    QVERIFY( !recorder->listen(this, SIGNAL(nonExistentSignal(int))) );

    SignalList signalList;
    signalList
            .append(this, SIGNAL(testSignal0()))
            .append(this, SIGNAL(nonExistentSignal()));
    QVERIFY( !recorder->listen(signalList) );
    // Verify that none of the signals were connected if there was even one invalid signal
    emit testSignal0();
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(signalList).count(), 0 );
}

void tst_SignalsRecorder::testWaitSignalOne()
{
    QVERIFY( recorder->listen(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );

    SignalList expectedSignals;
    expectedSignals.append(testSignalEmitter, SIGNAL(testSignal0()), 1);
    testSignalEmitter->postTestSignal0();
    QCOMPARE( recorder->allSignalData().count(), 0 );

    QVERIFY( recorder->waitForSignals(expectedSignals, 100) );

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), 1 );

    const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( sd0.count(), 1 );
    QVERIFY( sd0[0].matches(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( sd0[0].params().count(), 0 );

    QCOMPARE( ad, sd0 );
    QCOMPARE( sd0, recorder->signalData(SignalList().append(testSignalEmitter, SIGNAL(testSignal0()))) );
}

void tst_SignalsRecorder::testWaitSignalMany()
{
    QVERIFY( recorder->listen(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );

    const int signalCount = 3;
    SignalList expectedSignals;
    expectedSignals.append(testSignalEmitter, SIGNAL(testSignal0()), signalCount);
    for (int i=0; i < signalCount; ++i)
        testSignalEmitter->postTestSignal0();
    QCOMPARE( recorder->allSignalData().count(), 0 );

    QVERIFY( recorder->waitForSignals(expectedSignals, 100) );
    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), signalCount );
    const QList<SignalsRecorder::SignalData> sd = recorder->signalData(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( sd.count(), signalCount );
    for(int i = 0; i < signalCount; ++i) {
        QVERIFY( sd[i].matches(testSignalEmitter, SIGNAL(testSignal0())) );
        QCOMPARE( sd[i].params().count(), 0 );
        if (i > 0) {
            QVERIFY( sd[i-1].seqId() < sd[i].seqId() );
        }
    }
    QCOMPARE( ad, sd );
    QCOMPARE( sd, recorder->signalData(SignalList().append(testSignalEmitter, SIGNAL(testSignal0()))) );
}

void tst_SignalsRecorder::testWaitMultipleSignals()
{
    SignalList expectedSignals;
    expectedSignals
            .append(this, SIGNAL(testSignal0()))
            .append(testSignalEmitter, SIGNAL(testSignal0()));
    QVERIFY( recorder->listen(expectedSignals) );
    QCOMPARE( recorder->allSignalData().count(), 0 );

    expectedSignals.append(testSignalEmitter, SIGNAL(testSignal0()));
    emit postSignal0();  // posts testSignal0
    testSignalEmitter->postTestSignal0();
    testSignalEmitter->postTestSignal0();
    QCOMPARE( recorder->allSignalData().count(), 0 );

    QVERIFY( recorder->waitForSignals(expectedSignals, 100) );

    const QList<SignalsRecorder::SignalData> ad = recorder->allSignalData();
    QCOMPARE( ad.count(), 3 );

    const QList<SignalsRecorder::SignalData> sd0 = recorder->signalData(this, SIGNAL(testSignal0()));
    QCOMPARE( sd0.count(), 1 );
    QVERIFY( sd0[0].matches(this, SIGNAL(testSignal0())) );
    QCOMPARE( sd0[0].params().count(), 0 );

    const QList<SignalsRecorder::SignalData> sd1 = recorder->signalData(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( sd1.count(), 2 );
    QVERIFY( sd1[0].matches(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( sd1[0].params().count(), 0 );
    QVERIFY( sd0[0].seqId() < sd1[0].seqId() );
    QVERIFY( sd1[1].matches(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( sd1[1].params().count(), 0 );
    QVERIFY( sd1[0].seqId() < sd1[1].seqId() );

    QCOMPARE( ad, qSorted(sd0+sd1, sequenceOrdering) );
    QCOMPARE( ad, recorder->signalData(expectedSignals) );
}

void tst_SignalsRecorder::testWaitEmpty()
{
    QVERIFY( recorder->listen(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );

    SignalList expectedSignals;
    QCOMPARE( recorder->allSignalData().count(), 0 );

    QVERIFY( recorder->waitForSignals(expectedSignals, 20) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
}

void tst_SignalsRecorder::testWaitTimeout()
{
    QVERIFY( recorder->listen(testSignalEmitter, SIGNAL(testSignal0())) );
    QCOMPARE( recorder->allSignalData().count(), 0 );

    SignalList expectedSignals;
    expectedSignals.append(testSignalEmitter, SIGNAL(testSignal0()), 1);
    QCOMPARE( recorder->allSignalData().count(), 0 );

    QVERIFY( !recorder->waitForSignals(expectedSignals, 20) );
    QCOMPARE( recorder->allSignalData().count(), 0 );
}

void tst_SignalsRecorder::testClearSignalData()
{
    recorder->clearAllSignalData();
    QCOMPARE( recorder->allSignalData().count(), 0 );

    recorder->clearSignalData(this, SIGNAL(testSignal0()));
    QCOMPARE( recorder->allSignalData().count(), 0 );

    SignalList signalsToClear;
    signalsToClear
            .append(this, SIGNAL(testSignal1(int)))
            .append(this, SIGNAL(testSignal0()));
    recorder->clearSignalData(signalsToClear);
    QCOMPARE( recorder->allSignalData().count(), 0 );

    recorder->listen(this, SIGNAL(testSignal0()));
    recorder->clearSignalData(this, SIGNAL(testSignal0()));
    QCOMPARE( recorder->allSignalData().count(), 0 );

    emit testSignal0();
    emit testSignal0();
    QCOMPARE( recorder->allSignalData().count(), 2 );

    recorder->clearSignalData(this, SIGNAL(testSignal0()));
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );

    recorder->listen(this, SIGNAL(testSignal1(int)));
    recorder->listen(testSignalEmitter, SIGNAL(testSignal0()));
    emit testSignal0();
    emit testSignal0();
    emit testSignal1(42);
    testSignalEmitter->emitTestSignal0();
    QCOMPARE( recorder->allSignalData().count(), 4 );

    recorder->clearSignalData(this, SIGNAL(testSignal0()));
    QCOMPARE( recorder->allSignalData().count(), 2 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal0())).count(), 0 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal1(int))).count(), 1 );
    QCOMPARE( recorder->signalData(testSignalEmitter, SIGNAL(testSignal0())).count(), 1 );

    recorder->clearSignalData(this, SIGNAL(testSignal1(int)));
    QCOMPARE( recorder->allSignalData().count(), 1 );
    QCOMPARE( recorder->signalData(this, SIGNAL(testSignal1(int))).count(), 0 );
    QCOMPARE( recorder->signalData(testSignalEmitter, SIGNAL(testSignal0())).count(), 1 );

    recorder->clearSignalData(testSignalEmitter, SIGNAL(testSignal0()));
    QCOMPARE( recorder->allSignalData().count(), 0 );
    QCOMPARE( recorder->signalData(testSignalEmitter, SIGNAL(testSignal0())).count(), 0 );

    emit testSignal0();
    emit testSignal0();
    emit testSignal1(42);
    testSignalEmitter->emitTestSignal0();
    QCOMPARE( recorder->allSignalData().count(), 4 );

    recorder->clearSignalData(signalsToClear);
    QCOMPARE( recorder->signalData(signalsToClear).count(), 0 );
    QCOMPARE( recorder->signalData(testSignalEmitter, SIGNAL(testSignal0())).count(), 1 );

    recorder->clearAllSignalData();
    QCOMPARE( recorder->allSignalData().count(), 0 );
}

void tst_SignalsRecorder::testCompareNullSignalData()
{
    const SignalsRecorder::SignalData sd1(0, QByteArray(), QList<QVariant>(), 0);
    const SignalsRecorder::SignalData sd2(0, QByteArray(), QList<QVariant>(), 0);
    QCOMPARE( sd1, sd2 );
    QCOMPARE( sd2, sd1 );
}

void tst_SignalsRecorder::testCompareNonNullSignalData()
{
    QStandardItemModel model(2,3);

    QObject* const object = this;
    const QByteArray sig("testSignal1(QModelIndex,QModelIndex,int,int)");
    const QList<QVariant> params(QList<QVariant>()
                                    << QVariant::fromValue<QModelIndex>(QModelIndex())
                                    << QVariant::fromValue<QModelIndex>(model.index(1,2))
                                    << QVariant(3)
                                    << QVariant(4));
    const int seqid = 5;

    const SignalsRecorder::SignalData sd1(object, sig, params, seqid);
    const SignalsRecorder::SignalData sd2(object, sig, params, seqid);
    const SignalsRecorder::SignalData sdNull(0, QByteArray(), QList<QVariant>(), 0);
    QCOMPARE( sd1, sd2 );
    QCOMPARE( sd2, sd1 );
    QVERIFY( sd1 != sdNull );
    QVERIFY( sdNull != sd2 );
}

void tst_SignalsRecorder::testSignalDataDebugOutput()
{
    QStandardItemModel model(2,3);

    QObject* const object = this;
    const QByteArray sig("testSignal1(QModelIndex,QModelIndex,int,int)");
    const QList<QVariant> params(QList<QVariant>()
                                    << QVariant::fromValue<QModelIndex>(QModelIndex())
                                    << QVariant::fromValue<QModelIndex>(model.index(1,2))
                                    << QVariant(3)
                                    << QVariant(4));
    const int seqid = 5;

    const SignalsRecorder::SignalData sd1(object, sig, params, seqid);
    const SignalsRecorder::SignalData sdNull(0, QByteArray(), QList<QVariant>(), 0);

    qDebug() << "Non-null SignalData:" << sd1;
    qDebug() << "Null SignalData:" << sdNull;
}

QTEST_MAIN(tst_SignalsRecorder);

#include "tst_signalsrecorder.moc"
