/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef MODELEQUATOR__H
#define MODELEQUATOR__H

#include "modeltesthelpers.h"

#include <QAbstractItemModel>
#include <QDebug>

// Helper class:
// 1) Ensures that 2 models, reference and model, emit the same signals.
// 2) Ensures that after each rows(Inserted|Removed|Moved) signal
// and dataChanged signal, the data of the model correspond to the data of the
// reference.
class ModelEquator : public QObject
{
    Q_OBJECT
public:
    ModelEquator(QAbstractItemModel* model, QAbstractItemModel* reference)
      : isGood(true), model(model), reference(reference),
      lastSender(0)
    {
        connect(model, SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                SLOT(onRowsInserted(const QModelIndex&, int, int)));
        connect(model, SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
                SLOT(onRowsRemoved(const QModelIndex&, int, int)));
        connect(model, SIGNAL(rowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
                SLOT(onRowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)));
        connect(model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
                SLOT(onDataChanged(const QModelIndex&, const QModelIndex&)));

        connect(reference, SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                SLOT(onRowsInserted(const QModelIndex&, int, int)));
        connect(reference, SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
                SLOT(onRowsRemoved(const QModelIndex&, int, int)));
        connect(reference, SIGNAL(rowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
                SLOT(onRowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)));
        connect(reference, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
                SLOT(onDataChanged(const QModelIndex&, const QModelIndex&)));
    }
    bool good() const
    {
        // check if we have any signals which are sent by only one of the models
        if (lastSender != 0) {
            qDebug() << "Pending signal:" << lastSignalName << "from"
                     << (lastSender == model ? "model" : "reference");
            return false;
        }
        return isGood;
    }
private Q_SLOTS:
    void doCheck()
    {
        if (!equalModels(model, reference)) {
            qDebug() << "Model:";
            printModel(model);
            qDebug() << "Reference:";
            printModel(reference);
            isGood = false;
        }
    }

    void storeSignal(QObject* sender, const QString& signalName,
                     const QVariantList& params)
    {
        lastSender = sender;
        lastSignalName = signalName;
        lastParams = params;
    }

    bool compareToLastSignal(const QObject* sender, const QString& signalName,
                             const QVariantList& params)
    {
        if (lastSender == sender) {
            qDebug() << "no signal corresponding to" << lastSignalName << "arrived.";
            return false;
        }
        if (lastSignalName != signalName) {
            qDebug() << "signal mismatch:" << lastSignalName << " != " << signalName;
            return false;
        }
        if (lastParams != params) {
            qDebug() << "parameter mismatch for rowsInserted:"
                     << lastParams << " != " << params;
            return false;
        }
        return true;
    }

    void clearLastSignalInfo()
    {
        lastSender = 0;
        lastParams.clear();
    }

    void onRowsInserted(const QModelIndex& parent, int start, int end)
    {
        if (parent.isValid()) {
            qDebug() << "rowsInserted with valid parent";
            isGood = false;
        }
        if (lastSender) {
            // The other model has already sent the corresponding signal; do the
            // parameter check now.
            if (!compareToLastSignal(sender(), "rowsInserted",
                                     QVariantList() << start << end))
                isGood = false;
            clearLastSignalInfo();
            doCheck();
        }
        else {
            // The other model hasn't sent the signal yet; store the data for
            // the further check.
            storeSignal(sender(), "rowsInserted",
                        QVariantList() << start << end);
        }
    }
    void onRowsRemoved(const QModelIndex& parent, int start, int end)
    {
        if (parent.isValid()) {
            qDebug() << "rowsRemoved with valid parent";
            isGood = false;
        }
        if (lastSender) {
            // The other model has already sent the corresponding signal; do the
            // parameter check now.
            if (!compareToLastSignal(sender(), "rowsRemoved",
                                     QVariantList() << start << end))
                isGood = false;
            clearLastSignalInfo();
            doCheck();
        }
        else {
            // The other model hasn't sent the signal yet; store the data for
            // the further check.
            storeSignal(sender(), "rowsRemoved",
                        QVariantList() << start << end);
        }
    }
    void onRowsMoved(const QModelIndex& sourceParent, int sourceFirst, int sourceLast,
                   const QModelIndex& destinationParent, int destinationStart)
    {
        if (sourceParent.isValid() || destinationParent.isValid()) {
            qDebug() << "rowsMoved with valid parent";
            isGood = false;
        }
        if (lastSender) {
            // The other model has already sent the corresponding signal; do the
            // parameter check now.
            if (!compareToLastSignal(sender(), "rowsMoved",
                                     QVariantList() << sourceFirst << sourceLast << destinationStart))
                isGood = false;
            clearLastSignalInfo();
            doCheck();
        }
        else {
            // The other model hasn't sent the signal yet; store the data for
            // the further check.
            storeSignal(sender(), "rowsMoved",
                        QVariantList() << sourceFirst << sourceLast << destinationStart);
        }
    }
    void onDataChanged(const QModelIndex& topLeft, const QModelIndex& bottomRight)
    {
        if (lastSender) {
            // The other model has already sent the corresponding signal; do the
            // parameter check now.
            if (!compareToLastSignal(sender(), "dataChanged",
                                     QVariantList() << topLeft.row() << topLeft.column()
                                     << bottomRight.row() << bottomRight.column()))
                isGood = false;
            clearLastSignalInfo();
            doCheck();
        }
        else {
            // The other model hasn't sent the signal yet; store the data for
            // the further check.
            storeSignal(sender(), "dataChanged",
                        QVariantList() << topLeft.row() << topLeft.column()
                        << bottomRight.row() << bottomRight.column());
        }
    }

private:
    bool isGood;
    QAbstractItemModel *model, *reference;

    QObject* lastSender; // model or reference
    QString lastSignalName;
    QVariantList lastParams;
};

#endif
