/****************************************************************************
**
** Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "modelsignalrecorder.h"

#include <QAbstractItemModel>
#include <QTest>
#include "signalsrecorder.h"
#include "modeltesthelpers.h"

Q_DECLARE_METATYPE(QModelIndex)

namespace
{

    void verifyNoSignalsReceived_impl(bool& result, SignalsRecorder* signalsRecorder)
    {
        result = false;
        QCOMPARE( signalsRecorder->allSignalData().count(), 0 );
        result = true;
    }

    void verifyModelSignalPairReceived_impl
            (bool& result, QAbstractItemModel* model, SignalsRecorder* signalsRecorder, const char* preSignal, const char* postSignal,
             const QModelIndex& parent, int start, int end,
             const QModelIndex & destinationParent, int destinationRow)
    {
        result = false;
        SignalList expectedSignals;
        expectedSignals.append(model, preSignal);
        expectedSignals.append(model, postSignal);

        QList<SignalsRecorder::SignalData> receivedSignals = signalsRecorder->signalData(expectedSignals);
        QCOMPARE( receivedSignals.count(), 2 );
        QVERIFY( receivedSignals[0].matches(model, preSignal) );
        QVERIFY( receivedSignals[1].matches(model, postSignal) );
        QVERIFY( verifySignalParameters(receivedSignals[0].params(), parent, start, end, destinationParent, destinationRow) );
        QVERIFY( verifySignalParameters(receivedSignals[1].params(), parent, start, end, destinationParent, destinationRow) );
        result = true;
    }

    void verifyDataChangedSignal_impl(bool& result, QAbstractItemModel* model, SignalsRecorder* signalsRecorder,
                                                           const QModelIndex & topLeft, const QModelIndex & bottomRight)
    {
        result = false;
        QList<SignalsRecorder::SignalData> dataChangedSignals =
                signalsRecorder->signalData(model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
        QCOMPARE( dataChangedSignals.count(), 1 );
        QCOMPARE( dataChangedSignals.back().params().count(), 2 );
        QCOMPARE( dataChangedSignals.back().params()[0].value<QModelIndex>(), topLeft );
        QCOMPARE( dataChangedSignals.back().params()[1].value<QModelIndex>(), bottomRight );
        result = true;
    }

}

ModelSignalRecorder::ModelSignalRecorder(QAbstractItemModel* model)
    : model(model), signalsRecorder(new SignalsRecorder)
{
    qRegisterMetaType<QModelIndex>("QModelIndex");
}

ModelSignalRecorder::~ModelSignalRecorder()
{
    delete signalsRecorder;
}

bool ModelSignalRecorder::verifyNoSignalsReceived() const
{
    bool result = false;
    verifyNoSignalsReceived_impl(result, signalsRecorder);
    return result;
}

bool ModelSignalRecorder::verifyModelSignalPairReceived
        (const char* preSignal, const char* postSignal,
         const QModelIndex& parent, int start, int end,
         const QModelIndex & destinationParent, int destinationRow) const
{
    bool result = false;
    verifyModelSignalPairReceived_impl(result, model, signalsRecorder, preSignal, postSignal,
                                       parent, start, end, destinationParent, destinationRow);
    return result;
}

void ModelSignalRecorder::connectInsertColumnSignals()
{
    signalsRecorder->listen(model, SIGNAL(columnsAboutToBeInserted(const QModelIndex&, int, int)));
    signalsRecorder->listen(model, SIGNAL(columnsInserted(const QModelIndex&, int, int)));
}

void ModelSignalRecorder::connectInsertRowSignals()
{
    signalsRecorder->listen(model, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)));
    signalsRecorder->listen(model, SIGNAL(rowsInserted(const QModelIndex&, int, int)));
}

void ModelSignalRecorder::connectRemoveColumnSignals()
{
    signalsRecorder->listen(model, SIGNAL(columnsAboutToBeRemoved(const QModelIndex&, int, int)));
    signalsRecorder->listen(model, SIGNAL(columnsRemoved(const QModelIndex&, int, int)));
}

void ModelSignalRecorder::connectRemoveRowSignals()
{
    signalsRecorder->listen(model, SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)));
    signalsRecorder->listen(model, SIGNAL(rowsRemoved(const QModelIndex&, int, int)));
}

void ModelSignalRecorder::connectMoveRowSignals()
{
    signalsRecorder->listen(model, SIGNAL(rowsAboutToBeMoved(const QModelIndex&, int, int, const QModelIndex&, int)));
    signalsRecorder->listen(model, SIGNAL(rowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)));
}

void ModelSignalRecorder::connectDataChangedSignal()
{
    signalsRecorder->listen(model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)));
}

bool ModelSignalRecorder::verifyInsertColumnSignalPairReceived(const QModelIndex& parent, int start, int end) const
{
    return verifyModelSignalPairReceived(
            SIGNAL(columnsAboutToBeInserted(const QModelIndex&, int, int)),
            SIGNAL(columnsInserted(const QModelIndex&, int, int)),
            parent, start, end);
}

bool ModelSignalRecorder::verifyInsertRowSignalPairReceived(const QModelIndex& parent, int start, int end) const
{
    return verifyModelSignalPairReceived(
            SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)),
            SIGNAL(rowsInserted(const QModelIndex&, int, int)),
            parent, start, end);
}

bool ModelSignalRecorder::verifyRemoveRowSignalPairReceived(const QModelIndex& parent, int start, int end) const
{
    return verifyModelSignalPairReceived(
            SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)),
            SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
            parent, start, end);
}

bool ModelSignalRecorder::verifyMoveRowSignalPairReceived
        (const QModelIndex & sourceParent, int sourceStart, int sourceEnd, const QModelIndex & destinationParent, int destinationRow) const
{
    return verifyModelSignalPairReceived(
            SIGNAL(rowsAboutToBeMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
            SIGNAL(rowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
            sourceParent, sourceStart, sourceEnd, destinationParent, destinationRow);
}

bool ModelSignalRecorder::verifyDataChangedSignalReceived(const QModelIndex & topLeft, const QModelIndex & bottomRight) const
{
    bool result = false;
    verifyDataChangedSignal_impl(result, model, signalsRecorder, topLeft, bottomRight);
    return result;
}

void ModelSignalRecorder::clearAllSignalData()
{
    signalsRecorder->clearAllSignalData();
}
