/****************************************************************************
**
** Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (ivan.frade@nokia.com)
**
** This file is part of QtSparqlTracker library.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at ivan.frade@nokia.com.
** $QT_END_LICENSE$
**
****************************************************************************/

#include "signalsrecorder.h"
#include <QCoreApplication>
#include <QModelIndex>
#include <QDebug>

Q_DECLARE_METATYPE(QModelIndex)

namespace {

	QByteArray normalizeSignature(const char *aSignal)
	{
		// From QSignalSpy::QSignalSpy(QObject*,const char*)
		Q_ASSERT(aSignal);

		if (((aSignal[0] - '0') & 0x03) != QSIGNAL_CODE) {
			qWarning("SignalsRecorder: Not a valid signal, use the SIGNAL macro");
			return QByteArray();
		}

		return QMetaObject::normalizedSignature(aSignal + 1);
	}

}


SignalList::SignalList()
{ }

SignalList& SignalList::append(QObject* object, const char* signal, int numExpected)
{
	Q_ASSERT(object);
	Q_ASSERT(signal);
	Q_ASSERT(numExpected > 0);
	QByteArray normalizedSignal(normalizeSignature(signal));
	const int index = indexOf(object, normalizedSignal);
	if (index < 0) {
		const SignalDesc sd = { object, QByteArray(signal), normalizedSignal, numExpected };
		signalDescs_.append(sd);
	}
	else {
		signalDescs_[index].numExpected += numExpected;
	}
	return *this;
}

bool SignalList::isEmpty() const
{
	return signalDescs_.isEmpty();
}

int SignalList::indexOf(QObject* object, const QByteArray& signal) const
{
	const int count = signalDescs_.count();
	for (int i=0; i < count; ++i) {
		const SignalDesc& sd = signalDescs_[i];
		if (sd.object == object && sd.signal == signal)
			return i;
	}
	return -1;
}

void SignalList::remove(QObject* object, const QByteArray& signal)
{
	const int index = indexOf(object, signal);
	if (index >= 0 && --(signalDescs_[index].numExpected) == 0)
		signalDescs_.removeAt(index);
}

QList<SignalList::SignalDesc> SignalList::asList() const
{
	return signalDescs_;
}

class SignalsRecorder::SignalsRecorderSpy : public QSignalSpy
{
public:
	SignalsRecorderSpy(SignalsRecorder* recorder_, QObject *obj, const char *aSignal)
		: QSignalSpy(obj, aSignal)
		, recorder(recorder_)
		, object(obj)  // Store object because QSignalSpy does not provide access to it
	{ }

	bool hasObject(QObject* obj) const { return(object == obj); }
	bool hasSignal(const QByteArray& sig) const { return(signal() == sig); }
	int  seqId(int sigIndex) const { return seqIds[sigIndex]; }
	void clear() { seqIds.clear(); QSignalSpy::clear(); }
	SignalsRecorder::SignalData data(int signalIndex) const;

	// From QObject
	int qt_metacall(QMetaObject::Call call, int methodId, void **a);

private:
	SignalsRecorder* recorder;
	QObject* object;
	QList<int> seqIds;
};

SignalsRecorder::SignalData SignalsRecorder::SignalsRecorderSpy::data(int signalIndex) const
{
	return SignalData(object, signal(), (*this)[signalIndex], seqIds[signalIndex]);
}

int SignalsRecorder::SignalsRecorderSpy::qt_metacall
		(QMetaObject::Call call, int methodId, void **a)
{
	const int countBefore = count();
	const int result = QSignalSpy::qt_metacall(call, methodId, a);
	if (count() > countBefore)
		seqIds.append(recorder->signalReceived(object, signal()));
	return result;
}

SignalsRecorder::SignalData::SignalData(QObject* object, const QByteArray& signal, const QList<QVariant>& params, int seqId)
	: object_(object), signal_(signal), params_(params), seqId_(seqId)
{
}

bool SignalsRecorder::SignalData::hasObject(QObject* object) const
{
	return (object_ == object);
}

bool SignalsRecorder::SignalData::hasSignal(const char* signal) const
{
	return (signal_ == normalizeSignature(signal));
}

bool SignalsRecorder::SignalData::matches(QObject* object, const char* signal) const
{
	return (hasObject(object) && hasSignal(signal));
}

QByteArray SignalsRecorder::SignalData::signal() const
{
	return signal_;
}

SignalsRecorder::SignalsRecorder(QObject *parent)
	: QObject(parent)
	, nextSeqId_(0)
{
}

SignalsRecorder::~SignalsRecorder()
{
	while (!signalSpies_.isEmpty())
		delete signalSpies_.takeFirst();
}

bool SignalsRecorder::listen(QObject* object, const char* signal)
{
	QScopedPointer<SignalsRecorderSpy> newSpy(new SignalsRecorderSpy(this, object, signal));
	if (!newSpy->isValid())
		return false;
	signalSpies_.append(newSpy.data());
	newSpy.take();
	return true;
}

bool SignalsRecorder::listen(const SignalList& signalList)
{
	QList<SignalsRecorderSpy*> newSpies;
	const QList<SignalList::SignalDesc>& signalDescs = signalList.asList();
	signalSpies_.reserve(signalSpies_.size() + signalDescs.size());
	foreach( const SignalList::SignalDesc& signalDesc, signalDescs ) {
        SignalsRecorderSpy* newSpy =
                new SignalsRecorderSpy(this, signalDesc.object, signalDesc.originalSignal.constData());
        if (!newSpy->isValid()) {
            delete newSpy;
            qDeleteAll(newSpies);
			return false;
        }
        newSpies.append(newSpy);
	}
	signalSpies_.append(newSpies);
	return true;
}

bool SignalsRecorder::waitForSignals(const SignalList& expectedSignals, int timeoutMsec)
{
	expectedSignals_ = expectedSignals;
	timeoutTimer_.setSingleShot(true);
	timeoutTimer_.start(timeoutMsec);
	bool timerActive;
	while ((timerActive = timeoutTimer_.isActive()) && !expectedSignals_.isEmpty()) {
		QCoreApplication::sendPostedEvents();
		QCoreApplication::sendPostedEvents(0, QEvent::DeferredDelete);
		QCoreApplication::processEvents();
	}
	timeoutTimer_.stop();
	return timerActive;
}

QList<SignalsRecorder::SignalData> SignalsRecorder::allSignalData() const
{
	QList<SignalData> result;
	foreach(const SignalsRecorderSpy* spy, signalSpies_) {
		for (int sigIndex = 0; sigIndex < spy->count(); ++sigIndex) {
			result.append(spy->data(sigIndex));
		}
	}
	qSort(result.begin(), result.end(), sequenceOrdering);
	return result;
}

QList<SignalsRecorder::SignalData> SignalsRecorder::signalData(QObject* object, const char* signal) const
{
	return signalData(object, normalizeSignature(signal));
}

QList<SignalsRecorder::SignalData> SignalsRecorder::signalData(const SignalList& signalList) const
{
	QList<SignalData> result;
	const QList<SignalList::SignalDesc>& signalDescs = signalList.asList();
	foreach( const SignalList::SignalDesc& signalDesc, signalDescs ) {
		result.append(signalData(signalDesc.object, signalDesc.signal));
	}
	qSort(result.begin(), result.end(), sequenceOrdering);
	return result;
}

void SignalsRecorder::clearAllSignalData()
{
	foreach(SignalsRecorderSpy* spy, signalSpies_) {
		spy->clear();
	}
}

void SignalsRecorder::clearSignalData(QObject* object, const char* signal)
{
	clearSignalData(object, normalizeSignature(signal));
}

void SignalsRecorder::clearSignalData(const SignalList& signalList)
{
	const QList<SignalList::SignalDesc>& signalDescs = signalList.asList();
	foreach( const SignalList::SignalDesc& signalDesc, signalDescs ) {
		clearSignalData(signalDesc.object, signalDesc.signal);
	}
}

int SignalsRecorder::signalReceived(QObject* object, const QByteArray& signal)
{
	expectedSignals_.remove(object, signal);
	return nextSeqId_++;
}

QList<SignalsRecorder::SignalData> SignalsRecorder::signalData(QObject* object, const QByteArray& signal) const
{
	QList<SignalData> result;
	foreach(const SignalsRecorderSpy* spy, signalSpies_) {
		if (spy->hasObject(object) && spy->hasSignal(signal)) {
			for (int sigIndex = 0; sigIndex < spy->count(); ++sigIndex) {
				result.append(spy->data(sigIndex));
			}
		}
	}
	return result;
}

void SignalsRecorder::clearSignalData(QObject* object, const QByteArray& signal)
{
	foreach(SignalsRecorderSpy* spy, signalSpies_) {
		if (spy->hasObject(object) && spy->hasSignal(signal))
			spy->clear();
	}
}

bool equalParameterLists(const QList<QVariant>& lhs, const QList<QVariant>& rhs)
{
    if (lhs.size() != rhs.size())
        return false;

    for (int i = 0; i < lhs.size(); ++i) {
        if (lhs[i].canConvert<QModelIndex>() && rhs[i].canConvert<QModelIndex>()) {
            if (!(lhs[i].value<QModelIndex>() == rhs[i].value<QModelIndex>()))
                return false;
        }
        else {
            if (!(lhs[i] == rhs[i]))
                return false;
        }
    }

    return true;
}

bool operator==(const SignalsRecorder::SignalData& lhs, const SignalsRecorder::SignalData& rhs)
{
    return (lhs.object_ == rhs.object_ && lhs.signal_ == rhs.signal_ &&
        lhs.seqId_ == rhs.seqId_ && equalParameterLists(lhs.params_, rhs.params_));
}

bool operator!=(const SignalsRecorder::SignalData& lhs, const SignalsRecorder::SignalData& rhs)
{
	return !(lhs == rhs);
}

QDebug operator<<(QDebug dbg, const SignalsRecorder::SignalData& sd)
{
	dbg.nospace() << sd.signal() << " with params (";
	bool first = true;
	Q_FOREACH(const QVariant& v, sd.params()) {
		if (!first)
			dbg.nospace() << ", ";
		first = false;
		if (v.canConvert<QModelIndex>()) {
			QModelIndex qmi = v.value<QModelIndex>();
			if (qmi == QModelIndex())
				dbg.nospace() << "QModelIndex()";
			else
				dbg.nospace() << "QModelIndex(" << qmi.row() << ", " << qmi.column() << ")";
		}
		else
			dbg.nospace() << v;
	}
	dbg.nospace() << ")";
	return dbg.space();
}

bool sequenceOrdering(const SignalsRecorder::SignalData& lhs, const SignalsRecorder::SignalData& rhs)
{
	return (lhs.seqId() < rhs.seqId());
}
